Package: taxodist
Title: Taxonomic Distance and Phylogenetic Lineage Computation
Version: 0.5.0
LazyData: true
Authors@R: person("Rodrigo Fonseca", "Villa", email = "rodrigo03.villa@gmail.com",
           role = c("aut", "cre"),
           comment = c(ORCID = "0009-0005-2938-2270"))
Maintainer: Rodrigo Fonseca Villa <rodrigo03.villa@gmail.com>
Description: Computes phylogenetic distances between any two taxa using 
    hierarchical lineage data retrieved from The Taxonomicon 
    <http://taxonomicon.taxonomy.nl>, a comprehensive curated classification 
    of all life based on Systema Naturae 2000 (Brands, 1989 
    <http://taxonomicon.taxonomy.nl>). Given any two taxon names, retrieves their 
    full lineages, identifies the most recent common ancestor (MRCA), 
    and computes a dissimilarity index based on lineage depth. Outputs 
    native dist objects, enabling direct integration with the R statistical 
    ecosystem for hierarchical clustering, principal coordinate analysis (PCoA), 
    and multivariate ecological analyses. Supports individual distance queries, 
    pairwise distance matrices, clade filtering, and lineage utilities.
Language: en-US
License: GPL (>= 3)
URL: https://github.com/rodrigosqrt3/taxodist
BugReports: https://github.com/rodrigosqrt3/taxodist/issues
Encoding: UTF-8
RoxygenNote: 7.3.3
SystemRequirements: internet access
Depends: R (>= 4.1.0)
Imports: httr (>= 1.4.0), rvest (>= 1.0.0), stringr (>= 1.4.0), purrr
        (>= 0.3.0), cli (>= 3.0.0), utils, stats, graphics
Suggests: testthat (>= 3.0.0), mockery, knitr, rmarkdown, xml2, ape,
        vegan, webmockr
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-06-02 19:13:41 UTC; rodri
Author: Rodrigo Fonseca Villa [aut, cre] (ORCID:
    <https://orcid.org/0009-0005-2938-2270>)
Repository: CRAN
Date/Publication: 2026-06-02 19:30:02 UTC
Built: R 4.5.2; ; 2026-06-02 21:29:38 UTC; unix

Package: taxodist
Title: Taxonomic Hierarchy Distances and Lineage Analysis
Version: 0.6.0
LazyData: true
Authors@R: person("Rodrigo Fonseca", "Villa", email = "rodrigo03.villa@gmail.com",
           role = c("aut", "cre"),
           comment = c(ORCID = "0009-0005-2938-2270"))
Maintainer: Rodrigo Fonseca Villa <rodrigo03.villa@gmail.com>
Description: Computes distances between taxonomic hierarchy nodes using lineage
    data retrieved from The Taxonomicon <http://taxonomicon.taxonomy.nl>. For
    distinct nodes, distance is defined as the reciprocal of the depth of their
    most recent common ancestor; identical nodes have distance zero. This
    definition yields an ultrametric within each connected hierarchy. Functions
    are provided for lineage retrieval and comparison, clade membership,
    pairwise and matrix distance calculation, hierarchical clustering,
    principal coordinates analysis, and cache management. Distance matrices are
    returned as base R 'dist' objects. The distances represent classification
    depth rather than evolutionary time or phylogenetic branch length.
Language: en-US
License: GPL (>= 3)
URL: https://github.com/rodrigosqrt3/taxodist
BugReports: https://github.com/rodrigosqrt3/taxodist/issues
Encoding: UTF-8
RoxygenNote: 7.3.3
SystemRequirements: internet access
Depends: R (>= 4.1.0)
Imports: httr (>= 1.4.0), rvest (>= 1.0.0), stringr (>= 1.4.0), purrr
        (>= 0.3.0), cli (>= 3.0.0), utils, stats, graphics
Suggests: testthat (>= 3.0.0), mockery, knitr, rmarkdown, xml2, ape,
        vegan, webmockr
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-08-19 12:14:14 UTC; rodri
Author: Rodrigo Fonseca Villa [aut, cre] (ORCID:
    <https://orcid.org/0009-0005-2938-2270>)
Repository: CRAN
Date/Publication: 2026-08-19 13:40:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(plot,taxodist_cluster)
S3method(plot,taxodist_ord)
S3method(print,taxodist_focal)
S3method(print,taxodist_path)
S3method(print,taxodist_result)
S3method(summary,taxodist_ord)
export(cache_info)
export(check_coverage)
export(clear_cache)
export(closest_relative)
export(compare_lineages)
export(distance_matrix)
export(filter_clade)
export(focal_distances)
export(get_lineage)
export(get_lineage_by_id)
export(get_taxonomicon_id)
export(is_member)
export(lineage_depth)
export(load_cache)
export(mrca)
export(save_cache)
export(shared_clades)
export(taxo_cluster)
export(taxo_distance)
export(taxo_heatmap)
export(taxo_ordinate)
export(taxo_path)
export(taxo_search)
importFrom(httr,GET)
importFrom(httr,add_headers)
importFrom(httr,content)
importFrom(httr,status_code)
importFrom(rvest,html_attr)
importFrom(rvest,html_nodes)
importFrom(rvest,html_text)
importFrom(rvest,read_html)
importFrom(stringr,str_extract)
importFrom(stringr,str_remove)
importFrom(stringr,str_remove_all)
importFrom(stringr,str_trim)

# taxodist

[![CRAN status](https://www.r-pkg.org/badges/version/taxodist)](https://CRAN.R-project.org/package=taxodist)
[![R-CMD-check](https://github.com/rodrigosqrt3/taxodist/actions/workflows/r.yml/badge.svg)](https://github.com/rodrigosqrt3/taxodist/actions/workflows/r.yml)
[![Coverage](https://app.codecov.io/gh/rodrigosqrt3/taxodist/branch/main/graph/badge.svg)](https://app.codecov.io/gh/rodrigosqrt3/taxodist)

**Taxonomic distance and phylogenetic lineage computation for any taxon on Earth.**

`taxodist` retrieves full hierarchical lineages from [The Taxonomicon](http://taxonomicon.taxonomy.nl) and computes a tree metric distance between any two taxa: a pair of dinosaurs, a dinosaur and a fungus, two species of fly, or an oak tree and a human.

## Installation

```r
devtools::install_github("rodrigosqrt3/taxodist")
```

## Basic usage

```r
library(taxodist)

# Get a full lineage
get_lineage("Tyrannosaurus")

# Distance between two taxa
taxo_distance("Tyrannosaurus", "Velociraptor")

# Most recent common ancestor
mrca("Tyrannosaurus", "Triceratops")   # "Dinosauria"
mrca("Tyrannosaurus", "Homo")          # "Amniota"

# Pairwise distance matrix
theropods <- c("Tyrannosaurus", "Velociraptor", "Spinosaurus", "Allosaurus")
distance_matrix(theropods)

# Filter taxa by clade
taxa <- c("Tyrannosaurus", "Triceratops", "Homo", "Quercus")
filter_clade(taxa, "Dinosauria")

# Get the path between two taxa
taxo_path("Tyrannosaurus", "Velociraptor")

# Save and restore the lineage cache across sessions
save_cache("my_cache.rds")
load_cache("my_cache.rds")
```

## The distance metric

`taxodist` measures relatedness by asking a single question: how deep is the most recent common ancestor (MRCA)?

$$d(A, B) = \frac{1}{\text{depth}(\text{MRCA}(A, B))}$$

The deeper the shared ancestor, the smaller the distance and the more related the two taxa are. A shallow MRCA means the two taxa diverged early; a deep MRCA means they share a long common history. The metric returns 0 when one taxon is ancestral to the other, and satisfies the triangle inequality.

The Taxonomicon provides substantially deeper lineage resolution than other programmatic sources, e.g., *Tyrannosaurus* has over 70 nodes in its lineage, which is what makes the distances meaningful across all of life.

## Caching

Lineages are cached in memory automatically during a session. To persist the
cache across sessions and avoid redundant network requests, use
`save_cache("file.rds")` and `load_cache("file.rds")`.

## Data source

All lineage data is sourced from **The Taxonomicon** (taxonomy.nl), based on *Systema Naturae 2000* by Sheila J. Brands (1989 onwards). Please cite this resource in any published work using `taxodist`:

> Brands, S.J. (1989 onwards). *Systema Naturae 2000*. Amsterdam, The Netherlands. Retrieved from The Taxonomicon, http://taxonomicon.taxonomy.nl.

## Contributing

Found a taxon with an incorrect lineage? Please [open an issue](https://github.com/rodrigosqrt3/taxodist/issues),
lineage corrections are the most valuable contribution to this package.

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
library(taxodist)

## ----lineage------------------------------------------------------------------
# lin <- get_lineage("Tyrannosaurus")
# tail(lin, 8)
# #> [1] "Avetheropoda"     "Coelurosauria"    "Tyrannoraptora"   "Tyrannosauroidea"
# #> [5] "Tyrannosauridae"  "Tyrannosaurinae"  "Tyrannosaurini"   "Tyrannosaurus"

## ----lineage-species----------------------------------------------------------
# lin <- get_lineage("Drosophila melanogaster")
# tail(lin, 4)
# #> [1] "Ephydroidea"             "Drosophilidae"           "Drosophilinae"
# #> [4] "Drosophila melanogaster"

## ----distance-----------------------------------------------------------------
# result <- taxo_distance("Tyrannosaurus", "Velociraptor")
# print(result)
# #> -- Taxonomic Distance --
# #>
# #> * Tyrannosaurus vs Velociraptor
# #>   Distance : 0.0153846153846154
# #>   MRCA : Tyrannoraptora (depth 65)
# #>   Depth A : 70
# #>   Depth B : 73

## ----distance-far-------------------------------------------------------------
# taxo_distance("Tyrannosaurus", "Homo")$distance        # 0.02777778
# taxo_distance("Tyrannosaurus", "Drosophila")$distance  # 0.06666667
# taxo_distance("Tyrannosaurus", "Quercus")$distance     # 0.25
# taxo_distance("Escherichia", "Homo")$distance          # 1

## ----mrca---------------------------------------------------------------------
# mrca("Tyrannosaurus", "Velociraptor")  # "Tyrannoraptora"
# mrca("Tyrannosaurus", "Triceratops")   # "Dinosauria"
# mrca("Tyrannosaurus", "Homo")          # "Amniota"
# mrca("Tyrannosaurus", "Drosophila")    # "Nephrozoa"
# mrca("Tyrannosaurus", "Quercus")       # "discaria"

## ----matrix-------------------------------------------------------------------
# taxa <- c(
#   "Tyrannosaurus", "Carnotaurus", "Triceratops",
#   "Parasaurolophus", "Stegosaurus", "Brachiosaurus",
#   "Homo sapiens", "Homo neanderthalensis", "Pan troglodytes",
#   "Panthera leo", "Canis lupus",
#   "Ornithorhynchus anatinus",
#   "Loxodonta africana",
#   "Struthio camelus",
#   "Aptenodytes forsteri",
#   "Ara ararauna",
#   "Crocodylus niloticus",
#   "Chelonia mydas",
#   "Ambystoma mexicanum",
#   "Octopus vulgaris",
#   "Carcharodon carcharias",
#   "Balaenoptera musculus",
#   "Drosophila melanogaster",
#   "Apis mellifera",
#   "Arabidopsis thaliana",
#   "Quercus robur",
#   "Ginkgo biloba",
#   "Welwitschia mirabilis",
#   "Saccharomyces cerevisiae",
#   "Escherichia coli",
#   "Bacillus subtilis",
#   "Plasmodium falciparum"
# )
# mat <- distance_matrix(taxa)
# print(mat)

## ----clustering---------------------------------------------------------------
# cl <- taxo_cluster(taxa)
# plot(cl)

## ----PCoA---------------------------------------------------------------------
# ord <- taxo_ordinate(taxa)
# summary(ord)
# plot(ord)

## -----------------------------------------------------------------------------
# taxo_heatmap(taxa)

## ----closest------------------------------------------------------------------
# closest_relative(
#   "Carnotaurus",
#   c("Aucasaurus", "Velociraptor", "Triceratops",
#     "Brachiosaurus", "Homo sapiens", "Apis mellifera")
# )
# #>            taxon   distance
# #> 1     Aucasaurus 0.01515152
# #> 2   Velociraptor 0.01666667
# #> 4  Brachiosaurus 0.01754386
# #> 3    Triceratops 0.01818182
# #> 5   Homo sapiens 0.02777778
# #> 6 Apis mellifera 0.06666667

## ----compare------------------------------------------------------------------
# compare_lineages("Carnotaurus", "Tyrannosaurus")
# #> -- Lineage Comparison --
# #> MRCA: Averostra at depth 60
# #>
# #> Shared lineage (60 nodes):
# #>   Biota ... Theropoda
# #>
# #> Carnotaurus only (7 nodes):
# #> Ceratosauria
# #> Neoceratosauria
# #> Abelisauroidea
# #> Abelisauria
# #> Abelisauridae
# #> Carnotaurinae
# #> Carnotaurus
# #>
# #> Tyrannosaurus only (10 nodes):
# #> Tetanurae
# #> Orionides
# #> ...

## ----shared-------------------------------------------------------------------
# # what do a fly and a beetle have in common?
# shared_clades("Drosophila melanogaster", "Tribolium castaneum")
# # returns their shared lineage from Biota down to their MRCA
# 
# # what do T. rex and a rose share?
# shared_clades("Tyrannosaurus rex", "Rosa agrestis")

## ----membership---------------------------------------------------------------
# is_member("Tyrannosaurus", "Theropoda")          # TRUE
# is_member("Carnotaurus", "Abelisauridae")        # TRUE
# is_member("Triceratops", "Theropoda")            # FALSE
# is_member("Homo sapiens", "Amniota")             # TRUE
# is_member("Drosophila melanogaster", "Insecta")  # TRUE
# is_member("Quercus robur", "Animalia")           # FALSE

## ----filter-------------------------------------------------------------------
# taxa <- c("Tyrannosaurus", "Carnotaurus", "Triceratops",
#           "Velociraptor", "Homo sapiens", "Drosophila melanogaster",
#           "Quercus robur", "Saccharomyces cerevisiae")
# 
# filter_clade(taxa, "Dinosauria")
# #> [1] "Tyrannosaurus" "Carnotaurus"   "Triceratops"   "Velociraptor"
# 
# filter_clade(taxa, "Theropoda")
# #> [1] "Tyrannosaurus" "Carnotaurus"   "Velociraptor"
# 
# filter_clade(taxa, "Animalia")
# #> [1] "Tyrannosaurus"          "Carnotaurus"
# #> [3] "Triceratops"            "Velociraptor"
# #> [5] "Homo sapiens"           "Drosophila melanogaster"

## ----coverage-----------------------------------------------------------------
# taxa <- c("Tyrannosaurus", "Velociraptor", "Apis mellifera", "Fakeosaurus")
# check_coverage(taxa)
# #>  Tyrannosaurus  Velociraptor  Apis mellifera    Fakeosaurus
# #>           TRUE          TRUE            TRUE          FALSE

## ----cache--------------------------------------------------------------------
# clear_cache()

## ----save-cache---------------------------------------------------------------
# save_cache("my_taxa_cache.rds") # at the end of a session
# 
# load_cache("my_taxa_cache.rds") # at the start of the next session, before any distance calls

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.align = "center"
)

library(taxodist)

if (is.null(taxobase$matrix)) {
  stop(
    "The installed taxobase object does not contain the reference matrix. ",
    "Rebuild data/taxobase.rda before building this vignette."
  )
}

## ----methodology-link, eval=FALSE---------------------------------------------
# vignette("methodological-notes", package = "taxodist")

## ----taxobase-----------------------------------------------------------------
names(taxobase)
taxobase$metadata

## ----stored-lineages----------------------------------------------------------
tail(taxobase$lineage_tyrannosaurus)
tail(taxobase$lineage_homo)

## ----stored-pairwise----------------------------------------------------------
taxobase$pairwise

## ----result-structure---------------------------------------------------------
class(taxobase$pairwise)
names(taxobase$pairwise)

## ----reference-matrix---------------------------------------------------------
reference_matrix <- taxobase$matrix

inherits(reference_matrix, "dist")
attr(reference_matrix, "Size")
head(attr(reference_matrix, "Labels"))

round(
  as.matrix(reference_matrix)[1:6, 1:6],
  digits = 4
)

## ----live-lineage, eval=FALSE-------------------------------------------------
# get_lineage("Tyrannosaurus")
# get_lineage("Drosophila melanogaster")

## ----lineage-id, eval=FALSE---------------------------------------------------
# get_lineage("67263")

## ----live-distance, eval=FALSE------------------------------------------------
# result <- taxo_distance(
#   "Tyrannosaurus",
#   "Velociraptor"
# )
# 
# result
# result$distance
# result$mrca

## ----live-mrca, eval=FALSE----------------------------------------------------
# mrca("Tyrannosaurus", "Velociraptor")
# mrca("Tyrannosaurus", "Triceratops")
# mrca("Tyrannosaurus", "Homo")

## ----live-path, eval=FALSE----------------------------------------------------
# taxo_path(
#   "Tyrannosaurus",
#   "Triceratops"
# )
# 
# taxo_path(
#   "Dinosauria",
#   "Tyrannosaurus"
# )

## ----live-matrix, eval=FALSE--------------------------------------------------
# taxa <- c(
#   "Tyrannosaurus",
#   "Velociraptor",
#   "Spinosaurus",
#   "Allosaurus"
# )
# 
# mat <- distance_matrix(
#   taxa,
#   progress = TRUE
# )
# 
# mat

## ----stored-closest-----------------------------------------------------------
taxobase$closest

## ----live-closest, eval=FALSE-------------------------------------------------
# closest_relative(
#   "Carnotaurus",
#   c(
#     "Aucasaurus",
#     "Velociraptor",
#     "Triceratops",
#     "Brachiosaurus"
#   )
# )

## ----live-focal, eval=FALSE---------------------------------------------------
# focal_distances(
#   focal = "Tyrannosaurus",
#   community = c(
#     "Velociraptor",
#     "Triceratops",
#     "Spinosaurus"
#   )
# )

## ----live-membership, eval=FALSE----------------------------------------------
# is_member("Tyrannosaurus", "Dinosauria")
# is_member("Tyrannosaurus", "Theropoda")
# is_member("Tyrannosaurus", "Ornithischia")

## ----stored-filter------------------------------------------------------------
taxobase$filter

## ----live-filter, eval=FALSE--------------------------------------------------
# taxa <- c(
#   "Tyrannosaurus",
#   "Carnotaurus",
#   "Triceratops",
#   "Velociraptor",
#   "Homo",
#   "Drosophila"
# )
# 
# filter_clade(taxa, "Dinosauria")
# filter_clade(taxa, "Theropoda")

## ----live-comparison, eval=FALSE----------------------------------------------
# shared_clades(
#   "Tyrannosaurus",
#   "Triceratops"
# )
# 
# compare_lineages(
#   "Carnotaurus",
#   "Tyrannosaurus"
# )

## ----analysis-matrix----------------------------------------------------------
labels <- attr(reference_matrix, "Labels")[1:8]

example_matrix <- stats::as.dist(
  as.matrix(reference_matrix)[labels, labels]
)

example_matrix

## ----clustering, fig.width=7, fig.height=5------------------------------------
clustering <- taxo_cluster(
  example_matrix,
  method = "average"
)

plot(
  clustering,
  main = "Taxonomic hierarchy distance clustering",
  xlab = "",
  sub = ""
)

## ----ordination, fig.width=7, fig.height=5------------------------------------
ordination <- taxo_ordinate(
  example_matrix,
  k = 2
)

summary(ordination)
plot(
  ordination,
  main = "Taxonomic hierarchy distance space"
)

## ----heatmap, fig.width=7, fig.height=6---------------------------------------
taxo_heatmap(
  example_matrix,
  main = "Taxonomic hierarchy distance matrix"
)

## ----statistics-link, eval=FALSE----------------------------------------------
# vignette("statistical-applications", package = "taxodist")

## ----live-search, eval=FALSE--------------------------------------------------
# taxo_search("Panthera")
# taxo_search("Bacteria")

## ----live-coverage, eval=FALSE------------------------------------------------
# taxa <- c(
#   "Tyrannosaurus",
#   "Velociraptor",
#   "Quercus",
#   "Not_a_real_taxon"
# )
# 
# check_coverage(taxa)

## ----live-cache, eval=FALSE---------------------------------------------------
# cache_info()
# clear_cache()

## ----persistent-cache, eval=FALSE---------------------------------------------
# save_cache("taxodist-cache.rds")
# clear_cache()
# load_cache("taxodist-cache.rds")

## ----citation, eval=FALSE-----------------------------------------------------
# citation("taxodist")

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.align = "center"
)

library(taxodist)

vegan_available <- requireNamespace("vegan", quietly = TRUE)
ape_available <- requireNamespace("ape", quietly = TRUE)

if (is.null(taxobase$statistical_matrix)) {
  stop(
    "The installed taxobase object does not contain statistical_matrix. ",
    "Rebuild data/taxobase.rda before building this vignette."
  )
}

## ----data---------------------------------------------------------------------
taxa <- taxobase$statistical_taxa
mat <- taxobase$statistical_matrix

taxobase$metadata
length(taxa)
inherits(mat, "dist")
identical(attr(mat, "Labels"), taxa)

## ----taxa---------------------------------------------------------------------
taxa

## ----matrix-summary-----------------------------------------------------------
summary(as.vector(mat))
range(mat)

round(
  as.matrix(mat)[1:6, 1:6],
  digits = 5
)

## ----clustering, fig.width=7, fig.height=5------------------------------------
clustering <- taxo_cluster(mat, method = "average")

clustering$hclust
summary(clustering$hclust$height)

plot(
  clustering,
  main = "Average-linkage clustering of taxonomic hierarchy distances",
  xlab = "",
  sub = ""
)

## ----ape-tree, eval=ape_available, fig.width=7, fig.height=5------------------
tree <- ape::as.phylo(clustering$hclust)

plot(
  tree,
  main = "Tree representation of a taxonomic distance dendrogram",
  cex = 0.8
)

## ----ape-export, eval=FALSE---------------------------------------------------
# ape::write.tree(tree, file = "taxonomic-distance-dendrogram.nwk")

## ----pcoa---------------------------------------------------------------------
ordination <- taxo_ordinate(mat, k = 2)

positive_eigenvalues <- ordination$eig[ordination$eig > 0]
variance_percent <- 100 * positive_eigenvalues / sum(positive_eigenvalues)

ordination_summary <- data.frame(
  Axis = c("PC1", "PC2"),
  Eigenvalue = ordination$eig[1:2],
  Variance_percent = variance_percent[1:2],
  Cumulative_percent = cumsum(variance_percent)[1:2]
)

ordination_summary
100 * ordination$GOF[1]
sum(ordination$eig < -sqrt(.Machine$double.eps))

## ----pcoa-plot, fig.width=7, fig.height=5-------------------------------------
plot(
  ordination,
  main = "PCoA of taxonomic hierarchy distances"
)

## ----community----------------------------------------------------------------
comm <- matrix(
  c(
    1,1,0,0,1,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,1,1,1,1,1
  ),
  nrow = 3,
  byrow = TRUE,
  dimnames = list(
    c("community_A", "community_B", "community_C"),
    taxa
  )
)

comm

## ----taxondive, eval=vegan_available------------------------------------------
taxonomic_diversity <- vegan::taxondive(comm, mat)

taxonomic_diversity_table <- data.frame(
  Species = taxonomic_diversity$Species,
  Delta = taxonomic_diversity$D,
  Delta_star = taxonomic_diversity$Dstar,
  Lambda_plus = taxonomic_diversity$Lambda,
  Delta_plus = taxonomic_diversity$Dplus,
  SD_Delta_plus = taxonomic_diversity$sd.Dplus
)

taxonomic_diversity_table

## ----mantel, eval=vegan_available---------------------------------------------
set.seed(42)

coordinates <- matrix(
  rnorm(2 * length(taxa)),
  ncol = 2,
  dimnames = list(taxa, c("x", "y"))
)

geographic_distance <- stats::dist(coordinates)

mantel_result <- vegan::mantel(
  mat,
  geographic_distance,
  method = "pearson",
  permutations = 999
)

mantel_result

## ----groups-------------------------------------------------------------------
groups <- factor(c(
  "xenarthran", "xenarthran",
  "carnivoran",
  "ungulate",
  "marsupial",
  "primate", "primate",
  "carnivoran", "carnivoran", "carnivoran",
  "cetacean", "cetacean",
  "sirenian",
  "ungulate", "ungulate"
))

table(groups)

## ----permanova, eval=vegan_available------------------------------------------
set.seed(42)

permanova_result <- vegan::adonis2(
  mat ~ groups,
  permutations = 999
)

permanova_result

## ----dispersion, eval=vegan_available-----------------------------------------
dispersion <- vegan::betadisper(mat, groups)

anova(dispersion)

set.seed(42)
vegan::permutest(
  dispersion,
  permutations = 999
)

## ----live-workflow, eval=FALSE------------------------------------------------
# library(taxodist)
# 
# taxa <- c(
#   "Tyrannosaurus",
#   "Velociraptor",
#   "Spinosaurus",
#   "Allosaurus"
# )
# 
# coverage <- check_coverage(taxa)
# stopifnot(all(coverage))
# 
# mat <- distance_matrix(taxa)
# 
# clustering <- taxo_cluster(mat, method = "average")
# ordination <- taxo_ordinate(mat, k = 2)
# 
# plot(clustering)
# plot(ordination)

## ----citations, eval=FALSE----------------------------------------------------
# citation("taxodist")
# citation("vegan")
# citation("ape")

#' Print method for taxodist distance results
#'
#' @param x A list returned by [taxo_distance()].
#' @param ... Additional arguments (ignored).
#' @return Invisibly returns \code{x}. Called for side effects (printing).
#' @export
print.taxodist_result <- function(x, ...) {
  cli::cli_h2("Taxonomic Distance")
  cli::cli_bullets(c(
    "*" = "{x$taxon_a} vs {x$taxon_b}",
    " " = "Distance : {x$distance}",
    " " = "MRCA     : {x$mrca} (depth {x$mrca_depth})",
    " " = "Depth A  : {x$depth_a}",
    " " = "Depth B  : {x$depth_b}"
  ))
  invisible(x)
}

#' Compare lineages of two taxa side by side
#'
#' Prints the lineages of two taxa aligned at their most recent common
#' ancestor, making the point of divergence easy to identify.
#'
#' @param taxon_a A character string giving the first taxon name.
#' @param taxon_b A character string giving the second taxon name.
#' @param verbose Logical. If `TRUE`, prints progress messages. Default `FALSE`.
#'
#' @return Invisibly returns a list with elements `lineage_a`, `lineage_b`,
#'   and `mrca_depth`.
#'
#' @export
#' @examples
#' \donttest{
#' compare_lineages("Tyrannosaurus", "Velociraptor")
#' compare_lineages("Tyrannosaurus", "Triceratops")
#' }
compare_lineages <- function(taxon_a, taxon_b, verbose = FALSE) {
  lin_a <- get_lineage(taxon_a, verbose = verbose)
  lin_b <- get_lineage(taxon_b, verbose = verbose)

  if (is.null(lin_a) || is.null(lin_b)) {
    cli::cli_alert_danger("Could not retrieve one or both lineages")
    return(invisible(NULL))
  }

  result <- .compute_distance(lin_a, lin_b, taxon_a, taxon_b)
  mrca_d <- result$mrca_depth

  cli::cli_h2("Lineage Comparison")
  cli::cli_text("MRCA: {result$mrca} at depth {mrca_d}")
  cli::cli_text("")

  # shared trunk
  if (mrca_d > 0) {
    cli::cli_text("{.strong Shared lineage ({mrca_d} nodes):}")
    for (node in lin_a[seq_len(mrca_d)]) {
      cli::cli_bullets(c(" " = "{node}"))
    }
  }

  # divergence
  cli::cli_text("")
  cli::cli_text("{.strong {taxon_a} only ({length(lin_a) - mrca_d} nodes):}")
  if (length(lin_a) > mrca_d) {
    for (node in lin_a[(mrca_d + 1):length(lin_a)]) {
      cli::cli_bullets(c(">" = "{node}"))
    }
  }

  cli::cli_text("")
  cli::cli_text("{.strong {taxon_b} only ({length(lin_b) - mrca_d} nodes):}")
  if (length(lin_b) > mrca_d) {
    for (node in lin_b[(mrca_d + 1):length(lin_b)]) {
      cli::cli_bullets(c(">" = "{node}"))
    }
  }

  invisible(list(
    lineage_a  = lin_a,
    lineage_b  = lin_b,
    mrca_depth = mrca_d
  ))
}

#' List all clades shared between two taxa
#'
#' Returns the vector of clade names forming the shared trunk of two taxa's
#' lineages, from root down to (and including) their MRCA.
#'
#' @param taxon_a A character string giving the first taxon name.
#' @param taxon_b A character string giving the second taxon name.
#' @param verbose Logical. If `TRUE`, prints progress messages. Default `FALSE`.
#'
#' @return A character vector of shared clade names ordered from root to MRCA,
#'   or `NULL` if either taxon cannot be found.
#'
#' @export
#' @examples
#' \donttest{
#' shared_clades("Tyrannosaurus", "Velociraptor")
#' shared_clades("Tyrannosaurus", "Homo")
#' }
shared_clades <- function(taxon_a, taxon_b, verbose = FALSE) {
  lin_a <- get_lineage(taxon_a, verbose = verbose)
  lin_b <- get_lineage(taxon_b, verbose = verbose)
  if (is.null(lin_a) || is.null(lin_b)) return(NULL)

  result <- .compute_distance(lin_a, lin_b, taxon_a, taxon_b)
  if (result$mrca_depth == 0) return(character(0))
  lin_a[seq_len(result$mrca_depth)]
}

#' Test whether one taxon is nested within another
#'
#' Returns `TRUE` if `taxon` is a member of `clade` — i.e., if the complete
#' clade name appears in the taxon's lineage. Matching is case-insensitive and
#' ignores leading and trailing whitespace.
#'
#' @param taxon A character string giving the taxon name to test.
#' @param clade A character string giving the clade name to test membership in.
#' @param verbose Logical. If `TRUE`, prints progress messages. Default `FALSE`.
#'
#' @return A logical value, or `NULL` if the taxon cannot be found.
#'
#' @export
#' @examples
#' \donttest{
#' is_member("Tyrannosaurus", "Theropoda")    # TRUE
#' is_member("Triceratops", "Theropoda")      # FALSE
#' is_member("Homo", "Amniota")               # TRUE
#' }
is_member <- function(taxon, clade, verbose = FALSE) {
  lin <- get_lineage(taxon, verbose = verbose)
  if (is.null(lin)) return(NULL)
  normalized_lineage <- tolower(trimws(lin))
  normalized_clade   <- tolower(trimws(clade))
  any(normalized_lineage %in% normalized_clade)
}

#' Filter a vector of taxa to those belonging to a given clade
#'
#' Given a vector of taxon names and a clade name, returns only those taxa
#' whose lineage includes the specified clade.
#'
#' @param taxa A character vector of taxon names.
#' @param clade A character string giving the clade to filter by.
#' @param verbose Logical. If `TRUE`, prints progress messages. Default `FALSE`.
#'
#' @return A character vector of taxa that are members of the specified clade.
#'
#' @export
#' @examples
#' \donttest{
#' taxa <- c("Tyrannosaurus", "Triceratops", "Velociraptor",
#'           "Brachiosaurus", "Homo")
#' filter_clade(taxa, "Theropoda")
#' filter_clade(taxa, "Dinosauria")
#' }
filter_clade <- function(taxa, clade, verbose = FALSE) {
  keep <- purrr::map_lgl(taxa, function(t) {
    result <- is_member(t, clade, verbose = verbose)
    if (is.null(result)) FALSE else result
  })
  taxa[keep]
}

#' Plot method for taxodist_cluster objects
#'
#' Plots the hierarchical clustering dendrogram produced by [taxo_cluster()].
#'
#' @param x A `taxodist_cluster` object from [taxo_cluster()].
#' @param main Plot title. Default: `"Taxonomic Clustering"`.
#' @param xlab X-axis label. Default: `""`.
#' @param sub Subtitle. Default: clustering method used.
#' @param ... Additional arguments passed to [stats::plot.hclust()].
#'
#' @return Invisibly returns `x`. Called for side effects (plotting).
#' @export
#' @examples
#' \donttest{
#' taxa <- c("Tyrannosaurus", "Velociraptor", "Triceratops", "Brachiosaurus")
#' cl <- taxo_cluster(taxa)
#' plot(cl)
#' }
plot.taxodist_cluster <- function(x, main = "Taxonomic Clustering",
                                  xlab = "",
                                  sub  = paste("Method:", if(!is.null(x$hclust)) x$hclust$method else "None"),
                                  ...) {
  if (is.null(x$hclust)) return(invisible(x))
  plot(x$hclust, main = main, xlab = xlab, sub = sub, ...)
  invisible(x)
}

#' Plot method for taxodist_ord objects
#'
#' Plots the PCoA ordination produced by [taxo_ordinate()].
#'
#' @param x A `taxodist_ord` object from [taxo_ordinate()].
#' @param main Plot title. Default: `"Taxonomic Ordination (PCoA)"`.
#' @param xlab X-axis label. Default: `"PC1"`.
#' @param ylab Y-axis label. Default: `"PC2"`.
#' @param labels Character vector of labels. Default: rownames of points.
#' @param ... Additional arguments passed to [graphics::text()].
#'
#' @return Invisibly returns `x`. Called for side effects (plotting).
#' @export
#' @examples
#' \donttest{
#' taxa <- c("Tyrannosaurus", "Velociraptor", "Triceratops", "Brachiosaurus")
#' ord <- taxo_ordinate(taxa)
#' plot(ord)
#' }
plot.taxodist_ord <- function(x, main   = "Taxonomic Ordination (PCoA)",
                              xlab   = "PC1",
                              ylab   = "PC2",
                              labels = if(!is.null(x$points)) rownames(x$points) else NULL,
                              ...) {
  if (is.null(x$points)) return(invisible(x))
  gof <- round(x$GOF[1], 3)
  plot_points <- if (ncol(x$points) == 1L) {
    cbind(PC1 = x$points[, 1], PC2 = 0)
  } else {
    x$points[, 1:2, drop = FALSE]
  }
  graphics::plot(plot_points, type = "n",
                 main = paste0(main, "  (GOF = ", gof, ")"),
                 xlab = xlab, ylab = ylab)
  graphics::text(plot_points, labels = labels, ...)
  invisible(x)
}

#' Summary method for taxodist_ord objects
#'
#' Prints a summary of the PCoA ordination, including the goodness-of-fit
#' and the proportion of variance explained by each principal coordinate.
#'
#' @param object A \code{taxodist_ord} object from \code{\link{taxo_ordinate}}.
#' @param ... Additional arguments (ignored).
#'
#' @return Invisibly returns a data frame containing the eigenvalues and
#'   variance explained by each dimension.
#' @export
summary.taxodist_ord <- function(object, ...) {
  if (is.null(object$points)) return(invisible(NULL))

  cli::cli_h2("Taxonomic Ordination Summary (PCoA)")
  gof <- round(object$GOF[1] * 100, 2)
  cli::cli_text("Goodness-of-Fit (GOF): {gof}%")
  cli::cli_text("")

  if (!is.null(object$eig)) {
    eig_pos <- object$eig[object$eig > 0]
    var_exp <- (eig_pos / sum(eig_pos)) * 100
    k <- ncol(object$points)

    df <- data.frame(
      Axis           = paste0("PC", 1:k),
      Eigenvalue     = object$eig[1:k],
      Variance_Pct   = var_exp[1:k],
      Cumulative_Pct = cumsum(var_exp)[1:k]
    )
    print(df, row.names = FALSE, digits = 4)
    invisible(df)
  } else {
    cli::cli_alert_warning("Eigenvalues not found in the object.")
    invisible(NULL)
  }
}

#' Plot a taxonomic heatmap
#'
#' Computes pairwise taxonomic distances and plots a heatmap with hierarchical
#' clustering dendrograms on the margins. Darker/hotter colors typically
#' represent smaller distances (closer relatives).
#'
#' @param taxa A character vector of taxon names, or a \code{dist} object from
#'   \code{\link{distance_matrix}}.
#' @param ... Additional arguments passed to \code{\link[stats]{heatmap}}.
#'
#' @return Invisibly returns the underlying \code{dist} object.
#'   Called primarily for its side effect (plotting).
#'   If fewer than two taxa are supplied, plotting is skipped.
#' @export
#' @examples
#' \donttest{
#' taxa <- c("Tyrannosaurus", "Velociraptor", "Homo", "Panthera", "Quercus")
#' taxo_heatmap(taxa)
#' }
taxo_heatmap <- function(taxa, ...) {
  d <- if (inherits(taxa, "dist")) taxa else distance_matrix(taxa, ...)
  if (any(is.na(d))) {
    cli::cli_warn("Distance matrix contains NA values. Heatmap skipped.")
    return(invisible(d))
  }
  if (any(!is.finite(d))) {
    cli::cli_warn("Distance matrix contains infinite values (no shared ancestor). Heatmap skipped.")
    return(invisible(d))
  }
  if (attr(d, "Size") < 2L) {
    cli::cli_warn("At least two taxa are required for a heatmap. Heatmap skipped.")
    return(invisible(d))
  }
  mat <- as.matrix(d)
  stats::heatmap(mat, symm = TRUE, margins = c(10, 10), ...)
  invisible(d)
}

#' Get the taxonomic path between two taxa
#'
#' Returns the full node-by-node path from one taxon up to their most recent
#' common ancestor (MRCA) and back down to the other taxon. The result is a
#' data frame with one row per node, making it easy to inspect, filter, or
#' pipe into other functions.
#'
#' @param taxon_a A character string giving the first taxon name.
#' @param taxon_b A character string giving the second taxon name.
#' @param verbose Logical. If `TRUE`, prints progress messages. Default `FALSE`.
#'
#' @return A data frame of class `"taxodist_path"` with columns:
#' \describe{
#'   \item{`node`}{Character. The clade or taxon name at this step.}
#'   \item{`depth`}{Integer. The depth of this node in the full lineage of
#'     its side (or the shared lineage for the MRCA).}
#'   \item{`direction`}{Character. One of `"a"` (ascending from taxon A to
#'     MRCA), `"mrca"` (the shared ancestor), or `"b"` (descending from MRCA
#'     to taxon B).}
#' }
#' Returns `NULL` if either taxon cannot be found or if their lineages have no
#' common ancestor.
#'
#' @seealso [mrca()], [shared_clades()], [compare_lineages()]
#' @export
#' @examples
#' \donttest{
#' taxo_path("Tyrannosaurus", "Velociraptor")
#' taxo_path("Tyrannosaurus", "Homo")
#' }
taxo_path <- function(taxon_a, taxon_b, verbose = FALSE) {
  lin_a <- get_lineage(taxon_a, verbose = verbose)
  lin_b <- get_lineage(taxon_b, verbose = verbose)

  if (is.null(lin_a)) {
    cli::cli_alert_danger("Could not retrieve lineage for {taxon_a}")
    return(NULL)
  }
  if (is.null(lin_b)) {
    cli::cli_alert_danger("Could not retrieve lineage for {taxon_b}")
    return(NULL)
  }

  result  <- .compute_distance(lin_a, lin_b, taxon_a, taxon_b)
  mrca_d  <- result$mrca_depth

  if (mrca_d == 0L) {
    cli::cli_warn("No common ancestor found for {.val {taxon_a}} and {.val {taxon_b}}.")
    return(NULL)
  }

  a_idx <- if (mrca_d < length(lin_a)) {
    seq.int(length(lin_a), mrca_d + 1L)
  } else {
    integer(0)
  }

  side_a <- data.frame(
    node      = lin_a[a_idx],
    depth     = a_idx,
    direction = rep("a", length(a_idx)),
    stringsAsFactors = FALSE
  )

  mrca_row <- data.frame(
    node      = result$mrca,
    depth     = mrca_d,
    direction = "mrca",
    stringsAsFactors = FALSE
  )

  b_idx <- if (mrca_d < length(lin_b)) {
    seq.int(mrca_d + 1L, length(lin_b))
  } else {
    integer(0)
  }

  side_b <- data.frame(
    node      = lin_b[b_idx],
    depth     = b_idx,
    direction = rep("b", length(b_idx)),
    stringsAsFactors = FALSE
  )

  path <- rbind(side_a, mrca_row, side_b)
  rownames(path) <- NULL
  structure(path, class = c("taxodist_path", "data.frame"),
            taxon_a = taxon_a, taxon_b = taxon_b)
}

#' Print method for taxodist_path objects
#'
#' @param x A `taxodist_path` object from [taxo_path()].
#' @param ... Additional arguments (ignored).
#' @return Invisibly returns `x`. Called for side effects (printing).
#' @export
print.taxodist_path <- function(x, ...) {
  taxon_a <- attr(x, "taxon_a")
  taxon_b <- attr(x, "taxon_b")
  cli::cli_h2("Taxonomic Path: {taxon_a} \u2192 {taxon_b}")
  for (i in seq_len(nrow(x))) {
    node <- x$node[i]
    dir  <- x$direction[i]
    dep  <- x$depth[i]
    if (dir == "a") {
      cli::cli_bullets(c(" " = "{.emph {node}}  (depth {dep})  \u2191"))
    } else if (dir == "mrca") {
      cli::cli_bullets(c("*" = "{.strong {node}}  (MRCA, depth {dep})"))
    } else {
      cli::cli_bullets(c(" " = "{.emph {node}}  (depth {dep})  \u2193"))
    }
  }
  invisible(x)
}

#' Print method for taxodist_focal objects
#'
#' @param x A `taxodist_focal` object from [focal_distances()].
#' @param ... Additional arguments (ignored).
#' @return Invisibly returns `x`. Called for side effects (printing).
#' @export
print.taxodist_focal <- function(x, ...) {
  focal <- attr(x, "focal")
  cli::cli_h2("Focal distances from {focal}")
  cli::cli_text("{nrow(x)} taxa \u2022 sorted closest \u2192 most distant")
  cli::cli_text("")
  print.data.frame(x, ...)
  invisible(x)
}

#' Reference Taxonomic Dataset (taxobase)
#'
#' A pre-computed reference dataset for 52 candidate genera distributed across
#' major taxonomic groups. The data were retrieved from The Taxonomicon using
#' the package functions and are intended for examples, documentation, and
#' offline inspection of package output structures.
#'
#' @docType data
#' @name taxobase
#' @usage data(taxobase)
#' @format A structured list containing the following pre-computed components:
#' \describe{
#'   \item{taxa}{Character vector of 52 candidate genera.}
#'   \item{found_taxa}{Character vector of candidates found in The Taxonomicon.}
#'   \item{coverage}{Named logical vector indicating database presence of each taxon.}
#'   \item{matrix}{A pre-computed taxonomic `dist` object for `found_taxa`.}
#'   \item{pairwise}{A pre-computed distance result for *Tyrannosaurus* and *Homo*.}
#'   \item{lineage_homo}{Taxonomic lineage returned for *Homo*.}
#'   \item{lineage_tyrannosaurus}{Taxonomic lineage returned for *Tyrannosaurus*.}
#'   \item{closest}{Result of a closest-relative query for *Tyrannosaurus*.}
#'   \item{filter}{Taxa in `found_taxa` assigned to Theropoda.}
#'   \item{search}{Reference search result for the term `"Bacteria"`.}
#'   \item{statistical_taxa}{Character vector used by the statistical applications vignette.}
#'   \item{statistical_matrix}{Pre-computed `dist` object for `statistical_taxa`.}
#'   \item{metadata}{Data source, generation date, package version, and distance definition.}
#' }
#' @source \url{http://taxonomicon.taxonomy.nl}
#' @examples
#' data(taxobase)
#' length(taxobase$found_taxa)
#' head(labels(taxobase$matrix))
"taxobase"

#' @importFrom httr GET content status_code add_headers
#' @importFrom rvest read_html html_nodes html_text html_attr
#' @importFrom stringr str_remove str_remove_all str_extract str_trim
NULL

regexEscape <- function(x) gsub("([.\\^$*+?{}\\[\\]|()])", "\\\\\\1", x)

# -- Internal cache -----------------------------------------------------------

.taxodist_cache <- new.env(parent = emptyenv())

#' Clear the taxodist lineage cache
#'
#' Clears all cached lineages stored in the current R session. Useful when
#' you suspect cached data is stale or want to force fresh retrieval.
#'
#' @return Invisibly returns `NULL`.
#' @export
#' @examples
#' \donttest{
#' clear_cache()
#' }
clear_cache <- function() {
  rm(list = ls(.taxodist_cache), envir = .taxodist_cache)
  invisible(NULL)
}

#' Save the taxodist lineage cache to disk
#'
#' Serialises the current session cache to an `.rds` file so it can be
#' restored in a future session with [load_cache()]. Useful for
#' reproducibility and for avoiding repeated network requests.
#'
#' @param file Path to the `.rds` file to write.
#'
#' @return Invisibly returns `NULL`.
#' @seealso [load_cache()], [clear_cache()]
#' @export
#' @examples
#' \donttest{
#' tmp <- tempfile(fileext = ".rds")
#' save_cache(tmp)
#' unlink(tmp)
#' }
save_cache <- function(file) {
  data <- as.list(.taxodist_cache)
  saveRDS(data, file = file)
  cli::cli_alert_success("Cache saved to {.file {file}} ({length(data)} entries).")
  invisible(NULL)
}

#' Load a previously saved taxodist cache from disk
#'
#' Restores lineage data saved with [save_cache()] into the current session
#' cache, avoiding network requests for taxa already retrieved in a previous
#' session. The file structure is validated before the current cache is
#' modified.
#'
#' @param file Path to an `.rds` file created by [save_cache()].
#'
#' @return Invisibly returns `NULL`.
#' @seealso [save_cache()], [clear_cache()]
#' @export
#' @examples
#' \donttest{
#' tmp <- tempfile(fileext = ".rds")
#' save_cache(tmp)
#' load_cache(tmp)
#' unlink(tmp)
#' }
load_cache <- function(file) {
  if (!file.exists(file)) {
    cli::cli_abort("Cache file not found: {.file {file}}")
  }
  data <- readRDS(file)

  if (!is.list(data)) {
    cli::cli_abort("Invalid cache file: expected a named list created by {.fn save_cache}.")
  }

  data_names <- names(data)
  invalid_names <- length(data) > 0L && (
    is.null(data_names) || anyNA(data_names) ||
      any(!nzchar(data_names)) || anyDuplicated(data_names) > 0L
  )
  if (invalid_names) {
    cli::cli_abort("Invalid cache file: cache entries must have unique, non-empty names.")
  }

  # An empty cache is stored as list() and legitimately has NULL names.
  # Normalize it so the prefix checks below receive a character vector.
  if (length(data) == 0L) {
    data_names <- character(0)
  }

  id_entries <- startsWith(data_names, "id_")
  invalid_ids <- any(!vapply(data[id_entries], function(x) {
    is.character(x) && length(x) == 1L && !is.na(x) && nzchar(x)
  }, logical(1)))
  if (invalid_ids) {
    cli::cli_abort("Invalid cache file: taxon ID entries must be non-empty character scalars.")
  }

  lineage_entries <- startsWith(data_names, "lin_")
  invalid_lineages <- any(!vapply(data[lineage_entries], is.character, logical(1)))
  if (invalid_lineages) {
    cli::cli_abort("Invalid cache file: lineage entries must be character vectors.")
  }

  list2env(data, envir = .taxodist_cache)
  cli::cli_alert_success("Cache loaded from {.file {file}} ({length(data)} entries).")
  invisible(NULL)
}

#' Inspect the current taxodist lineage cache
#'
#' Reports the number of cached entries, their total memory footprint, and
#' the names of all taxa whose lineages are stored in the current session.
#' Useful for understanding what has already been retrieved before running
#' further computations.
#'
#' @return Invisibly returns a list with elements:
#' \describe{
#'   \item{`n_lineages`}{Integer. Number of cached lineages.}
#'   \item{`n_ids`}{Integer. Number of cached taxon IDs.}
#'   \item{`taxa`}{Character vector of taxa with cached lineages.}
#'   \item{`size_bytes`}{Numeric. Total memory used by the cache.}
#' }
#'
#' @seealso [save_cache()], [load_cache()], [clear_cache()]
#' @export
#' @examples
#' \donttest{
#' get_lineage("Tyrannosaurus")
#' get_lineage("Velociraptor")
#' cache_info()
#' }
cache_info <- function() {
  keys <- ls(.taxodist_cache)

  lin_keys <- keys[startsWith(keys, "lin_")]
  id_keys  <- keys[startsWith(keys, "id_")]

  taxa_names <- sub("^lin_", "", lin_keys)

  size_bytes <- utils::object.size(as.list(.taxodist_cache))

  cli::cli_h2("taxodist Cache")
  cli::cli_bullets(c(
    "*" = "Lineages cached : {length(lin_keys)}",
    "*" = "IDs cached      : {length(id_keys)}",
    "*" = "Memory used     : {format(size_bytes, units = 'auto')}"
  ))

  if (length(taxa_names) > 0) {
    cli::cli_text("")
    cli::cli_text("{.strong Cached taxa:}")
    cli::cli_bullets(stats::setNames(
      paste0("{.val ", taxa_names, "}"),
      rep(" ", length(taxa_names))
    ))
  } else {
    cli::cli_text("")
    cli::cli_alert_info("No lineages cached yet.")
  }

  invisible(list(
    n_lineages = length(lin_keys),
    n_ids      = length(id_keys),
    taxa       = taxa_names,
    size_bytes = as.numeric(size_bytes)
  ))
}

# -- ID lookup ----------------------------------------------------------------

#' Find the Taxonomicon ID for a taxon name
#'
#' Queries The Taxonomicon (taxonomy.nl) to retrieve the internal numeric
#' identifier for a given taxon name. The search filters out non-biological
#' entities such as astronomical objects that may share the same name.
#'
#' @param taxon A character string giving the taxon name to search for.
#'   Typically a genus name (e.g., `"Tyrannosaurus"`) but species and higher
#'   ranks are also supported.
#' @param verbose Logical. If `TRUE`, prints status messages during retrieval.
#'   Default is `FALSE`.
#'
#' @return A character string containing the Taxonomicon numeric ID, or `NULL`
#'   if the taxon is not found.
#'
#' @details
#' The function queries the static search endpoint at
#' `taxonomicon.taxonomy.nl/TaxonList.aspx` and parses the resulting HTML
#' to extract the taxon ID from the hierarchy link. When multiple matches
#' exist (e.g., a genus name shared with an astronomical object), biological
#' entries are prioritised by filtering for entries annotated as dinosaur,
#' reptile, archosaur, animal, plant, fungus, or bacterium.
#'
#' @seealso [get_lineage()], [taxo_distance()]
#'
#' @export
#' @examples
#' \donttest{
#' get_taxonomicon_id("Tyrannosaurus")   # returns "50841"
#' get_taxonomicon_id("Homo")
#' get_taxonomicon_id("Quercus")
#' }
get_taxonomicon_id <- function(taxon, verbose = FALSE) {
  cache_key <- paste0("id_", taxon)
  if (exists(cache_key, envir = .taxodist_cache)) {
    if (verbose) cli::cli_alert_info("Using cached ID for {taxon}")
    return(get(cache_key, envir = .taxodist_cache))
  }

  if (verbose) cli::cli_alert_info("Searching Taxonomicon for {taxon}...")

  url <- paste0(
    "http://taxonomicon.taxonomy.nl/TaxonList.aspx",
    "?subject=Entity&by=ScientificName&search=",
    utils::URLencode(taxon)
  )

  res <- tryCatch(
    httr::GET(url, httr::add_headers("User-Agent" = "taxodist R package/0.3"), httr::timeout(30)),
    error = function(e) NULL
  )

  if (is.null(res) || httr::status_code(res) != 200) {
    cli::cli_warn(c(
      "!" = "Cannot reach The Taxonomicon server.",
      "x" = "The website (taxonomy.nl) appears to be offline or unreachable.",
      "i" = "Please try again later."
    ))
    return(NULL)
  }

  page <- rvest::read_html(httr::content(res, "text", encoding = "UTF-8"))
  rows <- rvest::html_nodes(page, "tr")
  bio_ids <- list()

  for (row in rows) {
    text <- rvest::html_text(row, trim = TRUE)
    if (grepl("\\bastronomical\\b|\\bplanet\\b|\\bMinor planet\\b|\\bcomet\\b|\\basteroid\\b",
              text, ignore.case = TRUE)) next

    links_nodes <- rvest::html_nodes(row, "a[href*='TaxonTree']")
    if (length(links_nodes) == 0) next

    classes <- rvest::html_attr(links_nodes, "class")
    valid_idx <- which(!is.na(classes) & classes == "Valid")
    if (length(valid_idx) == 0) next

    target_link <- links_nodes[[valid_idx[1]]]

    id <- stringr::str_remove(
      stringr::str_extract(
        rvest::html_attr(target_link, "href"), "id=([0-9]+)"
      ), "id="
    )
    if (is.na(id)) next

    text_entry <- trimws(gsub("\\s+", " ", text))
    text_entry <- stringr::str_remove(text_entry, "^N\\|T\\|P\\|R\\|B\\|L\\s*")

    candidate_lin <- get_lineage_by_id(id, clean = TRUE, verbose = FALSE)
    if (is.null(candidate_lin) || !"Biota" %in% candidate_lin) next

    bio_ids <- c(bio_ids, list(list(id = id, text = text_entry)))
  }

  if (length(bio_ids) > 1) {
    matched <- Filter(function(x) {
      lin <- get_lineage_by_id(x$id, clean = TRUE, verbose = FALSE)
      !is.null(lin) && any(grepl(paste0("\\b", regexEscape(taxon), "\\b"), lin, ignore.case = TRUE))
    }, bio_ids)
    # nocov start
    if (length(matched) > 0) bio_ids <- matched
    # nocov end
  }

  if (length(bio_ids) == 0L) {
    if (verbose) cli::cli_alert_warning("{taxon} not found in Taxonomicon")
    return(NULL)
  }

  unique_ids <- unique(sapply(bio_ids, function(x) x$id))
  bio_ids <- lapply(unique_ids, function(uid) {
    matches <- bio_ids[sapply(bio_ids, function(x) x$id == uid)]
    matches[[1]]
  })

  if (length(bio_ids) > 1L) {
    warn_msg <- c(
      "!" = "Multiple valid biological entries found for {.val {taxon}}.",
      "i" = paste0("Using: ", bio_ids[[1]]$text, " (ID: ", bio_ids[[1]]$id, ")"),
      "i" = paste0("To use a different entry, pass its numeric ID directly, e.g. `get_lineage(\"", bio_ids[[2]]$id, "\")`."),
      "i" = "Other available IDs:"
    )
    for (i in 2:length(bio_ids)) {
      warn_msg <- c(warn_msg, "*" = paste0("ID ", bio_ids[[i]]$id, ": ", bio_ids[[i]]$text))
    }
    cli::cli_warn(warn_msg)
  }

  id <- bio_ids[[1]]$id
  assign(cache_key, id, envir = .taxodist_cache)
  if (verbose) cli::cli_alert_success("Found {taxon} with ID {id}")
  return(id)
}

# -- Lineage retrieval --------------------------------------------------------

#' Retrieve the full taxonomic lineage of a taxon
#'
#' Given a Taxonomicon numeric ID, retrieves and parses the complete
#' hierarchical lineage from root (Natura) to the taxon itself. The lineage
#' is returned as a character vector ordered from root to tip.
#'
#' @param taxon_id A numeric or character string giving the Taxonomicon ID.
#'   Obtain this with [get_taxonomicon_id()].
#' @param clean Logical. If `TRUE` (default), removes philosophical root nodes
#'   above `Biota` (i.e., Natura, actualia, Mundus, naturalia) and strips
#'   dagger and superscript markers from names.
#' @param verbose Logical. If `TRUE`, prints status messages. Default `FALSE`.
#'
#' @return A character vector of clade names from root to tip, or `NULL` if
#'   retrieval fails.
#'
#' @details
#' Lineage data is sourced from The Taxonomicon, which is based on
#' Systema Naturae 2000 (Brands, S.J., 1989 onwards). Lineage resolution varies
#' among taxonomic groups and may change when the source classification is
#' updated. Because hierarchy distances depend on lineage depth, analyses
#' should record the retrieval date and package version.
#'
#' @seealso [get_lineage()], [taxo_distance()]
#'
#' @export
#' @examples
#' \donttest{
#' id <- get_taxonomicon_id("Tyrannosaurus")
#' lin <- get_lineage_by_id(id)
#' print(lin)
#' }
get_lineage_by_id <- function(taxon_id, clean = TRUE, verbose = FALSE) {
  if (is.null(taxon_id) || is.na(taxon_id) || taxon_id == "" || !grepl("^[0-9]+$", as.character(taxon_id))) {
    return(NULL)
  }
  cache_key <- paste0("lin_", taxon_id)
  if (exists(cache_key, envir = .taxodist_cache)) {
    if (verbose) cli::cli_alert_info("Using cached lineage for ID {taxon_id}")
    return(get(cache_key, envir = .taxodist_cache))
  }
  url <- paste0(
    "http://taxonomicon.taxonomy.nl/TaxonTree.aspx?id=",
    taxon_id, "&src=0"
  )
  res <- tryCatch(
    httr::GET(
      url,
      httr::add_headers("User-Agent" = "taxodist R package/0.3"),
      httr::timeout(30)
    ),
    error = function(e) NULL
  )
  if (is.null(res) || httr::status_code(res) != 200) {
    if (verbose) cli::cli_alert_warning("Could not retrieve lineage for ID {taxon_id}")
    return(NULL)
  }
  page  <- rvest::read_html(httr::content(res, "text", encoding = "UTF-8"))
  regexEscape <- function(x) gsub("([.\\^$*+?{}\\[\\]|()])", "\\\\\\1", x)

  current_name <- rvest::html_text(
    rvest::html_node(page, "#ctl00_divSubject b"), trim = TRUE
  )
  content_node <- rvest::html_node(page, "#divPageContent")

  if (!is.na(current_name) && !is.null(content_node)) {
    tree_text <- rvest::html_text(content_node, trim = TRUE)
    raw_lines <- strsplit(tree_text, "\n")[[1]]
    raw_lines <- trimws(raw_lines)
    raw_lines <- raw_lines[nzchar(raw_lines)]
    tree_start <- grep("^Natura", raw_lines)[1]
    if (!is.na(tree_start)) raw_lines <- raw_lines[tree_start:length(raw_lines)]
    raw_lines_search <- stringr::str_remove_all(raw_lines, "[\u2020\u1D40]")
    cutoff <- grep(paste0("\\b", regexEscape(current_name), "\\b"), raw_lines_search)[1]
    texts <- if (!is.na(cutoff)) raw_lines[seq_len(cutoff)] else raw_lines
  } else {
    links <- rvest::html_nodes(page, "a[href*='TaxonTree']")
    hrefs_all <- rvest::html_attr(links, "href")
    links <- links[grepl("id=[0-9]", hrefs_all)]
    hrefs <- rvest::html_attr(links, "href")
    own_idx <- which(grepl(paste0("id=", taxon_id, "($|&)"), hrefs))
    if (length(own_idx) > 0) links <- links[seq_len(max(own_idx))]
    texts <- rvest::html_text(links, trim = TRUE)
  }
  lineage <- texts |>
    stringr::str_remove_all("[\u2020\u1D40]") |>
    stringr::str_remove("^\\[crown\\]\\s+(Clade|Grandorder|Order|Superorder|Infraorder|Suborder|Class|Superclass|Subclass|Infraclass|Family|Superfamily|Subfamily|Tribe|Subtribe|Kingdom|Subkingdom|Infrakingdom|Domain|Superkingdom|Phylum|Subphylum|Genus|Species)?\\s*") |>
    stringr::str_remove("^(Clade |Kingdom |Phylum |Superphylum |Subphylum |Infraphylum |Class |Order |Suborder |Infraorder |Parvorder |Grandorder |Magnorder |Cohort |Subcohort |Legion |Family |Subfamily |Tribe |Subtribe |Genus |Species |Subkingdom |Infrakingdom |Superclass |Subclass |Infraclass |Superorder |Superfamily |Domain |Superkingdom |Grade |Subgrade |Supergrade )") |>
    stringr::str_remove("\\s+[A-Z][a-z\u00e1\u00e0\u00e2\u00e3\u00e9\u00e8\u00ea\u00ed\u00ef\u00f3\u00f4\u00f5\u00f6\u00fa\u00fc\u00e7].*$") |>
    stringr::str_remove("\\s+[A-Z]\\.[A-Z]\\..*$") |>
    stringr::str_remove("\\s+auct\\..*$") |>
    stringr::str_remove("\\s+von.*$") |>
    stringr::str_remove("\\s+\\([A-Z][a-z].*$") |>
    stringr::str_remove("\\s+\\(\\d{4}\\).*$") |>
    stringr::str_remove("\\s+\\[.*$") |>
    stringr::str_remove("\\s+[A-Z]\\.$") |>
    stringr::str_remove("\\s+\\([a-z].*$") |>
    stringr::str_remove('\\s+".*$') |>
    stringr::str_remove("^\".*") |>
    stringr::str_trim()
  bare_ranks <- c(
    "Go to", "Superphylum", "Subfamily", "Suborder", "Epifamily",
    "Infraorder", "Superclass", "Subclass", "Superfamily",
    "Subgenus", "Section", "Division", "Candidatus", "Parvphylum",
    "Branch", "Supercohort", "Infracohort", "Subdivision", "Subsection",
    "Grade", "[unranked]", "(Supercluster)", "(Region)",
    "[crown]", ""
  )
  lineage <- lineage[!lineage %in% bare_ranks]
  lineage <- lineage[lineage != "" & !grepl("^\\s*$", lineage)]
  lineage <- lineage[!grepl("^\"", lineage)]
  lineage <- lineage[!grepl("^Population", lineage)]
  lineage <- unique(lineage)
  if (clean) {
    idx_biota <- which(lineage == "Biota")
    if (length(idx_biota) > 0) {
      lineage <- lineage[idx_biota[1]:length(lineage)]
    }
  }
  if (length(lineage) == 0) return(NULL)
  assign(cache_key, lineage, envir = .taxodist_cache)
  lineage
}

#' Retrieve the full taxonomic lineage of a taxon by name
#'
#' A convenience wrapper that combines [get_taxonomicon_id()] and
#' [get_lineage_by_id()] into a single call. Given a taxon name, returns
#' its complete lineage from root to tip.
#'
#' @param taxon A character string giving the taxon name.
#' @param clean Logical. If `TRUE` (default), removes philosophical root nodes
#'   and cleans formatting markers.
#' @param verbose Logical. If `TRUE`, prints progress messages. Default `FALSE`.
#'
#' @return A character vector of clade names ordered from root to tip, or
#'   `NULL` if the taxon cannot be found.
#'
#' @export
#' @examples
#' \donttest{
#' get_lineage("Tyrannosaurus")
#' get_lineage("Homo sapiens")
#' get_lineage("Quercus robur")
#' }
get_lineage <- function(taxon, clean = TRUE, verbose = FALSE) {
  is_id <- grepl("^[0-9]+$", as.character(taxon))

  if (is_id) {
    id <- as.character(taxon)
  } else {
    id <- get_taxonomicon_id(taxon, verbose = verbose)
  }

  if (is.null(id)) return(NULL)

  lineage <- get_lineage_by_id(id, clean = clean, verbose = verbose)
  if (is.null(lineage)) return(NULL)

  if (!is_id) {
    if (!grepl("\\s", taxon)) {
      lineage <- lineage[!grepl(" ", lineage)]
      lineage <- lineage[!grepl("^\\[", lineage)]
      target_idx <- which(lineage == taxon)
      if (length(target_idx) > 0) {
        lineage <- lineage[seq_len(target_idx[length(target_idx)])]
      } else {
        lineage <- c(lineage, taxon)
      }
    } else {
      lineage <- lineage[!grepl(" ", lineage) | lineage == taxon]
      target_idx <- which(lineage == taxon)
      if (length(target_idx) > 0) {
        lineage <- lineage[seq_len(target_idx[1])]
      } else {
        lineage <- c(lineage, taxon) # nocov
      }
    }
  }

  if (length(lineage) == 0) return(NULL) # nocov
  lineage
}

#' Search The Taxonomicon for a taxon name
#'
#' Queries The Taxonomicon database and returns a data frame of all available
#' biological entries matching the search string. This is particularly useful
#' for exploring homonyms, historical ranks, or taxonomic synonyms before
#' computing distances.
#'
#' @param taxon A character string giving the taxon name to search for.
#' @param verbose Logical. If `TRUE`, prints status messages. Default `FALSE`.
#'
#' @return A data frame of class `"data.frame"` with columns:
#' \describe{
#'   \item{`id`}{Character. The numeric Taxonomicon ID.}
#'   \item{`name`}{Character. The full taxon description, including rank and author.}
#' }
#' Returns `NULL` if no matches are found.
#'
#' @seealso [get_lineage()], [taxo_distance()]
#' @export
#' @examples
#' \donttest{
#' taxo_search("Bacteria")
#' taxo_search("Nereis")
#' taxo_search("Tyrannosaurus")
#' }
taxo_search <- function(taxon, verbose = FALSE) {
  if (verbose) cli::cli_alert_info("Searching Taxonomicon for {.val {taxon}}...")

  url <- paste0(
    "http://taxonomicon.taxonomy.nl/TaxonList.aspx",
    "?subject=Entity&by=ScientificName&search=",
    utils::URLencode(taxon)
  )

  res <- tryCatch(
    httr::GET(url, httr::add_headers("User-Agent" = "taxodist R package/0.3"), httr::timeout(30)),
    error = function(e) NULL
  )

  if (is.null(res) || httr::status_code(res) != 200) {
    if (verbose) cli::cli_alert_warning("Could not reach Taxonomicon")
    return(NULL)
  }

  page <- rvest::read_html(httr::content(res, "text", encoding = "UTF-8"))
  rows <- rvest::html_nodes(page, "tr")

  results <- list()
  for (row in rows) {
    text <- rvest::html_text(row, trim = TRUE)
    if (grepl("\\bastronomical\\b|\\bplanet\\b|\\bMinor planet\\b|\\bcomet\\b|\\basteroid\\b", text, ignore.case = TRUE)) next

    links <- rvest::html_nodes(row, "a[href*='TaxonTree']")
    if (length(links) == 0) next

    classes <- rvest::html_attr(links, "class")
    valid_idx <- which(!is.na(classes) & classes == "Valid")
    if (length(valid_idx) == 0) next

    id <- stringr::str_remove(stringr::str_extract(rvest::html_attr(links[[valid_idx[1]]], "href"), "id=([0-9]+)"), "id=")
    if (is.na(id)) next

    text_entry <- trimws(gsub("\\s+", " ", text))
    text_entry <- stringr::str_remove(text_entry, "^N\\|T\\|P\\|R\\|B\\|L\\s*")

    results <- c(results, list(data.frame(id = id, name = text_entry, stringsAsFactors = FALSE)))
  }

  if (length(results) == 0) {
    if (verbose) cli::cli_alert_warning("No matches found.")
    return(NULL)
  }

  df <- do.call(rbind, results)
  df <- df[!duplicated(df$id), ]
  rownames(df) <- NULL

  if (verbose) cli::cli_alert_success("Found {nrow(df)} entries.")
  return(df)
}

#' taxodist: Taxonomic Hierarchy Distances and Lineage Analysis
#'
#' @description
#' taxodist retrieves taxonomic lineages from The Taxonomicon (taxonomy.nl)
#' and computes distances from the depth of the most recent common ancestor.
#' The resulting values describe separation within the source classification;
#' they are not estimates of evolutionary time or phylogenetic branch length.
#'
#' ## Core functions
#'
#' - [get_lineage()] — retrieve the full lineage of any taxon
#' - [taxo_distance()] — compute the ultrametric distance between two taxa
#' - [mrca()] — find the most recent common ancestor
#' - [distance_matrix()] — compute all pairwise distances for a set of taxa
#' - [closest_relative()] — find the closest relative among candidates
#' - [compare_lineages()] — print a side-by-side lineage comparison
#' - [shared_clades()] — list clades shared between two taxa
#' - [is_member()] — test clade membership
#' - [filter_clade()] — filter taxa by clade membership
#' - [check_coverage()] — check Taxonomicon coverage for a list of taxa
#' - [lineage_depth()] — get the lineage depth of a taxon
#' - [clear_cache()] — clear the session lineage cache
#'
#' ## Mathematical background
#'
#' The distance metric is based on the depth of the most recent common
#' ancestor (MRCA):
#'
#' \deqn{d(A,A) = 0}{d(A,A) = 0}
#' \deqn{d(A,B) = 1 / h(A,B)}{d(A,B) = 1 / h(A,B)}
#'
#' Here, `h(A,B)` denotes the depth of the MRCA. The deeper the shared
#' ancestor, the smaller the distance. Zero is reserved
#' for identical hierarchy nodes; distinct ancestor-descendant pairs therefore
#' have positive distance. The measure is an ultrametric on each connected
#' taxonomic hierarchy. Its numerical values depend on the resolution of the
#' classification returned by the data source.
#'
#' ## Data source
#'
#' All lineage data is sourced from The Taxonomicon (taxonomy.nl), based on
#' Systema Naturae 2000 by S.J. Brands (1989 onwards). Please cite this
#' resource when using taxodist in published work.
#'
#' @references
#' Brands, S.J. (1989 onwards). Systema Naturae 2000. Amsterdam,
#' The Netherlands. Retrieved from The Taxonomicon,
#' \url{http://taxonomicon.taxonomy.nl}.
#'
#' @docType package
#' @name taxodist-package
#' @aliases taxodist
"_PACKAGE"

#' Compute the taxonomic hierarchy distance between two taxa
#'
#' Given two taxon names, retrieves their lineages from The Taxonomicon and
#' computes a taxonomic distance based on the depth of their most recent
#' common ancestor (MRCA):
#'
#' \deqn{d(A,A) = 0}{d(A,A) = 0}
#' \deqn{d(A,B) = 1 / h(A,B)}{d(A,B) = 1 / h(A,B)}
#'
#' Here, `h(A,B)` is the depth of the MRCA of `A` and `B`. The deeper the
#' shared ancestor, the smaller (closer to zero) the distance.
#' Distinct nodes always have positive distance, including when one taxon is
#' an ancestor of the other. This makes the measure an ultrametric on each
#' connected taxonomic hierarchy. Use [is_member()] when the question is
#' whether one taxon belongs to a clade rather than how far apart the two
#' hierarchy nodes are.
#'
#' @param taxon_a A character string giving the first taxon name.
#' @param taxon_b A character string giving the second taxon name.
#' @param verbose Logical. If `TRUE`, prints progress messages. Default `FALSE`.
#'
#' @return A named list of class `"taxodist_result"` with the following elements:
#' \describe{
#'   \item{`distance`}{Numeric. The distance between the two taxa. Returns 0
#'     only when their complete lineages are identical.}
#'   \item{`mrca`}{Character. The name of the most recent common ancestor.}
#'   \item{`mrca_depth`}{Integer. The depth of the MRCA node.}
#'   \item{`depth_a`}{Integer. The lineage depth of taxon A.}
#'   \item{`depth_b`}{Integer. The lineage depth of taxon B.}
#'   \item{`taxon_a`}{Character. Name of the first taxon.}
#'   \item{`taxon_b`}{Character. Name of the second taxon.}
#' }
#' Returns `NULL` if either taxon cannot be found.
#'
#' @seealso [mrca()], [distance_matrix()], [get_lineage()], [is_member()],
#'   [taxo_path()]
#'
#' @references
#' Brands, S.J. (1989 onwards). Systema Naturae 2000. Amsterdam, The
#' Netherlands. Retrieved from The Taxonomicon,
#' \url{http://taxonomicon.taxonomy.nl}.
#'
#' @export
#' @examples
#' \donttest{
#' # Distance between two theropods
#' taxo_distance("Tyrannosaurus", "Velociraptor")
#'
#' # Distance between very distantly related taxa
#' taxo_distance("Tyrannosaurus", "Quercus")
#'
#' # Distance between two oviraptorid genera
#' taxo_distance("Nomingia", "Huanansaurus")
#' }
taxo_distance <- function(taxon_a, taxon_b, verbose = FALSE) {
  lin_a <- get_lineage(taxon_a, verbose = verbose)
  lin_b <- get_lineage(taxon_b, verbose = verbose)

  if (is.null(lin_a)) {
    cli::cli_alert_danger("Could not retrieve lineage for {taxon_a}")
    return(NULL)
  }
  if (is.null(lin_b)) {
    cli::cli_alert_danger("Could not retrieve lineage for {taxon_b}")
    return(NULL)
  }

  .compute_distance(lin_a, lin_b, taxon_a, taxon_b)
}

#' Compute the most recent common ancestor of two taxa
#'
#' Retrieves the lineages of two taxa and returns the name of their most
#' recent common ancestor (MRCA) — the deepest node shared by both lineages.
#'
#' @param taxon_a A character string giving the first taxon name.
#' @param taxon_b A character string giving the second taxon name.
#' @param verbose Logical. If `TRUE`, prints progress messages. Default `FALSE`.
#'
#' @return A character string giving the name of the MRCA, or `NULL` if
#'   either taxon cannot be found or no common ancestor exists.
#'
#' @export
#' @examples
#' \donttest{
#' mrca("Tyrannosaurus", "Velociraptor")   # "Tyrannoraptora"
#' mrca("Tyrannosaurus", "Triceratops")    # "Dinosauria"
#' mrca("Tyrannosaurus", "Homo")           # "Amniota"
#' }
mrca <- function(taxon_a, taxon_b, verbose = FALSE) {
  result <- taxo_distance(taxon_a, taxon_b, verbose = verbose)
  if (is.null(result)) return(NULL)
  result$mrca
}

#' Compute pairwise taxonomic distances for a set of taxa
#'
#' Given a vector of taxon names, computes all pairwise taxonomic hierarchy distances
#' and returns a symmetric distance matrix. Lineages are cached after first
#' retrieval to minimise redundant network requests.
#'
#' @param taxa A character vector of taxon names.
#' @param verbose Logical. If `TRUE`, prints progress for each pair.
#'   Default `FALSE`.
#' @param progress Logical. If `TRUE`, shows a progress bar. Default `TRUE`.
#'
#' @return A symmetric numeric matrix of class `"dist"` containing pairwise
#'   distances. Row and column names are set to the input taxon names.
#'   Taxa that could not be found are included with `NA` distances.
#'   An empty input returns an empty `"dist"` object.
#'
#' @seealso [taxo_distance()], [closest_relative()]
#'
#' @export
#' @examples
#' \donttest{
#' theropods <- c("Tyrannosaurus", "Velociraptor", "Spinosaurus",
#'                "Allosaurus", "Carnotaurus")
#' mat <- distance_matrix(theropods)
#' print(mat)
#' }
distance_matrix <- function(taxa, verbose = FALSE, progress = TRUE) {
  n <- length(taxa)
  mat <- matrix(NA_real_, nrow = n, ncol = n,
                dimnames = list(taxa, taxa))
  diag(mat) <- 0

  if (n == 0L) {
    return(stats::as.dist(mat))
  }

  # fetch lineages sequentially in main process (cache is shared)
  if (progress) cli::cli_alert_info("Fetching {n} lineages...")
  lineages <- lapply(taxa, function(t) get_lineage(t, verbose = verbose))
  names(lineages) <- taxa
  if (progress) cli::cli_alert_success("Lineages fetched.")

  # compute pairwise distances
  total_pairs <- n * (n - 1) / 2
  if (progress) cli::cli_progress_bar("Computing distances", total = total_pairs)
  for (i in seq_len(n - 1)) {
    for (j in (i + 1):n) {
      if (!is.null(lineages[[i]]) && !is.null(lineages[[j]])) {
        result <- .compute_distance(
          lineages[[i]], lineages[[j]], taxa[i], taxa[j]
        )
        mat[i, j] <- mat[j, i] <- result$distance
      }
      if (progress) cli::cli_progress_update()
    }
  }
  if (progress) cli::cli_progress_done()
  stats::as.dist(mat)
}

#' Find the closest relative of a taxon among a set of candidates
#'
#' Given a query taxon and a vector of candidate taxa, returns the candidate
#' with the smallest taxonomic hierarchy distance to the query.
#'
#' @param taxon A character string giving the query taxon name.
#' @param candidates A character vector of candidate taxon names to compare
#'   against.
#' @param verbose Logical. If `TRUE`, prints progress messages. Default `FALSE`.
#'
#' @return A data frame with columns `taxon` (candidate name) and `distance`
#'   (ultrametric distance), sorted by distance ascending. Returns `NULL` if
#'   the query taxon cannot be found. An empty candidate vector returns an
#'   empty data frame with the same columns.
#'
#' @export
#' @examples
#' \donttest{
#' closest_relative("Tyrannosaurus",
#'   c("Velociraptor", "Triceratops", "Brachiosaurus", "Allosaurus"))
#' }
closest_relative <- function(taxon, candidates, verbose = FALSE) {
  query_lin <- get_lineage(taxon, verbose = verbose)
  if (is.null(query_lin)) {
    cli::cli_alert_danger("Could not retrieve lineage for {taxon}")
    return(NULL)
  }

  if (length(candidates) == 0L) {
    return(data.frame(
      taxon = character(0),
      distance = numeric(0),
      stringsAsFactors = FALSE
    ))
  }

  results <- do.call(rbind, purrr::map(
    candidates,
    function(cand) {
      cand_lin <- get_lineage(cand, verbose = verbose)
      if (is.null(cand_lin)) {
        return(data.frame(taxon = cand, distance = NA_real_))
      }
      dist_result <- .compute_distance(query_lin, cand_lin, taxon, cand)
      data.frame(taxon = cand, distance = dist_result$distance)
    }
  ))

  results[order(results$distance, na.last = TRUE), ]
}

#' Compute distances from a focal taxon to a community of taxa
#'
#' Given a focal taxon and a vector of community taxa, retrieves all lineages
#' and computes the taxonomic distance from the focal taxon to each member of
#' the community. Returns a sorted data frame that also reports the most recent
#' common ancestor (MRCA) for each pair, making it easy to interpret *why* a
#' taxon is close or distant.
#'
#' @param focal A character string giving the focal taxon name.
#' @param community A character vector of community taxon names to compare
#'   against. The focal taxon may be included — it will receive a distance
#'   of `0`.
#' @param verbose Logical. If `TRUE`, prints progress messages. Default `FALSE`.
#' @param progress Logical. If `TRUE`, shows a progress bar. Default `TRUE`.
#'
#' @return A data frame of class `"taxodist_focal"` with columns:
#' \describe{
#'   \item{`taxon`}{Character. Community taxon name.}
#'   \item{`distance`}{Numeric. Taxonomic distance to the focal taxon.}
#'   \item{`mrca`}{Character. Name of the most recent common ancestor.}
#'   \item{`mrca_depth`}{Integer. Depth of the MRCA node.}
#' }
#' Rows are sorted by `distance` ascending (closest relatives first).
#' Returns `NULL` if the focal taxon cannot be found. An empty community
#' returns an empty object with the same columns and class.
#'
#' @seealso [closest_relative()], [distance_matrix()], [taxo_distance()]
#' @export
#' @examples
#' \donttest{
#' community <- c("Velociraptor", "Triceratops", "Brachiosaurus",
#'                "Allosaurus", "Homo", "Quercus")
#' focal_distances("Tyrannosaurus", community)
#' }
focal_distances <- function(focal, community, verbose = FALSE, progress = TRUE) {
  focal_lin <- get_lineage(focal, verbose = verbose)
  if (is.null(focal_lin)) {
    cli::cli_alert_danger("Could not retrieve lineage for {focal}")
    return(NULL)
  }

  if (length(community) == 0L) {
    out <- data.frame(
      taxon = character(0),
      distance = numeric(0),
      mrca = character(0),
      mrca_depth = integer(0),
      stringsAsFactors = FALSE
    )
    return(structure(
      out,
      class = c("taxodist_focal", "data.frame"),
      focal = focal
    ))
  }

  if (progress) cli::cli_progress_bar("Computing focal distances", total = length(community))

  results <- vector("list", length(community))
  for (i in seq_along(community)) {
    taxon <- community[[i]]

    if (identical(taxon, focal)) {
      results[[i]] <- data.frame(
        taxon      = taxon,
        distance   = 0,
        mrca       = focal,
        mrca_depth = length(focal_lin),
        stringsAsFactors = FALSE
      )
    } else {
      cand_lin <- get_lineage(taxon, verbose = verbose)
      if (is.null(cand_lin)) {
        results[[i]] <- data.frame(
          taxon      = taxon,
          distance   = NA_real_,
          mrca       = NA_character_,
          mrca_depth = NA_integer_,
          stringsAsFactors = FALSE
        )
      } else {
        res <- .compute_distance(focal_lin, cand_lin, focal, taxon)
        results[[i]] <- data.frame(
          taxon      = taxon,
          distance   = res$distance,
          mrca       = res$mrca,
          mrca_depth = res$mrca_depth,
          stringsAsFactors = FALSE
        )
      }
    }

    if (progress) cli::cli_progress_update()
  }

  if (progress) cli::cli_progress_done()

  out <- do.call(rbind, results)
  out <- out[order(out$distance, na.last = TRUE), ]
  rownames(out) <- NULL

  structure(out, class = c("taxodist_focal", "data.frame"), focal = focal)
}

#' Get the lineage depth of a taxon
#'
#' Returns the number of nodes in the lineage of a taxon, from root to tip.
#' This reflects how deeply nested the taxon is within the taxonomic hierarchy.
#'
#' @param taxon A character string giving the taxon name.
#' @param verbose Logical. If `TRUE`, prints progress messages. Default `FALSE`.
#'
#' @return An integer giving the lineage depth, or `NULL` if the taxon cannot
#'   be found.
#'
#' @export
#' @examples
#' \donttest{
#' lineage_depth("Tyrannosaurus")   # deep — many intermediate clades
#' lineage_depth("Biota")           # shallow — near root
#' }
lineage_depth <- function(taxon, verbose = FALSE) {
  lin <- get_lineage(taxon, verbose = verbose)
  if (is.null(lin)) return(NULL)
  length(lin)
}

#' Check whether a taxon is covered by The Taxonomicon
#'
#' Queries The Taxonomicon for a taxon name and returns a logical indicating
#' whether the taxon was found. Useful for pre-screening a list of names
#' before running distance computations.
#'
#' @param taxa A character vector of one or more taxon names.
#' @param verbose Logical. If `TRUE`, prints progress messages. Default `FALSE`.
#'
#' @return A named logical vector. `TRUE` indicates the taxon was found,
#'   `FALSE` indicates it was not.
#'
#' @export
#' @examples
#' \donttest{
#' check_coverage(c("Tyrannosaurus", "Velociraptor", "Fakeosaurus"))
#' }
check_coverage <- function(taxa, verbose = FALSE) {
  result <- purrr::map_lgl(taxa, function(t) {
    !is.null(get_taxonomicon_id(t, verbose = verbose))
  })
  names(result) <- taxa
  result
}

#' Cluster taxa by taxonomic distance
#'
#' Computes pairwise taxonomic distances and performs hierarchical clustering.
#'
#' @param taxa A character vector of taxon names, or a `dist` object from
#'[distance_matrix()].
#' @param method Clustering method passed to [stats::hclust()]. Default
#'   `"average"` (UPGMA), which works well with taxonomic distances.
#' @param ... Additional arguments passed to [distance_matrix()] (e.g.
#'   `verbose`, `progress`).
#'
#' @return An object of class `"taxodist_cluster"` — a list with:
#' \describe{
#'   \item{`hclust`}{The [stats::hclust()] result.}
#'   \item{`dist`}{The underlying distance matrix.}
#' }
#' If fewer than two taxa are supplied, `hclust` is `NULL` and the distance
#' object is preserved.
#'
#' @seealso[taxo_ordinate()], [distance_matrix()]
#' @export
#' @examples
#' \donttest{
#' taxa <- c("Tyrannosaurus", "Velociraptor", "Triceratops", "Brachiosaurus")
#' cl <- taxo_cluster(taxa)
#' if (!is.null(cl$hclust)) {
#'   plot(cl)
#' }
#' }
taxo_cluster <- function(taxa, method = "average", ...) {
  d <- if (inherits(taxa, "dist")) taxa else distance_matrix(taxa, ...)
  if (any(is.na(d))) {
    cli::cli_warn("Distance matrix contains NA values (taxa not found or server offline). Clustering skipped.")
    return(structure(list(hclust = NULL, dist = d), class = "taxodist_cluster"))
  }
  if (any(!is.finite(d))) {
    cli::cli_warn("Distance matrix contains infinite values (no shared ancestor). Clustering skipped.")
    return(structure(list(hclust = NULL, dist = d), class = "taxodist_cluster"))
  }
  if (attr(d, "Size") < 2L) {
    cli::cli_warn("At least two taxa are required for clustering. Clustering skipped.")
    return(structure(list(hclust = NULL, dist = d), class = "taxodist_cluster"))
  }
  hc <- stats::hclust(d, method = method)
  structure(list(hclust = hc, dist = d), class = "taxodist_cluster")
}

#' Ordinate taxa in taxonomic distance space
#'
#' Computes pairwise taxonomic distances and applies classical multidimensional
#' scaling (PCoA) to project taxa into a low-dimensional space.
#'
#' @param taxa A character vector of taxon names, or a `dist` object from
#'[distance_matrix()].
#' @param k Number of dimensions. Default `2`.
#' @param ... Additional arguments passed to [distance_matrix()].
#'
#' @return An object of class `"taxodist_ord"` — a list with:
#' \describe{
#'   \item{`points`}{A matrix of coordinates (taxa x k dimensions).}
#'   \item{`dist`}{The underlying distance matrix.}
#'   \item{`GOF`}{Goodness-of-fit from[stats::cmdscale()].}
#'   \item{`eig`}{The eigenvalues computed during PCoA.}
#' }
#' At least two taxa are required. When `k` is greater than the maximum
#' possible dimension (`n - 1`), it is reduced automatically.
#'
#' @seealso [taxo_cluster()], [distance_matrix()]
#' @export
#' @examples
#' \donttest{
#' taxa <- c("Tyrannosaurus", "Velociraptor", "Triceratops", "Brachiosaurus")
#' ord <- taxo_ordinate(taxa)
#' if (!is.null(ord$points)) {
#'   plot(ord$points, type = "n")
#'   text(ord$points, labels = rownames(ord$points))
#' }
#' }
taxo_ordinate <- function(taxa, k = 2, ...) {
  d <- if (inherits(taxa, "dist")) taxa else distance_matrix(taxa, ...)
  if (any(is.na(d))) {
    cli::cli_warn("Distance matrix contains NA values. Ordination skipped.")
    return(structure(list(points = NULL, dist = d, GOF = NULL, eig = NULL), class = "taxodist_ord"))
  }
  if (any(!is.finite(d))) {
    cli::cli_warn("Distance matrix contains infinite values (no shared ancestor). Ordination skipped.")
    return(structure(list(points = NULL, dist = d, GOF = NULL, eig = NULL), class = "taxodist_ord"))
  }
  n_taxa <- attr(d, "Size")
  if (n_taxa < 2L) {
    cli::cli_warn("At least two taxa are required for ordination. Ordination skipped.")
    return(structure(list(points = NULL, dist = d, GOF = NULL, eig = NULL), class = "taxodist_ord"))
  }
  valid_k <- length(k) == 1L &&
    (is.integer(k) || is.double(k)) &&
    is.finite(k) && k >= 1 && k == floor(k)
  if (!valid_k) {
    stop("`k` must be a single positive integer.", call. = FALSE)
  }
  max_k <- n_taxa - 1L
  if (k > max_k) {
    cli::cli_warn("`k` reduced from {k} to {max_k}, the maximum for {n_taxa} taxa.")
    k <- max_k
  }
  cmd <- stats::cmdscale(d, k = k, eig = TRUE)
  structure(list(
    points = cmd$points,
    dist   = d,
    GOF    = cmd$GOF,
    eig    = cmd$eig
  ), class = "taxodist_ord")
}

# ── Internal helpers ──────────────────────────────────────────────────────────

#' @keywords internal
.compute_distance <- function(lin_a, lin_b, name_a = "A", name_b = "B") {
  depth_a <- length(lin_a)
  depth_b <- length(lin_b)

  # The shared ancestry must be a continuous prefix from the root. A taxon
  # name repeated after the lineages diverge is a homonym, not a shared node.
  common_length <- min(depth_a, depth_b)
  common_idx    <- seq_len(common_length)
  matches       <- lin_a[common_idx] == lin_b[common_idx]
  first_mismatch <- which(is.na(matches) | !matches)

  mrca_depth <- if (length(first_mismatch) > 0L) {
    first_mismatch[1] - 1L
  } else {
    common_length
  }

  if (mrca_depth == 0L) {
    return(structure(list(
      distance   = Inf,
      mrca       = NA_character_,
      mrca_depth = 0L,
      depth_a    = depth_a,
      depth_b    = depth_b,
      taxon_a    = name_a,
      taxon_b    = name_b
    ), class = "taxodist_result"))
  }

  mrca_name    <- lin_a[mrca_depth]
  same_lineage <- identical(as.character(lin_a), as.character(lin_b))
  distance     <- if (same_lineage) 0 else 1 / mrca_depth

  structure(list(
    distance   = distance,
    mrca       = mrca_name,
    mrca_depth = mrca_depth,
    depth_a    = depth_a,
    depth_b    = depth_b,
    taxon_a    = name_a,
    taxon_b    = name_b
  ), class = "taxodist_result")
}

library(testthat)
library(taxodist)

# ── Pure logic tests ──────────────────────────────────────────────────────────

test_that("taxodist package loads", {
  expect_true(TRUE)
})

test_that(".compute_distance works correctly", {
  lin_a <- c("Biota", "Animalia", "Chordata", "Dinosauria",
             "Theropoda", "Tyrannosauridae", "Tyrannosaurus")
  lin_b <- c("Biota", "Animalia", "Chordata", "Dinosauria",
             "Theropoda", "Dromaeosauridae", "Velociraptor")
  result <- taxodist:::.compute_distance(lin_a, lin_b, "Tyrannosaurus",
                                         "Velociraptor")
  expect_equal(result$mrca, "Theropoda")
  expect_equal(result$mrca_depth, 5L)
  expect_equal(result$depth_a, 7L)
  expect_equal(result$depth_b, 7L)
  expect_gte(result$distance, 0)
  expect_lte(result$distance, 1)
})

test_that(".compute_distance distance is between 0 and 1", {
  lin_a <- c("Biota", "Animalia", "Chordata", "Dinosauria",
             "Theropoda", "Tyrannosauridae", "Tyrannosaurus")
  lin_b <- c("Biota", "Animalia", "Chordata", "Dinosauria",
             "Theropoda", "Dromaeosauridae", "Velociraptor")
  result <- taxodist:::.compute_distance(lin_a, lin_b)
  expect_gte(result$distance, 0)
  expect_lte(result$distance, 1)
})

test_that(".compute_distance is symmetric", {
  lin_a <- c("Biota", "Animalia", "Chordata", "Dinosauria", "Theropoda")
  lin_b <- c("Biota", "Animalia", "Chordata", "Dinosauria", "Ornithischia")
  r1 <- taxodist:::.compute_distance(lin_a, lin_b)
  r2 <- taxodist:::.compute_distance(lin_b, lin_a)
  expect_equal(r1$distance, r2$distance)
})

test_that(".compute_distance satisfies triangle inequality", {
  lin_a <- c("Biota", "Animalia", "Dinosauria", "Theropoda", "Tyrannosauridae")
  lin_b <- c("Biota", "Animalia", "Dinosauria", "Theropoda", "Dromaeosauridae")
  lin_c <- c("Biota", "Animalia", "Dinosauria", "Ornithischia")
  dAB <- taxodist:::.compute_distance(lin_a, lin_b)$distance
  dBC <- taxodist:::.compute_distance(lin_b, lin_c)$distance
  dAC <- taxodist:::.compute_distance(lin_a, lin_c)$distance
  expect_lte(dAC, dAB + dBC)
})

test_that(".compute_distance returns 0 for identical lineages", {
  lin <- c("Biota", "Animalia", "Dinosauria", "Tyrannosaurus")
  result <- taxodist:::.compute_distance(lin, lin)
  expect_equal(result$distance, 0)
  expect_equal(result$mrca, "Tyrannosaurus")
})

test_that(".compute_distance handles no common ancestor", {
  lin_a <- c("Biota", "Animalia")
  lin_b <- c("Fungi", "Ascomycota")
  result <- taxodist:::.compute_distance(lin_a, lin_b)
  expect_equal(result$mrca_depth, 0L)
  expect_true(is.na(result$mrca))
})

test_that(".compute_distance returns Inf for no shared ancestor", {
  lin_a <- c("Biota", "Animalia")
  lin_b <- c("Fungi", "Ascomycota")
  result <- taxodist:::.compute_distance(lin_a, lin_b)
  expect_equal(result$distance, Inf)
})

test_that(".compute_distance result has correct S3 class", {
  lin <- c("Biota", "Animalia", "Dinosauria", "Tyrannosaurus")
  result <- taxodist:::.compute_distance(lin, lin)
  expect_s3_class(result, "taxodist_result")
})

test_that(".compute_distance distance is between 0 and 1 for asymmetric lineages", {
  lin_a <- c("Biota", "Animalia", "Dinosauria", "Theropoda",
             "Abelisauridae", "Carnotaurus")
  lin_b <- c("Biota", "Animalia", "Dinosauria", "Ornithischia", "Triceratops")
  result <- taxodist:::.compute_distance(lin_a, lin_b)
  expect_gte(result$distance, 0)
  expect_lte(result$distance, 1)
})

test_that(".compute_distance returns 0 when one taxon is ancestor of other", {
  lin_a <- c("Biota", "Animalia", "Dinosauria")
  lin_b <- c("Biota", "Animalia", "Dinosauria", "Theropoda", "Carnotaurus")
  result <- taxodist:::.compute_distance(lin_a, lin_b)
  expect_equal(result$distance, 0)
})

test_that("clear_cache returns invisible NULL", {
  expect_invisible(clear_cache())
})

test_that("save_cache creates a file with cache contents", {
  clear_cache()
  assign("id_Carnotaurus", "12345", envir = taxodist:::.taxodist_cache)
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  expect_invisible(save_cache(tmp))
  expect_true(file.exists(tmp))
})

test_that("load_cache restores entries into the cache", {
  clear_cache()
  assign("id_Carnotaurus", "12345", envir = taxodist:::.taxodist_cache)
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  save_cache(tmp)
  clear_cache()
  expect_false(exists("id_Carnotaurus", envir = taxodist:::.taxodist_cache))
  load_cache(tmp)
  expect_true(exists("id_Carnotaurus", envir = taxodist:::.taxodist_cache))
  expect_equal(get("id_Carnotaurus", envir = taxodist:::.taxodist_cache), "12345")
})

test_that("save_cache / load_cache round-trip preserves all entries", {
  clear_cache()
  assign("id_Tyrannosaurus", "50841", envir = taxodist:::.taxodist_cache)
  assign("lin_50841", c("Biota", "Animalia", "Dinosauria", "Tyrannosaurus"),
         envir = taxodist:::.taxodist_cache)
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  save_cache(tmp)
  clear_cache()
  load_cache(tmp)
  expect_equal(get("id_Tyrannosaurus",  envir = taxodist:::.taxodist_cache), "50841")
  expect_equal(get("lin_50841", envir = taxodist:::.taxodist_cache),
               c("Biota", "Animalia", "Dinosauria", "Tyrannosaurus"))
})

test_that("load_cache errors on missing file", {
  expect_error(load_cache("nonexistent_file.rds"))
})

test_that("save_cache returns invisible NULL", {
  clear_cache()
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  expect_invisible(save_cache(tmp))
})

test_that("load_cache returns invisible NULL", {
  clear_cache()
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  save_cache(tmp)
  clear_cache()
  expect_invisible(load_cache(tmp))
})

test_that("filter_clade filters correctly with mocked lineages", {
  mockery::stub(filter_clade, "is_member", function(taxon, clade, ...) {
    memberships <- list(
      Tyrannosaurus = c("Dinosauria", "Theropoda"),
      Triceratops   = c("Dinosauria", "Ornithischia"),
      Homo          = c("Mammalia", "Amniota")
    )
    clade %in% memberships[[taxon]]
  })
  result <- filter_clade(
    c("Tyrannosaurus", "Triceratops", "Homo"), "Dinosauria"
  )
  expect_equal(result, c("Tyrannosaurus", "Triceratops"))
})

# ── taxo_path ─────────────────────────────────────────────────────────────────

test_that("taxo_path returns NULL when taxon_a not found", {
  mockery::stub(taxo_path, "get_lineage", function(...) NULL)
  expect_null(taxo_path("Fakeosaurus", "Carnotaurus"))
})

test_that("taxo_path returns NULL when taxon_b not found", {
  mockery::stub(taxo_path, "get_lineage", function(taxon, ...) {
    if (taxon == "Carnotaurus")
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Carnotaurus")
    else
      NULL
  })
  expect_null(taxo_path("Carnotaurus", "Fakeosaurus"))
})

test_that("taxo_path returns a taxodist_path data frame", {
  mockery::stub(taxo_path, "get_lineage", function(taxon, ...) {
    if (taxon == "Tyrannosaurus")
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Tyrannosaurus")
    else
      c("Biota", "Animalia", "Dinosauria", "Ornithischia", "Triceratops")
  })
  result <- taxo_path("Tyrannosaurus", "Triceratops")
  expect_s3_class(result, "taxodist_path")
  expect_s3_class(result, "data.frame")
  expect_named(result, c("node", "depth", "direction"))
})

test_that("taxo_path has exactly one mrca row", {
  mockery::stub(taxo_path, "get_lineage", function(taxon, ...) {
    if (taxon == "Tyrannosaurus")
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Tyrannosaurus")
    else
      c("Biota", "Animalia", "Dinosauria", "Ornithischia", "Triceratops")
  })
  result <- taxo_path("Tyrannosaurus", "Triceratops")
  expect_equal(sum(result$direction == "mrca"), 1L)
})

test_that("taxo_path mrca node is correct", {
  mockery::stub(taxo_path, "get_lineage", function(taxon, ...) {
    if (taxon == "Tyrannosaurus")
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Tyrannosaurus")
    else
      c("Biota", "Animalia", "Dinosauria", "Ornithischia", "Triceratops")
  })
  result <- taxo_path("Tyrannosaurus", "Triceratops")
  expect_equal(result$node[result$direction == "mrca"], "Dinosauria")
})

test_that("taxo_path direction column only contains valid values", {
  mockery::stub(taxo_path, "get_lineage", function(taxon, ...) {
    if (taxon == "Tyrannosaurus")
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Tyrannosaurus")
    else
      c("Biota", "Animalia", "Dinosauria", "Ornithischia", "Triceratops")
  })
  result <- taxo_path("Tyrannosaurus", "Triceratops")
  expect_true(all(result$direction %in% c("a", "mrca", "b")))
})

test_that("taxo_path preserves taxon attributes", {
  mockery::stub(taxo_path, "get_lineage", function(taxon, ...) {
    if (taxon == "Tyrannosaurus")
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Tyrannosaurus")
    else
      c("Biota", "Animalia", "Dinosauria", "Ornithischia", "Triceratops")
  })
  result <- taxo_path("Tyrannosaurus", "Triceratops")
  expect_equal(attr(result, "taxon_a"), "Tyrannosaurus")
  expect_equal(attr(result, "taxon_b"), "Triceratops")
})

test_that("print.taxodist_path runs without error and returns invisibly", {
  mockery::stub(taxo_path, "get_lineage", function(taxon, ...) {
    if (taxon == "Tyrannosaurus")
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Tyrannosaurus")
    else
      c("Biota", "Animalia", "Dinosauria", "Ornithischia", "Triceratops")
  })
  result <- taxo_path("Tyrannosaurus", "Triceratops")
  expect_no_error(print(result))
  expect_invisible(print(result))
})

# ── Mock tests ────────────────────────────────────────────────────────────────

test_that("get_taxonomicon_id returns NULL on network failure", {
  clear_cache()
  mockery::stub(get_taxonomicon_id, "httr::GET", function(...) NULL)
  expect_warning(result <- get_taxonomicon_id("Tyrannosaurus"), "Cannot reach")
  expect_null(result)
})

test_that("get_taxonomicon_id returns NULL on bad status", {
  clear_cache()
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_taxonomicon_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_taxonomicon_id, "httr::status_code", function(...) 404L)
  expect_warning(result <- get_taxonomicon_id("Tyrannosaurus"), "Cannot reach")
  expect_null(result)
})

test_that("get_lineage_by_id returns NULL on network failure", {
  clear_cache()
  mockery::stub(get_lineage_by_id, "httr::GET", function(...) NULL)
  result <- get_lineage_by_id("12345")
  expect_null(result)
})

test_that("get_lineage returns NULL when id not found", {
  clear_cache()
  mockery::stub(get_lineage, "get_taxonomicon_id", function(...) NULL)
  result <- get_lineage("Fakeosaurus")
  expect_null(result)
})

test_that("cache is used on second call to get_taxonomicon_id", {
  clear_cache()
  assign("id_Tyrannosaurus", "50841", envir = taxodist:::.taxodist_cache)
  call_count <- 0L
  mockery::stub(get_taxonomicon_id, "httr::GET", function(...) {
    call_count <<- call_count + 1L
    NULL
  })
  result <- get_taxonomicon_id("Tyrannosaurus")
  expect_equal(call_count, 0L)
  expect_equal(result, "50841")
})

test_that("taxo_distance works with mocked lineages", {
  mockery::stub(taxo_distance, "get_lineage", function(taxon, ...) {
    if (taxon == "Tyrannosaurus")
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Tyrannosaurus")
    else
      c("Biota", "Animalia", "Dinosauria", "Ornithischia", "Triceratops")
  })
  result <- taxo_distance("Tyrannosaurus", "Triceratops")
  expect_s3_class(result, "taxodist_result")
  expect_equal(result$mrca, "Dinosauria")
})

test_that("closest_relative works with mocked lineages", {
  mockery::stub(closest_relative, "get_lineage", function(taxon, ...) {
    lins <- list(
      Tyrannosaurus = c("Biota", "Animalia", "Dinosauria", "Theropoda", "Tyrannosaurus"),
      Velociraptor  = c("Biota", "Animalia", "Dinosauria", "Theropoda", "Velociraptor"),
      Triceratops   = c("Biota", "Animalia", "Dinosauria", "Ornithischia", "Triceratops")
    )
    lins[[taxon]]
  })
  result <- closest_relative("Tyrannosaurus", c("Velociraptor", "Triceratops"))
  expect_equal(nrow(result), 2)
  expect_equal(result$taxon[1], "Velociraptor")
})

test_that("lineage_depth works with mocked lineage", {
  mockery::stub(lineage_depth, "get_lineage",
                function(...) c("Biota", "Animalia", "Dinosauria", "Theropoda", "Tyrannosaurus"))
  expect_equal(lineage_depth("Tyrannosaurus"), 5L)
})

test_that("shared_clades works with mocked lineages", {
  mockery::stub(shared_clades, "get_lineage", function(taxon, ...) {
    if (taxon == "Tyrannosaurus")
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Tyrannosaurus")
    else
      c("Biota", "Animalia", "Dinosauria", "Ornithischia", "Triceratops")
  })
  result <- shared_clades("Tyrannosaurus", "Triceratops")
  expect_equal(result, c("Biota", "Animalia", "Dinosauria"))
})

test_that("is_member works with mocked lineage", {
  mockery::stub(is_member, "get_lineage",
                function(...) c("Biota", "Animalia", "Dinosauria", "Theropoda"))
  expect_true(is_member("Tyrannosaurus", "Dinosauria"))
  expect_false(is_member("Tyrannosaurus", "Mammalia"))
})

test_that("compare_lineages works with mocked lineages", {
  mockery::stub(compare_lineages, "get_lineage", function(taxon, ...) {
    if (taxon == "Tyrannosaurus")
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Tyrannosaurus")
    else
      c("Biota", "Animalia", "Dinosauria", "Ornithischia", "Triceratops")
  })
  result <- compare_lineages("Tyrannosaurus", "Triceratops")
  expect_equal(result$mrca_depth, 3L)
})

test_that("distance_matrix works with mocked lineages", {
  mockery::stub(distance_matrix, "get_lineage", function(taxon, ...) {
    lins <- list(
      Tyrannosaurus = c("Biota", "Animalia", "Dinosauria", "Theropoda", "Tyrannosaurus"),
      Velociraptor  = c("Biota", "Animalia", "Dinosauria", "Theropoda", "Velociraptor"),
      Triceratops   = c("Biota", "Animalia", "Dinosauria", "Ornithischia", "Triceratops")
    )
    lins[[taxon]]
  })
  mat <- distance_matrix(c("Tyrannosaurus", "Velociraptor", "Triceratops"),
                         progress = FALSE)
  expect_equal(nrow(as.matrix(mat)), 3)
  expect_equal(diag(as.matrix(mat)), c(0, 0, 0), ignore_attr = TRUE)
})

test_that("check_coverage returns named logical vector", {
  mockery::stub(check_coverage, "get_taxonomicon_id",
                function(taxon, ...) if (taxon == "Fakeosaurus") NULL else "12345")
  result <- check_coverage(c("Tyrannosaurus", "Fakeosaurus"))
  expect_type(result, "logical")
  expect_true(result["Tyrannosaurus"])
  expect_false(result["Fakeosaurus"])
})

test_that("taxo_distance returns NULL when taxon_a not found", {
  mockery::stub(taxo_distance, "get_lineage", function(taxon, ...) NULL)
  result <- taxo_distance("Fakeosaurus", "Carnotaurus")
  expect_null(result)
})

test_that("taxo_distance returns NULL when taxon_b not found", {
  mockery::stub(taxo_distance, "get_lineage", function(taxon, ...) {
    if (taxon == "Carnotaurus")
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Abelisauridae", "Carnotaurus")
    else
      NULL
  })
  result <- taxo_distance("Carnotaurus", "Fakeosaurus")
  expect_null(result)
})

test_that("mrca returns NULL when taxo_distance fails", {
  mockery::stub(mrca, "taxo_distance", function(...) NULL)
  result <- mrca("Fakeosaurus", "Carnotaurus")
  expect_null(result)
})

test_that("mrca returns correct value when taxo_distance succeeds", {
  mockery::stub(mrca, "taxo_distance", function(...) list(mrca = "Dinosauria"))
  result <- mrca("Carnotaurus", "Triceratops")
  expect_equal(result, "Dinosauria")
})

test_that("closest_relative returns NULL when query lineage not found", {
  mockery::stub(closest_relative, "get_lineage", function(...) NULL)
  result <- closest_relative("Fakeosaurus", c("Carnotaurus", "Velociraptor"))
  expect_null(result)
})

test_that("closest_relative handles NULL candidate lineage", {
  mockery::stub(closest_relative, "get_lineage", function(taxon, ...) {
    if (taxon == "Carnotaurus")
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Abelisauridae", "Carnotaurus")
    else if (taxon == "Velociraptor")
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Dromaeosauridae", "Velociraptor")
    else
      NULL
  })
  result <- closest_relative("Carnotaurus", c("Velociraptor", "Fakeosaurus"))
  expect_equal(nrow(result), 2)
  expect_true(is.na(result$distance[result$taxon == "Fakeosaurus"]))
})

test_that("distance_matrix handles NULL lineage for one taxon", {
  mockery::stub(distance_matrix, "get_lineage", function(taxon, ...) {
    if (taxon == "Fakeosaurus") NULL
    else if (taxon == "Carnotaurus")
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Abelisauridae", "Carnotaurus")
    else
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Dromaeosauridae", "Velociraptor")
  })
  mat <- distance_matrix(c("Carnotaurus", "Velociraptor", "Fakeosaurus"),
                         progress = FALSE)
  m <- as.matrix(mat)
  expect_true(is.na(m["Carnotaurus", "Fakeosaurus"]))
  expect_false(is.na(m["Carnotaurus", "Velociraptor"]))
})

test_that("distance_matrix with progress = TRUE runs without error", {
  mockery::stub(distance_matrix, "get_lineage", function(taxon, ...) {
    lins <- list(
      Carnotaurus  = c("Biota", "Animalia", "Dinosauria", "Theropoda",
                       "Abelisauridae", "Carnotaurus"),
      Velociraptor = c("Biota", "Animalia", "Dinosauria", "Theropoda",
                       "Dromaeosauridae", "Velociraptor"),
      Triceratops  = c("Biota", "Animalia", "Dinosauria",
                       "Ornithischia", "Triceratops")
    )
    lins[[taxon]]
  })
  expect_no_error(
    distance_matrix(c("Carnotaurus", "Velociraptor", "Triceratops"),
                    progress = TRUE)
  )
})

test_that("get_taxonomicon_id verbose prints messages on cache hit", {
  clear_cache()
  assign("id_Carnotaurus", "99999", envir = taxodist:::.taxodist_cache)
  expect_no_error(get_taxonomicon_id("Carnotaurus", verbose = TRUE))
})

test_that("get_taxonomicon_id verbose prints warning on network failure", {
  clear_cache()
  mockery::stub(get_taxonomicon_id, "httr::GET", function(...) NULL)
  expect_warning(get_taxonomicon_id("Drosophila", verbose = TRUE), "Cannot reach")
})

test_that("get_taxonomicon_id verbose prints warning on bad status", {
  clear_cache()
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_taxonomicon_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_taxonomicon_id, "httr::status_code", function(...) 503L)
  expect_warning(get_taxonomicon_id("Drosophila", verbose = TRUE), "Cannot reach")
})

test_that("get_lineage_by_id verbose prints messages on cache hit", {
  clear_cache()
  assign("lin_99999",
         c("Biota", "Animalia", "Dinosauria", "Abelisauridae", "Carnotaurus"),
         envir = taxodist:::.taxodist_cache)
  expect_no_error(get_lineage_by_id("99999", verbose = TRUE))
})

test_that("get_lineage_by_id verbose prints warning on network failure", {
  clear_cache()
  mockery::stub(get_lineage_by_id, "httr::GET", function(...) NULL)
  expect_no_error(get_lineage_by_id("00000", verbose = TRUE))
})

test_that("get_lineage_by_id returns NULL when lineage is empty after cleaning", {
  clear_cache()
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_lineage_by_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_lineage_by_id, "httr::status_code", function(...) 200L)
  mockery::stub(get_lineage_by_id, "httr::content",
                function(...) "<html><body></body></html>")
  mockery::stub(get_lineage_by_id, "rvest::read_html",
                function(...) xml2::read_html("<html><body></body></html>"))
  result <- get_lineage_by_id("empty_page")
  expect_null(result)
})

test_that("get_lineage_by_id returns NULL on bad HTTP status", {
  clear_cache()
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_lineage_by_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_lineage_by_id, "httr::status_code", function(...) 404L)
  result <- get_lineage_by_id("99999")
  expect_null(result)
})

test_that("get_lineage_by_id verbose prints warning on bad status", {
  clear_cache()
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_lineage_by_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_lineage_by_id, "httr::status_code", function(...) 503L)
  expect_no_error(get_lineage_by_id("99999", verbose = TRUE))
})

test_that("get_lineage_by_id cache hit with verbose prints message", {
  clear_cache()
  assign("lin_50841",
         c("Biota", "Animalia", "Dinosauria", "Theropoda", "Tyrannosaurus"),
         envir = taxodist:::.taxodist_cache)
  result <- get_lineage_by_id("50841", verbose = TRUE)
  expect_type(result, "character")
})

test_that("get_lineage_by_id with clean = FALSE returns NULL on network failure", {
  clear_cache()
  assign("lin_clean_test", NULL, envir = taxodist:::.taxodist_cache)
  mockery::stub(get_lineage_by_id, "httr::GET", function(...) NULL)
  result <- get_lineage_by_id("clean_test", clean = FALSE)
  expect_null(result)
})

test_that("print.taxodist_result displays output correctly", {
  lin_a <- c("Biota", "Animalia", "Dinosauria", "Theropoda",
             "Abelisauridae", "Carnotaurus")
  lin_b <- c("Biota", "Animalia", "Dinosauria", "Ornithischia", "Triceratops")
  result <- taxodist:::.compute_distance(lin_a, lin_b, "Carnotaurus", "Triceratops")
  expect_no_error(print(result))
  expect_invisible(print(result))
})

test_that("compare_lineages returns invisible NULL when lineage missing", {
  mockery::stub(compare_lineages, "get_lineage", function(...) NULL)
  result <- compare_lineages("Fakeosaurus", "Carnotaurus")
  expect_null(result)
})

test_that("compare_lineages handles mrca_depth == 0", {
  mockery::stub(compare_lineages, "get_lineage", function(taxon, ...) {
    if (taxon == "Drosophila")
      c("Biota", "Animalia", "Arthropoda", "Insecta", "Drosophila")
    else
      c("Fungi", "Ascomycota", "Saccharomyces")
  })
  result <- compare_lineages("Drosophila", "Saccharomyces")
  expect_equal(result$mrca_depth, 0L)
})

test_that("compare_lineages handles case where one lineage is subset of other", {
  mockery::stub(compare_lineages, "get_lineage", function(taxon, ...) {
    if (taxon == "Dinosauria")
      c("Biota", "Animalia", "Dinosauria")
    else
      c("Biota", "Animalia", "Dinosauria", "Theropoda", "Carnotaurus")
  })
  result <- compare_lineages("Dinosauria", "Carnotaurus")
  expect_equal(result$mrca_depth, 3L)
})

test_that("shared_clades returns character(0) when no common ancestor", {
  mockery::stub(shared_clades, "get_lineage", function(taxon, ...) {
    if (taxon == "Drosophila")
      c("Biota", "Animalia", "Arthropoda", "Insecta", "Drosophila")
    else
      c("Fungi", "Ascomycota", "Saccharomyces")
  })
  result <- shared_clades("Drosophila", "Saccharomyces")
  expect_equal(result, character(0))
})

test_that("shared_clades returns NULL when one lineage missing", {
  mockery::stub(shared_clades, "get_lineage", function(...) NULL)
  result <- shared_clades("Fakeosaurus", "Carnotaurus")
  expect_null(result)
})

test_that("is_member returns NULL when lineage not found", {
  mockery::stub(is_member, "get_lineage", function(...) NULL)
  result <- is_member("Fakeosaurus", "Dinosauria")
  expect_null(result)
})

test_that("filter_clade handles NULL result from is_member", {
  mockery::stub(filter_clade, "is_member", function(taxon, clade, ...) {
    if (taxon == "Fakeosaurus") NULL
    else clade %in% list(
      Carnotaurus = c("Dinosauria", "Theropoda"),
      Drosophila  = c("Animalia", "Insecta")
    )[[taxon]]
  })
  result <- filter_clade(
    c("Carnotaurus", "Fakeosaurus", "Drosophila"), "Dinosauria"
  )
  expect_equal(result, "Carnotaurus")
})

test_that("lineage_depth returns NULL when lineage not found", {
  mockery::stub(lineage_depth, "get_lineage", function(...) NULL)
  result <- lineage_depth("Fakeosaurus")
  expect_null(result)
})

test_that("get_taxonomicon_id parses HTML and returns id", {
  clear_cache()
  fake_html <- '
  <html><body><table>
    <tr>
      <td>Carnotaurus - animal - dinosaur</td>
      <td><a class="Valid" href="TaxonTree.aspx?id=12345&src=0">tree</a></td>
    </tr>
  </table></body></html>'
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_taxonomicon_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_taxonomicon_id, "httr::status_code", function(...) 200L)
  mockery::stub(get_taxonomicon_id, "httr::content", function(...) fake_html)
  mockery::stub(get_taxonomicon_id, "get_lineage_by_id", function(...) c("Biota", "Animalia"))
  result <- get_taxonomicon_id("Carnotaurus")
  expect_equal(result, "12345")
})

test_that("get_taxonomicon_id skips astronomical entries", {
  clear_cache()
  fake_html <- '
  <html><body><table>
    <tr>
      <td>Carnotaurus - asteroid - Minor planet</td>
      <td><a class="Valid" href="TaxonTree.aspx?id=99999&src=0">tree</a></td>
    </tr>
    <tr>
      <td>Carnotaurus - animal - dinosaur</td>
      <td><a class="Valid" href="TaxonTree.aspx?id=12345&src=0">tree</a></td>
    </tr>
  </table></body></html>'
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_taxonomicon_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_taxonomicon_id, "httr::status_code", function(...) 200L)
  mockery::stub(get_taxonomicon_id, "httr::content", function(...) fake_html)
  mockery::stub(get_taxonomicon_id, "get_lineage_by_id", function(...) c("Biota", "Animalia"))
  result <- get_taxonomicon_id("Carnotaurus", verbose = TRUE)
  expect_equal(result, "12345")
})

test_that("get_taxonomicon_id skips row matching both bio and astronomical keywords", {
  clear_cache()
  fake_html <- '
  <html><body><table>
    <tr>
      <td>Pterodactylus - animal - Minor planet asteroid</td>
      <td><a class="Valid" href="TaxonTree.aspx?id=99999&src=0">wrong</a></td>
    </tr>
    <tr>
      <td>Pterodactylus - animal - reptile</td>
      <td><a class="Valid" href="TaxonTree.aspx?id=42042&src=0">tree</a></td>
    </tr>
  </table></body></html>'
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_taxonomicon_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_taxonomicon_id, "httr::status_code", function(...) 200L)
  mockery::stub(get_taxonomicon_id, "httr::content", function(...) fake_html)
  mockery::stub(get_taxonomicon_id, "get_lineage_by_id", function(...) c("Biota", "Animalia"))
  result <- get_taxonomicon_id("Pterodactylus", verbose = TRUE)
  expect_equal(result, "42042")
})

test_that("get_taxonomicon_id returns NULL when bio row has no tree link", {
  clear_cache()
  fake_html <- '
  <html><body><table>
    <tr>
      <td>Quercus - plant</td>
      <td><a href="SomeOtherPage.aspx?id=999">no tree link</a></td>
    </tr>
  </table></body></html>'
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_taxonomicon_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_taxonomicon_id, "httr::status_code", function(...) 200L)
  mockery::stub(get_taxonomicon_id, "httr::content", function(...) fake_html)
  result <- get_taxonomicon_id("Quercus", verbose = TRUE)
  expect_null(result)
})

test_that("get_lineage_by_id parses HTML and returns lineage", {
  clear_cache()
  fake_html <- '
  <html><body>
    <a href="TaxonTree.aspx?id=1&src=0">Biota</a>
    <a href="TaxonTree.aspx?id=2&src=0">Animalia</a>
    <a href="TaxonTree.aspx?id=3&src=0">Dinosauria</a>
    <a href="TaxonTree.aspx?id=4&src=0">Theropoda</a>
    <a href="TaxonTree.aspx?id=5&src=0">Carnotaurus</a>
  </body></html>'
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_lineage_by_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_lineage_by_id, "httr::status_code", function(...) 200L)
  mockery::stub(get_lineage_by_id, "httr::content", function(...) fake_html)
  result <- get_lineage_by_id("12345", verbose = TRUE)
  expect_type(result, "character")
  expect_true("Dinosauria" %in% result)
  expect_true("Carnotaurus" %in% result)
})

test_that("get_lineage passes clean and verbose through to get_lineage_by_id", {
  clear_cache()
  mockery::stub(get_lineage, "get_taxonomicon_id", function(...) "12345")
  mockery::stub(get_lineage, "get_lineage_by_id",
                function(id, clean, verbose) {
                  expect_equal(id, "12345")
                  expect_false(clean)
                  expect_true(verbose)
                  c("Biota", "Animalia", "Plantae", "Quercus")
                })
  result <- get_lineage("Quercus", clean = FALSE, verbose = TRUE)
  expect_equal(result, c("Biota", "Animalia", "Plantae", "Quercus"))
})

test_that("plot.taxodist_cluster runs without error", {
  m <- matrix(c(0, 0.2, 0.5, 0.2, 0, 0.3, 0.5, 0.3, 0),
              nrow = 3,
              dimnames = list(c("A", "B", "C"), c("A", "B", "C")))
  cl <- taxo_cluster(stats::as.dist(m))
  expect_no_error(plot(cl))
})

test_that("plot.taxodist_ord runs without error", {
  m <- matrix(c(0, 0.2, 0.5, 0.2, 0, 0.3, 0.5, 0.3, 0),
              nrow = 3,
              dimnames = list(c("A","B","C"), c("A","B","C")))
  ord <- taxo_ordinate(stats::as.dist(m))
  expect_no_error(plot(ord))
  expect_invisible(plot(ord))
})

# ── Mocks for taxo_search ─────────────────────────────────────────────────────

test_that("taxo_search returns NULL on network failure and bad status", {
  clear_cache()
  mockery::stub(taxo_search, "httr::GET", function(...) stop("Network error"))
  expect_null(taxo_search("Bacteria", verbose = TRUE))

  fake_response <- structure(list(), class = "response")
  mockery::stub(taxo_search, "httr::GET", function(...) fake_response)
  mockery::stub(taxo_search, "httr::status_code", function(...) 503L)
  expect_null(taxo_search("Bacteria", verbose = TRUE))
})

test_that("taxo_search returns NULL when no matches are found", {
  clear_cache()
  fake_html <- '
  <html><body><table>
    <tr><td>Not a link</td></tr>
    <tr><td><a href="OtherPage.aspx">No ID here</a></td></tr>
  </table></body></html>'

  fake_response <- structure(list(), class = "response")
  mockery::stub(taxo_search, "httr::GET", function(...) fake_response)
  mockery::stub(taxo_search, "httr::status_code", function(...) 200L)
  mockery::stub(taxo_search, "httr::content", function(...) fake_html)

  expect_null(taxo_search("EmptyTaxon", verbose = TRUE))
})

test_that("taxo_search parses HTML, applies skips, dedups, and returns data.frame", {
  clear_cache()
  fake_html <- '
  <html><body><table>
    <tr>
      <td>Astronomical planet asteroid</td>
      <td><a class="Valid" href="TaxonTree.aspx?id=111">ignore</a></td>
    </tr>
    <tr>
      <td>No links here</td>
      <td>Just text</td>
    </tr>
    <tr>
      <td>Invalid class</td>
      <td><a class="Invalid" href="TaxonTree.aspx?id=222">ignore</a></td>
    </tr>
    <tr>
      <td>Missing ID</td>
      <td><a class="Valid" href="TaxonTree.aspx?wrong=333">ignore</a></td>
    </tr>
    <tr>
      <td><a class="Valid" href="TaxonTree.aspx?id=444">N|T|P|R|B|L Bacteria (Kingdom)</a></td>
    </tr>
    <tr>
      <td><a class="Valid" href="TaxonTree.aspx?id=444">N|T|P|R|B|L Bacteria (Kingdom) Duplicated</a></td>
    </tr>
    <tr>
      <td><a class="Valid" href="TaxonTree.aspx?id=555">N|T|P|R|B|L Bacteria (Domain)</a></td>
    </tr>
  </table></body></html>'

  fake_response <- structure(list(), class = "response")
  mockery::stub(taxo_search, "httr::GET", function(...) fake_response)
  mockery::stub(taxo_search, "httr::status_code", function(...) 200L)
  mockery::stub(taxo_search, "httr::content", function(...) fake_html)

  df <- taxo_search("Bacteria", verbose = TRUE)

  expect_s3_class(df, "data.frame")
  expect_equal(nrow(df), 2)
  expect_equal(df$id, c("444", "555"))
  expect_equal(df$name[1], "Bacteria (Kingdom)")
  expect_equal(df$name[2], "Bacteria (Domain)")
})

# ── Network tests (skipped on CRAN) ──────────────────────────────────────────

skip_if_taxonomicon_down <- function() {
  res <- tryCatch(httr::GET("http://taxonomicon.taxonomy.nl", httr::timeout(3)), error = function(e) NULL)
  if (is.null(res) || httr::status_code(res) != 200) {
    testthat::skip("Taxonomicon server is currently offline.")
  }
}

test_that("get_lineage returns correct lineage for Velociraptor", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  lin <- get_lineage("Velociraptor")
  expect_type(lin, "character")
  expect_true(length(lin) > 0)
  expect_true("Dinosauria" %in% lin)
  expect_true("Theropoda" %in% lin)
  expect_true("Dromaeosauridae" %in% lin)
})

test_that("get_lineage returns correct lineage for Tyrannosaurus", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  lin <- get_lineage("Tyrannosaurus")
  expect_type(lin, "character")
  expect_true("Coelurosauria" %in% lin)
  expect_true("Dinosauria" %in% lin)
})

test_that("get_lineage returns correct lineage for Carnotaurus", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  lin <- get_lineage("Carnotaurus")
  expect_type(lin, "character")
  expect_true("Dinosauria" %in% lin)
  expect_true("Theropoda" %in% lin)
})

test_that("get_lineage returns correct lineage for Homo", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  lin <- get_lineage("Homo")
  expect_true("Amniota" %in% lin)
  expect_true("Mammalia" %in% lin)
})

test_that("get_lineage returns correct lineage for Drosophila", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  lin <- suppressWarnings(get_lineage("Drosophila"))
  skip_if(is.null(lin), "Taxonomicon unstable")
  expect_type(lin, "character")
  expect_true(length(lin) > 0)
  expect_true("Animalia" %in% lin)
})

test_that("get_lineage returns NULL for unknown taxon", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  expect_null(get_lineage("Fakeosaurus"))
})

test_that("taxo_distance returns valid result for Tyrannosaurus vs Velociraptor", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  result <- taxo_distance("Tyrannosaurus", "Velociraptor")
  expect_s3_class(result, "taxodist_result")
  expect_gte(result$distance, 0)
  expect_lte(result$distance, 1)
  expect_equal(result$mrca, "Tyrannoraptora")
})

test_that("taxo_distance returns 0 when one taxon is ancestor of other", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  res1 <- taxo_distance("Tyrannosaurus", "Dinosauria")
  skip_if(is.null(res1), "Taxonomicon unstable")
  expect_equal(res1$distance, 0)

  res2 <- taxo_distance("Carnotaurus", "Ceratosauria")
  skip_if(is.null(res2), "Taxonomicon unstable")
  expect_equal(res2$distance, 0)
})

test_that("taxo_distance between Carnotaurus and Triceratops is valid", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  result <- taxo_distance("Carnotaurus", "Triceratops")
  expect_s3_class(result, "taxodist_result")
  expect_equal(result$mrca, "Dinosauria")
})

test_that("taxo_distance is larger between distant taxa than close ones", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  d_close  <- taxo_distance("Carnotaurus", "Tyrannosaurus")$distance
  d_distant <- taxo_distance("Carnotaurus", "Homo")$distance
  expect_gt(d_distant, d_close)
})

test_that("mrca of Tyrannosaurus and Triceratops is Dinosauria", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  expect_equal(mrca("Tyrannosaurus", "Triceratops"), "Dinosauria")
})

test_that("mrca of Tyrannosaurus and Homo is Amniota", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  expect_equal(mrca("Tyrannosaurus", "Homo"), "Amniota")
})

test_that("mrca of Velociraptor and Triceratops is Dinosauria", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  expect_equal(mrca("Velociraptor", "Triceratops"), "Dinosauria")
})

test_that("mrca of Carnotaurus and Tyrannosaurus is within Theropoda", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  ancestor <- mrca("Carnotaurus", "Tyrannosaurus")
  skip_if(is.null(ancestor), "Taxonomicon unstable")
  lin <- get_lineage("Tyrannosaurus")
  expect_true(ancestor %in% lin)
})

test_that("is_member correctly identifies clade membership", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  expect_true(is_member("Tyrannosaurus", "Theropoda"))
  expect_false(is_member("Tyrannosaurus", "Ornithischia"))
})

test_that("lineage_depth for Carnotaurus is reasonable", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  expect_gt(lineage_depth("Carnotaurus"), 10)
})

test_that("get_taxonomicon_id finds real ID and caches it", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  id <- get_taxonomicon_id("Carnotaurus", verbose = TRUE)
  expect_type(id, "character")
  expect_false(is.null(id))
  expect_equal(id, get("id_Carnotaurus", envir = taxodist:::.taxodist_cache))
})

test_that("get_lineage_by_id parses and caches lineage for Drosophila", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  id <- suppressWarnings(get_taxonomicon_id("Drosophila"))
  skip_if(is.null(id), "Taxonomicon unstable")
  result <- get_lineage_by_id(id, verbose = TRUE)
  expect_type(result, "character")
  expect_true("Animalia" %in% result)
})

test_that("get_lineage_by_id clean = FALSE keeps more nodes than clean = TRUE", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  id <- get_taxonomicon_id("Carnotaurus")
  result_clean <- get_lineage_by_id(id, clean = TRUE)
  clear_cache()
  result_no_clean <- get_lineage_by_id(id, clean = FALSE)
  expect_lte(length(result_clean), length(result_no_clean))
})

test_that("get_lineage verbose wrapper works for Quercus", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  result <- get_lineage("Quercus", verbose = TRUE)
  expect_type(result, "character")
  expect_true("Biota" %in% result)
})

test_that("get_taxonomicon_id returns NULL for genuinely unknown taxon", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  expect_null(get_taxonomicon_id("Zzzznotarealgenus99999", verbose = TRUE))
})

test_that("get_taxonomicon_id skips astronomical objects", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  id <- get_taxonomicon_id("Venus", verbose = TRUE)
  if (!is.null(id)) {
    lin <- get_lineage_by_id(id)
    expect_true(!is.null(lin))
  } else {
    expect_null(id)
  }
})

test_that("get_lineage_by_id works directly with verbose", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  id <- get_taxonomicon_id("Carnotaurus")
  expect_no_error(get_lineage_by_id(id, verbose = TRUE))
})

test_that("get_taxonomicon_id works with verbose for real taxon", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  expect_no_error(suppressWarnings(get_taxonomicon_id("Drosophila", verbose = TRUE)))
})

test_that("get_taxonomicon_id verbose prints not found warning", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  expect_null(get_taxonomicon_id("Zzzzfakeosaurus99999", verbose = TRUE))
})

test_that("get_lineage_by_id verbose success message fires on real taxon", {
  skip_if_offline()
  skip_if_taxonomicon_down()
  clear_cache()
  id <- get_taxonomicon_id("Carnotaurus")
  expect_no_error(get_lineage_by_id(id, verbose = TRUE))
})

test_that("get_taxonomicon_id skips entry whose lineage has no Biota", {
  clear_cache()
  fake_html <- '
  <html><body><table>
    <tr>
      <td>Carnotaurus - animal - dinosaur</td>
      <td><a class="Valid" href="TaxonTree.aspx?id=12345&src=0">tree</a></td>
    </tr>
  </table></body></html>'
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_taxonomicon_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_taxonomicon_id, "httr::status_code", function(...) 200L)
  mockery::stub(get_taxonomicon_id, "httr::content", function(...) fake_html)
  mockery::stub(get_taxonomicon_id, "get_lineage_by_id",
                function(...) c("NotBiota", "Animalia", "Dinosauria"))
  result <- get_taxonomicon_id("Carnotaurus")
  expect_null(result)
})

test_that("get_lineage_by_id truncates at own id when present in links", {
  clear_cache()
  fake_html <- '
  <html><body>
    <a href="TaxonTree.aspx?id=1&src=0">Biota</a>
    <a href="TaxonTree.aspx?id=2&src=0">Animalia</a>
    <a href="TaxonTree.aspx?id=99&src=0">Carnotaurus</a>
    <a href="TaxonTree.aspx?id=100&src=0">SomeChild</a>
  </body></html>'
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_lineage_by_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_lineage_by_id, "httr::status_code", function(...) 200L)
  mockery::stub(get_lineage_by_id, "httr::content", function(...) fake_html)
  result <- get_lineage_by_id("99")
  expect_true("Carnotaurus" %in% result)
  expect_false("SomeChild" %in% result)
})

test_that("get_lineage handles binomial taxon name correctly", {
  clear_cache()
  mockery::stub(get_lineage, "get_taxonomicon_id", function(...) "12345")
  mockery::stub(get_lineage, "get_lineage_by_id",
                function(...) c("Biota", "Animalia", "Dinosauria",
                                "Theropoda", "Carnotaurus sastrei"))
  result <- get_lineage("Carnotaurus sastrei")
  expect_true("Carnotaurus sastrei" %in% result)
})

test_that("get_lineage returns NULL when get_lineage_by_id returns NULL", {
  clear_cache()
  mockery::stub(get_lineage, "get_taxonomicon_id", function(...) "12345")
  mockery::stub(get_lineage, "get_lineage_by_id", function(...) NULL)
  result <- get_lineage("Carnotaurus")
  expect_null(result)
})

test_that("get_lineage returns single-node lineage when lineage_by_id returns empty", {
  clear_cache()
  mockery::stub(get_lineage, "get_taxonomicon_id", function(...) "12345")
  mockery::stub(get_lineage, "get_lineage_by_id", function(...) character(0))
  result <- get_lineage("Carnotaurus")
  expect_equal(result, "Carnotaurus")
})

test_that("get_lineage appends taxon name when not found in scraped lineage", {
  clear_cache()
  mockery::stub(get_lineage, "get_taxonomicon_id", function(...) "12345")
  mockery::stub(get_lineage, "get_lineage_by_id",
                function(...) c("Biota", "Animalia", "Dinosauria", "Theropoda"))
  result <- get_lineage("Carnotaurus")
  expect_equal(tail(result, 1), "Carnotaurus")
})

test_that("get_lineage_by_id returns NULL when all links are filtered out", {
  clear_cache()
  fake_html <- '
  <html><body>
    <a href="TaxonTree.aspx?id=1&src=0">Go to</a>
    <a href="TaxonTree.aspx?id=2&src=0">[unranked]</a>
  </body></html>'
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_lineage_by_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_lineage_by_id, "httr::status_code", function(...) 200L)
  mockery::stub(get_lineage_by_id, "httr::content", function(...) fake_html)
  result <- get_lineage_by_id("99999")
  expect_null(result)
})

test_that("get_taxonomicon_id warns on multiple biological entries (coverage)", {
  clear_cache()
  assign("lin_111", c("Biota", "Animalia", "Fake1"), envir = taxodist:::.taxodist_cache)
  assign("lin_222", c("Biota", "Animalia", "Fake2"), envir = taxodist:::.taxodist_cache)

  fake_html <- '
  <html><body><table>
    <tr>
      <td>Nereis - animal - one</td>
      <td><a class="Valid" href="TaxonTree.aspx?id=111&src=0">tree</a></td>
    </tr>
    <tr>
      <td>Nereis - animal - two</td>
      <td><a class="Valid" href="TaxonTree.aspx?id=222&src=0">tree</a></td>
    </tr>
  </table></body></html>'
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_taxonomicon_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_taxonomicon_id, "httr::status_code", function(...) 200L)
  mockery::stub(get_taxonomicon_id, "httr::content", function(...) fake_html)
  expect_warning(
    get_taxonomicon_id("Nereis"),
    "Multiple valid biological entries"
  )
})

test_that("deduplication preserves order", {
  clear_cache()
  mockery::stub(get_lineage_by_id, "httr::GET", function(...) {
    structure(list(), class = "response")
  })
  mockery::stub(get_lineage_by_id, "httr::status_code", function(...) 200L)
  mockery::stub(get_lineage_by_id, "httr::content", function(...) "x")
  mockery::stub(get_lineage_by_id, "rvest::read_html", function(...) {
    xml2::read_html('
      <html><body>
        <a href="TaxonTree.aspx?id=1&src=0">Biota</a>
        <a href="TaxonTree.aspx?id=2&src=0">Animalia</a>
        <a href="TaxonTree.aspx?id=3&src=0">Uropygi</a>
        <a href="TaxonTree.aspx?id=3&src=0">Uropygi</a>
        <a href="TaxonTree.aspx?id=4&src=0">Thelyphonida</a>
      </body></html>')
  })
  result <- get_lineage_by_id("4")
  expect_equal(result, c("Biota", "Animalia", "Uropygi", "Thelyphonida"))
})

test_that("taxo_cluster returns correct S3 class", {
  mockery::stub(taxo_cluster, "distance_matrix", function(...) {
    m <- matrix(c(0, 0.2, 0.5, 0.2, 0, 0.3, 0.5, 0.3, 0),
                nrow = 3,
                dimnames = list(c("A", "B", "C"), c("A", "B", "C")))
    stats::as.dist(m)
  })
  result <- taxo_cluster(c("A", "B", "C"), progress = FALSE)
  expect_s3_class(result, "taxodist_cluster")
})

test_that("taxo_cluster result contains hclust and dist", {
  mockery::stub(taxo_cluster, "distance_matrix", function(...) {
    m <- matrix(c(0, 0.2, 0.5, 0.2, 0, 0.3, 0.5, 0.3, 0),
                nrow = 3,
                dimnames = list(c("A", "B", "C"), c("A", "B", "C")))
    stats::as.dist(m)
  })
  result <- taxo_cluster(c("A", "B", "C"), progress = FALSE)
  expect_s3_class(result$hclust, "hclust")
  expect_s3_class(result$dist, "dist")
})

test_that("taxo_cluster accepts a dist object directly", {
  m <- matrix(c(0, 0.2, 0.5, 0.2, 0, 0.3, 0.5, 0.3, 0),
              nrow = 3,
              dimnames = list(c("A", "B", "C"), c("A", "B", "C")))
  d <- stats::as.dist(m)
  result <- taxo_cluster(d)
  expect_s3_class(result, "taxodist_cluster")
})

test_that("taxo_ordinate returns correct S3 class", {
  mockery::stub(taxo_ordinate, "distance_matrix", function(...) {
    m <- matrix(c(0, 0.2, 0.5, 0.2, 0, 0.3, 0.5, 0.3, 0),
                nrow = 3,
                dimnames = list(c("A","B","C"), c("A","B","C")))
    stats::as.dist(m)
  })
  result <- taxo_ordinate(c("A", "B", "C"), progress = FALSE)
  expect_s3_class(result, "taxodist_ord")
})

test_that("taxo_ordinate result contains points, dist and GOF", {
  mockery::stub(taxo_ordinate, "distance_matrix", function(...) {
    m <- matrix(c(0, 0.2, 0.5, 0.2, 0, 0.3, 0.5, 0.3, 0),
                nrow = 3,
                dimnames = list(c("A","B","C"), c("A","B","C")))
    stats::as.dist(m)
  })
  result <- taxo_ordinate(c("A", "B", "C"), progress = FALSE)
  expect_true(!is.null(result$points))
  expect_s3_class(result$dist, "dist")
  expect_true(!is.null(result$GOF))
})

test_that("taxo_ordinate points matrix has correct dimensions", {
  mockery::stub(taxo_ordinate, "distance_matrix", function(...) {
    m <- matrix(c(0, 0.2, 0.5, 0.2, 0, 0.3, 0.5, 0.3, 0),
                nrow = 3,
                dimnames = list(c("A","B","C"), c("A","B","C")))
    stats::as.dist(m)
  })
  result <- taxo_ordinate(c("A", "B", "C"), k = 2, progress = FALSE)
  expect_equal(ncol(result$points), 2)
  expect_equal(nrow(result$points), 3)
})

test_that("taxo_ordinate accepts a dist object directly", {
  m <- matrix(c(0, 0.2, 0.5, 0.2, 0, 0.3, 0.5, 0.3, 0),
              nrow = 3,
              dimnames = list(c("A","B","C"), c("A","B","C")))
  d <- stats::as.dist(m)
  result <- taxo_ordinate(d, k = 2)
  expect_s3_class(result, "taxodist_ord")
})

test_that("summary.taxodist_ord computes variance and handles missing eigenvalues", {
  mock_ord <- structure(list(
    points = matrix(1:4, ncol = 2, dimnames = list(c("A", "B"), NULL)),
    dist   = stats::dist(1:2),
    GOF    = c(0.95, 0.95),
    eig    = c(2.0, 1.0, -0.5)
  ), class = "taxodist_ord")
  out <- capture.output(res <- summary(mock_ord))
  expect_s3_class(res, "data.frame")
  expect_equal(nrow(res), 2)
  expect_equal(res$Axis, c("PC1", "PC2"))
  expect_equal(round(res$Variance_Pct[1]), 67)
  expect_equal(round(res$Variance_Pct[2]), 33)
  mock_bad <- mock_ord
  mock_bad$eig <- NULL
  expect_message(res_bad <- summary(mock_bad), "Eigenvalues not found")
  expect_null(res_bad)
})

test_that("taxo_heatmap plots correctly and returns dist invisibly", {
  mock_dist <- stats::dist(matrix(1:4, ncol = 2))
  attr(mock_dist, "Labels") <- c("TaxonA", "TaxonB")
  pdf(file = NULL)
  res <- taxo_heatmap(mock_dist)
  dev.off()
  expect_s3_class(res, "dist")
})

test_that("get_lineage accepts numeric IDs directly without searching", {
  clear_cache()
  mockery::stub(get_lineage, "get_lineage_by_id", function(...) c("Biota", "Bacteria"))
  result <- get_lineage("71320")
  expect_equal(result, c("Biota", "Bacteria"))
})

test_that("get_lineage_by_id returns NULL for non-numeric strings", {
  expect_null(get_lineage_by_id("Bacteria"))
  expect_null(get_lineage_by_id("123x"))
  expect_null(get_lineage_by_id("   "))
})

test_that("get_taxonomicon_id follows taxonomic redirects", {
  clear_cache()
  assign("lin_16197", c("Biota", "Animalia", "Uropygi"), envir = taxodist:::.taxodist_cache)
  fake_html <- '
  <html><body><table>
    <tr>
      <td>Thelyphonida see Uropygi</td>
      <td>
        <a class="Invalid" href="TaxonTree.aspx?id=123&src=0">old</a>
        <a class="Valid" href="TaxonTree.aspx?id=16197&src=0">tree</a>
      </td>
    </tr>
  </table></body></html>'
  fake_response <- structure(list(), class = "response")
  mockery::stub(get_taxonomicon_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_taxonomicon_id, "httr::status_code", function(...) 200L)
  mockery::stub(get_taxonomicon_id, "httr::content", function(...) fake_html)
  result <- get_taxonomicon_id("Thelyphonida")
  expect_equal(result, "16197")
})

test_that("get_taxonomicon_id skips rows with no Valid links", {
  clear_cache()
  assign("lin_222", c("Biota", "Animalia"), envir = taxodist:::.taxodist_cache)

  # HTML falso: A primeira linha NÃO tem classe Valid. O pacote deve pular (next).
  fake_html <- '
  <html><body><table>
    <tr>
      <td>Invalid taxon</td>
      <td><a class="Invalid" href="TaxonTree.aspx?id=111&src=0">skip me</a></td>
    </tr>
    <tr>
      <td>Good taxon - animal</td>
      <td><a class="Valid" href="TaxonTree.aspx?id=222&src=0">tree</a></td>
    </tr>
  </table></body></html>'

  fake_response <- structure(list(), class = "response")
  mockery::stub(get_taxonomicon_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_taxonomicon_id, "httr::status_code", function(...) 200L)
  mockery::stub(get_taxonomicon_id, "httr::content", function(...) fake_html)

  result <- get_taxonomicon_id("Good taxon")
  expect_equal(result, "222")
})

test_that("get_taxonomicon_id skips valid links missing numeric IDs", {
  clear_cache()
  assign("lin_333", c("Biota", "Animalia"), envir = taxodist:::.taxodist_cache)

  fake_html <- '
  <html><body><table>
    <tr>
      <td>Missing ID taxon</td>
      <td><a class="Valid" href="TaxonTree.aspx?wrongparam=abc">skip me</a></td>
    </tr>
    <tr>
      <td>Good taxon - animal</td>
      <td><a class="Valid" href="TaxonTree.aspx?id=333&src=0">tree</a></td>
    </tr>
  </table></body></html>'

  fake_response <- structure(list(), class = "response")
  mockery::stub(get_taxonomicon_id, "httr::GET", function(...) fake_response)
  mockery::stub(get_taxonomicon_id, "httr::status_code", function(...) 200L)
  mockery::stub(get_taxonomicon_id, "httr::content", function(...) fake_html)

  result <- get_taxonomicon_id("Good taxon")
  expect_equal(result, "333")
})

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(taxodist)

## ----ultrametric-check--------------------------------------------------------
m <- as.matrix(taxobase$statistical_matrix)
tolerance <- sqrt(.Machine$double.eps)

ultrametric_ok <- function(index) {
  i <- index[1]
  j <- index[2]
  k <- index[3]

  d_ij <- m[i, j]
  d_ik <- m[i, k]
  d_jk <- m[j, k]

  d_ij <= max(d_ik, d_jk) + tolerance &&
    d_ik <= max(d_ij, d_jk) + tolerance &&
    d_jk <= max(d_ij, d_ik) + tolerance
}

all(combn(seq_len(nrow(m)), 3, FUN = ultrametric_ok))

---
title: "Statistical Applications of taxodist"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Statistical Applications of taxodist}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
library(taxodist)
```

The `taxodist` distance matrix is a standard `dist` object and integrates
directly with the R ecosystem for multivariate analysis. This vignette shows
how to use it alongside `vegan` and `ape` for ecological and phylogenetic
analyses. All examples use a set of threatened mammals from Brazil spanning
five broad taxonomic groups.

---

## Example dataset: threatened mammals of Brazil

```{r taxa}
taxa_brazil <- c(
  "Priodontes", "Myrmecophaga", "Chrysocyon", "Tapirus", "Didelphis",
  "Leontopithecus", "Brachyteles",
  "Panthera", "Pteronura", "Puma",
  "Sotalia", "Pontoporia",
  "Trichechus", "Mazama", "Blastocerus"
)
```

The dataset covers xenarthrans (`Priodontes`, `Myrmecophaga`), a marsupial
(`Didelphis`), ungulates (`Tapirus`, `Mazama`, `Blastocerus`), primates
(`Leontopithecus`, `Brachyteles`), carnivores (`Chrysocyon`, `Panthera`,
`Pteronura`, `Puma`), cetaceans (`Sotalia`, `Pontoporia`), and a sirenian
(`Trichechus`).

The pairwise distance matrix is computed with a single call. Lineages are
cached after the first retrieval, so subsequent analyses on the same taxa
are instantaneous.

```{r matrix}
library(taxodist)
mat <- distance_matrix(taxa_brazil)
print(mat)
#>                Priodontes Myrmecophaga Chrysocyon    Tapirus  Didelphis
#> Myrmecophaga   0.01639344
#> Chrysocyon     0.01694915   0.01694915
#> Tapirus        0.01694915   0.01694915 0.01587302
#> Didelphis      0.01754386   0.01754386 0.01754386 0.01754386
#> Leontopithecus 0.01694915   0.01694915 0.01666667 0.01666667 0.01754386
#> Brachyteles    0.01694915   0.01694915 0.01666667 0.01666667 0.01754386
#> Panthera       0.01694915   0.01694915 0.01492537 0.01587302 0.01754386
#> Pteronura      0.01694915   0.01694915 0.01470588 0.01587302 0.01754386
#> Puma           0.01694915   0.01694915 0.01492537 0.01587302 0.01754386
#> Sotalia        0.01694915   0.01694915 0.01587302 0.01562500 0.01754386
#> Pontoporia     0.01694915   0.01694915 0.01587302 0.01562500 0.01754386
#> Trichechus     0.01666667   0.01666667 0.01694915 0.01694915 0.01754386
#> Mazama         0.01694915   0.01694915 0.01587302 0.01562500 0.01754386
#> Blastocerus    0.01694915   0.01694915 0.01587302 0.01562500 0.01754386
```

`Didelphis` shows the largest distances to all other taxa (0.01754),
consistent with the early divergence of marsupials from placental mammals.
Carnivores (`Panthera`, `Pteronura`, `Puma`) show smaller distances among
themselves, reflecting a more recent common ancestor within Carnivora.

---

## Phylogenetic tree with `ape`

`taxo_cluster()` performs hierarchical clustering on the distance matrix and
returns an object that `ape` can convert directly to a `phylo` tree for
publication-quality visualisation.

```{r ape, fig.width = 7, fig.height = 5}
library(ape)

cl   <- taxo_cluster(taxa_brazil)
tree <- ape::as.phylo(cl$hclust)

plot(tree,
     main      = "Threatened Mammals of Brazil",
     cex       = 0.85,
     tip.color = "gray20")
```

The `phylo` object is fully compatible with all `ape` tree manipulation and
plotting functions, including `nodelabels()`, `edgelabels()`, and export to
Newick format via `write.tree()`.

```{r ape-newick}
ape::write.tree(tree, file = "taxa_brazil.nwk")
```

---

## Taxonomic diversity with `vegan`

### Clarke and Warwick indices (`taxondive`)

`vegan::taxondive()` computes a family of taxonomic diversity indices
(Clarke and Warwick 1998, 2001) from a community matrix and a taxonomic
distance matrix. These indices capture not just species richness but how
distantly related the species in a community are: a community of closely
related species scores lower than one with the same richness but spanning
multiple orders.

The function expects a community matrix (sites x species) and the `dist`
object from `distance_matrix()`. Here we define three hypothetical
communities, each dominated by one broad taxonomic group.

```{r taxondive}
set.seed(123)
library(vegan)

comm <- matrix(c(
  1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   # xenarthrans + marsupial
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,   # primates + carnivores
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1    # cetaceans + sirenian + ungulates
), nrow = 3, byrow = TRUE)

colnames(comm) <- taxa_brazil
rownames(comm) <- c("community_A", "community_B", "community_C")

vegan::taxondive(comm, mat)
#>               Species      Delta       Delta*     Lambda+     Delta+     S Delta+
#> community_A  3.0000e+00  1.7160e-02  1.7160e-02  2.9410e-07  1.7160e-02   0.0515
#> community_B  5.0000e+00  1.5905e-02  1.5905e-02  8.8325e-07  1.5905e-02   0.0795
#> community_C  5.0000e+00  1.5750e-02  1.5750e-02  1.1834e-06  1.5750e-02   0.0788
#> Expected                -9.9265e-02  1.6544e-02              1.6457e-02         
```

The key indices are:

**Delta** ($\Delta$): mean taxonomic distance between all pairs of species in
the community.

**Delta*** ($\Delta^*$): mean pairwise distance excluding self-comparisons;
comparable across communities of different sizes.

**Delta+** ($\Delta^+$): mean pairwise distance weighted by species
abundances. For presence-absence data this equals $\Delta^*$.

**Lambda+** ($\Lambda^+$): variance in taxonomic distances, measuring how
evenly a community spans the taxonomic tree.

`community_A`, dominated by xenarthrans and a marsupial, scores the highest
$\Delta$ (0.01673), reflecting deeper divergences among its members.
`community_B` (primates and carnivores) scores the lowest (0.01591), as its
members share more recent common ancestors within Placentalia.

---

## Mantel test with `vegan`

The Mantel test assesses the correlation between two distance matrices by
permutation. A typical application is testing whether taxonomic distance
correlates with geographic distance, a signature of phylogeographic structure.

Here we use simulated coordinates for illustration. In practice, replace
`coords` with observed latitude and longitude values.

```{r mantel}
set.seed(42)
coords <- matrix(rnorm(30), ncol = 2)
rownames(coords) <- taxa_brazil
geo_dist <- dist(coords)

vegan::mantel(mat, geo_dist)
#> Mantel statistic based on Pearson's product-moment correlation
#>
#> Call:
#> vegan::mantel(xdis = mat, ydis = geo_dist)
#>
#> Mantel statistic r: -0.0569
#>       Significance: 0.653
#>
#> Upper quantiles of permutations (null model):
#>   90%   95% 97.5%   99% 
#> 0.188 0.237 0.272 0.336 
#> Permutation: free
#> Number of permutations: 999
```

The non-significant result (r = -0.27, p = 0.972) is expected with random
coordinates. For non-normal distance distributions, `"spearman"` is more
robust than the default `"pearson"`.

```{r mantel-spearman}
vegan::mantel(mat, geo_dist, method = "spearman", permutations = 9999)
#> Mantel statistic based on Spearman's rank correlation rho 
#>
#> Call:
#> vegan::mantel(xdis = mat, ydis = geo_dist, method = "spearman",
#>     permutations = 9999)
#>
#> Mantel statistic r: -0.07405
#>       Significance: 0.672
#>
#> Upper quantiles of permutations (null model):
#>   90%   95% 97.5%   99%
#> 0.189 0.244 0.293 0.353 
#> Permutation: free
#> Number of permutations: 9999
```

---

## PERMANOVA with `vegan`

PERMANOVA (Anderson 2001) tests whether groups of taxa differ significantly
in taxonomic distance space. It partitions total variance in the distance
matrix into within-group and between-group components and assesses
significance by permutation, with no assumption of multivariate normality.

```{r permanova}
groups <- c(
  "xenarthra", "xenarthra", "carnivore", "ungulate",  "marsupial",
  "primate",   "primate",
  "carnivore", "carnivore", "carnivore",
  "cetacean",  "cetacean",
  "sirenian",  "ungulate",  "ungulate"
)

vegan::adonis2(mat ~ groups)
#> Permutation test for adonis under reduced model
#> Permutation: free
#> Number of permutations: 999
#>
#> vegan::adonis2(formula = mat ~ groups)
#>          Df  SumOfSqs      R2      F Pr(>F)
#> Model     6 0.00100049 0.52646 1.4823  0.001 ***
#> Residual  8 0.00089992 0.47354  
#> Total    14 0.00190041 1.00000   
#> ---
#> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
```

The taxonomic grouping explains $49.6\%$ of the total variance in the distance
matrix ($\text{R}^2 = 0.496$, $\text{p} = 0.001$). This confirms that the broad mammalian
orders occupy distinct regions of taxonomic distance space, consistent with
their deep evolutionary divergences.

For pairwise comparisons between groups after a significant PERMANOVA,
see `pairwise.adonis2()` in the `pairwiseAdonis` package.

---

## A complete workflow

The following script chains all steps from taxon names to multivariate
analysis.

```{r workflow}
library(taxodist)
library(vegan)
library(ape)

# 1. define taxa
taxa_brazil <- c(
  "Priodontes", "Myrmecophaga", "Chrysocyon", "Tapirus", "Didelphis",
  "Leontopithecus", "Brachyteles",
  "Panthera", "Pteronura", "Puma",
  "Sotalia", "Pontoporia",
  "Trichechus", "Mazama", "Blastocerus"
)

# 2. compute distance matrix
mat <- distance_matrix(taxa_brazil)

# 3. hierarchical clustering and phylo plot
cl   <- taxo_cluster(taxa_brazil)
tree <- ape::as.phylo(cl$hclust)
plot(tree, main = "Threatened Mammals of Brazil", cex = 0.85)

# 4. PCoA ordination
ord <- taxo_ordinate(mat)
plot(ord, main = "PCoA: Taxonomic Distance Space")

# 5. PERMANOVA
groups <- c(
  "xenarthra", "xenarthra", "carnivore", "ungulate",  "marsupial",
  "primate",   "primate",
  "carnivore", "carnivore", "carnivore",
  "cetacean",  "cetacean",
  "sirenian",  "ungulate",  "ungulate"
)
vegan::adonis2(mat ~ groups)
```

---

## References

Anderson, M.J. (2001). A new method for non-parametric multivariate
analysis of variance. *Austral Ecology*, 26, 32--46.

Clarke, K.R. and Warwick, R.M. (1998). A taxonomic distinctness index and
its statistical properties. *Journal of Applied Ecology*, 35, 523--531.

Clarke, K.R. and Warwick, R.M. (2001). A further biodiversity index
applicable to species lists: variation in taxonomic distinctness.
*Marine Ecology Progress Series*, 216, 265--278.

Brands, S.J. (1989 onwards). *Systema Naturae 2000*. Amsterdam,
The Netherlands. Retrieved from The Taxonomicon,
<http://taxonomicon.taxonomy.nl>.

---
title: "Introduction to taxodist"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to taxodist}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
library(taxodist)
```

## What is taxodist?

`taxodist` answers a simple question: *how related are any two living things?*

Given any two taxon names, a pair of dinosaurs, a dinosaur and a fungus, two
species of fly, or an oak tree and a human, `taxodist` retrieves their full
hierarchical lineages from [The Taxonomicon](http://taxonomicon.taxonomy.nl) and
computes a dissimilarity index between them.

The Taxonomicon is based on *Systema Naturae 2000* (Brands, 1989 onwards) and
provides exceptionally deep lineage resolution, substantially exceeding other programmatic sources.

Searches work at any taxonomic level: genus, species, family, order, or
any clade. Both `"Tyrannosaurus"` and `"Tyrannosaurus rex"` are valid inputs, as
are `"Drosophila melanogaster"`, `"Homo sapiens"`, or `"Araucaria angustifolia"`.

---

## The distance metric

`taxodist` measures how related two taxa are by asking a single question:
*how deep is their most recent common ancestor?*

$$d(A, B) = \frac{1}{\text{depth}(\text{MRCA}(A,B))}$$

The deeper the shared ancestor, the smaller the distance, meaning the more
related the two taxa are. A shallow MRCA (close to the root) means the two
taxa diverged early and are distantly related; a deep MRCA means they share
a long common history and are closely related.

This has a key property: taxa that diverged at the same point
in the tree are always equidistant from any third taxon, regardless of how
many nodes each has in its lineage below the split. For example:

- *Tyrannosaurus* and *Velociraptor* are both Tetanurae: they diverged from
  *Carnotaurus* (Ceratosauria) at the same node (Averostra), so both have
  exactly the same distance to Carnotaurus;
  
- All dinosaurs diverged from the mammal lineage at the same
  node (Amniota), so *Homo sapiens* is equally distant from *Tyrannosaurus*, *Triceratops*, *Carnotaurus* and *Cyanocorax*.

The distance is bounded to $[0, 1]$, where 0 means identical taxa, 1 means the taxa share only the root as their most recent common ancestor, and smaller values reflect increasingly recent shared ancestry.

---

## Basic usage

### Getting a lineage

```{r lineage}
lin <- get_lineage("Tyrannosaurus")
tail(lin, 8)
#> [1] "Avetheropoda"     "Coelurosauria"    "Tyrannoraptora"   "Tyrannosauroidea"
#> [5] "Tyrannosauridae"  "Tyrannosaurinae"  "Tyrannosaurini"   "Tyrannosaurus" 
```

Species-level searches also work:

```{r lineage-species}
lin <- get_lineage("Drosophila melanogaster")
tail(lin, 4)
#> [1] "Ephydroidea"             "Drosophilidae"           "Drosophilinae"          
#> [4] "Drosophila melanogaster"
```

### Computing distance between two taxa

```{r distance}
result <- taxo_distance("Tyrannosaurus", "Velociraptor")
print(result)
#> -- Taxonomic Distance --
#> 
#> * Tyrannosaurus vs Velociraptor
#>   Distance : 0.0153846153846154
#>   MRCA : Tyrannoraptora (depth 65)
#>   Depth A : 70
#>   Depth B : 73
```

The distance between a dinosaur and a mammal or a bacteria and a human is larger:

```{r distance-far}
taxo_distance("Tyrannosaurus", "Homo")$distance        # 0.02777778
taxo_distance("Tyrannosaurus", "Drosophila")$distance  # 0.06666667
taxo_distance("Tyrannosaurus", "Quercus")$distance     # 0.25
taxo_distance("Escherichia", "Homo")$distance          # 1
```

### Finding the most recent common ancestor

```{r mrca}
mrca("Tyrannosaurus", "Velociraptor")  # "Tyrannoraptora"
mrca("Tyrannosaurus", "Triceratops")   # "Dinosauria"
mrca("Tyrannosaurus", "Homo")          # "Amniota"
mrca("Tyrannosaurus", "Drosophila")    # "Nephrozoa"
mrca("Tyrannosaurus", "Quercus")       # "discaria"
```

---

## Working with multiple taxa

### Pairwise distance matrix

```{r matrix}
taxa <- c(
  "Tyrannosaurus", "Carnotaurus", "Triceratops",
  "Parasaurolophus", "Stegosaurus", "Brachiosaurus",
  "Homo sapiens", "Homo neanderthalensis", "Pan troglodytes",
  "Panthera leo", "Canis lupus",
  "Ornithorhynchus anatinus",
  "Loxodonta africana",       
  "Struthio camelus",           
  "Aptenodytes forsteri",       
  "Ara ararauna",            
  "Crocodylus niloticus",
  "Chelonia mydas",
  "Ambystoma mexicanum",
  "Octopus vulgaris",
  "Carcharodon carcharias",    
  "Balaenoptera musculus",   
  "Drosophila melanogaster",
  "Apis mellifera",
  "Arabidopsis thaliana",
  "Quercus robur",
  "Ginkgo biloba",
  "Welwitschia mirabilis",
  "Saccharomyces cerevisiae",
  "Escherichia coli",
  "Bacillus subtilis",
  "Plasmodium falciparum"
)
mat <- distance_matrix(taxa)
print(mat)
```

The matrix is symmetric with zeros on the diagonal. 

### Built-in Clustering and Ordination

`taxodist` has wrapper functions for basic multivariate exploration of these matrices, saving you from manually writing boilerplate code:

#### Hierarchical Clustering

```{r clustering}
cl <- taxo_cluster(taxa)
plot(cl)
```

#### PCoA

```{r PCoA}
ord <- taxo_ordinate(taxa)
summary(ord)
plot(ord)
```

#### Heatmap visualization

```{r}
taxo_heatmap(taxa)
```

### Finding the closest relative

```{r closest}
closest_relative(
  "Carnotaurus",
  c("Aucasaurus", "Velociraptor", "Triceratops",
    "Brachiosaurus", "Homo sapiens", "Apis mellifera")
)
#>            taxon   distance
#> 1     Aucasaurus 0.01515152
#> 2   Velociraptor 0.01666667
#> 4  Brachiosaurus 0.01754386
#> 3    Triceratops 0.01818182
#> 5   Homo sapiens 0.02777778
#> 6 Apis mellifera 0.06666667
```

---

## Lineage utilities

### Comparing lineages side by side

```{r compare}
compare_lineages("Carnotaurus", "Tyrannosaurus")
#> -- Lineage Comparison --
#> MRCA: Averostra at depth 60
#>
#> Shared lineage (60 nodes):
#>   Biota ... Theropoda
#>
#> Carnotaurus only (7 nodes):
#> Ceratosauria
#> Neoceratosauria
#> Abelisauroidea
#> Abelisauria
#> Abelisauridae
#> Carnotaurinae
#> Carnotaurus
#>
#> Tyrannosaurus only (10 nodes):
#> Tetanurae
#> Orionides
#> ...
```

### Listing shared clades

```{r shared}
# what do a fly and a beetle have in common?
shared_clades("Drosophila melanogaster", "Tribolium castaneum")
# returns their shared lineage from Biota down to their MRCA

# what do T. rex and a rose share?
shared_clades("Tyrannosaurus rex", "Rosa agrestis")
```

### Testing clade membership

```{r membership}
is_member("Tyrannosaurus", "Theropoda")          # TRUE
is_member("Carnotaurus", "Abelisauridae")        # TRUE
is_member("Triceratops", "Theropoda")            # FALSE
is_member("Homo sapiens", "Amniota")             # TRUE
is_member("Drosophila melanogaster", "Insecta")  # TRUE
is_member("Quercus robur", "Animalia")           # FALSE
```

### Filtering a list of taxa by clade

```{r filter}
taxa <- c("Tyrannosaurus", "Carnotaurus", "Triceratops",
          "Velociraptor", "Homo sapiens", "Drosophila melanogaster",
          "Quercus robur", "Saccharomyces cerevisiae")

filter_clade(taxa, "Dinosauria")
#> [1] "Tyrannosaurus" "Carnotaurus"   "Triceratops"   "Velociraptor"

filter_clade(taxa, "Theropoda")
#> [1] "Tyrannosaurus" "Carnotaurus"   "Velociraptor"

filter_clade(taxa, "Animalia")
#> [1] "Tyrannosaurus"          "Carnotaurus"
#> [3] "Triceratops"            "Velociraptor"
#> [5] "Homo sapiens"           "Drosophila melanogaster"
```

---

## Coverage and caching

### Checking coverage before a large run

```{r coverage}
taxa <- c("Tyrannosaurus", "Velociraptor", "Apis mellifera", "Fakeosaurus")
check_coverage(taxa)
#>  Tyrannosaurus  Velociraptor  Apis mellifera    Fakeosaurus
#>           TRUE          TRUE            TRUE          FALSE
```

Use `check_coverage()` to pre-screen a list before running `distance_matrix()`
on a large dataset — taxa that return `FALSE` will produce `NA` distances.

### Caching

Lineages are automatically cached in memory during an R session to avoid
redundant network requests. This means the second call to `get_lineage()` for
the same taxon is instantaneous. Clear the cache with:

```{r cache}
clear_cache()
```

### Persisting the cache across sessions

By default the cache lives only in memory and is lost when you close R.
If you have retrieved lineages for a large set of taxa, you can save the
cache to disk and reload it in a future session to avoid hitting the network
again:

```{r save-cache}
save_cache("my_taxa_cache.rds") # at the end of a session

load_cache("my_taxa_cache.rds") # at the start of the next session, before any distance calls
```

This is especially useful before running `distance_matrix()` on a large
dataset.

---

## A note on lineage depth

The Taxonomicon provides substantially deeper lineage resolution than most
other programmatic sources. For example, *Tyrannosaurus* has 70 nodes
in its lineage, capturing intermediate clades at the level of superfamilies,
tribes, and named subclades that are absent from most sources. This depth is 
what makes the distance metric meaningful, shallower sources would produce 
coarser distances that conflate distantly related groups.

---

## Data source and citation

All lineage data is sourced from The Taxonomicon (taxonomy.nl), based on
*Systema Naturae 2000*:

> Brands, S.J. (1989 onwards). *Systema Naturae 2000*. Amsterdam,
> The Netherlands. Retrieved from The Taxonomicon,
> http://taxonomicon.taxonomy.nl.

Please cite this resource in any published work using `taxodist`.

---
title: "Statistical Applications of taxodist"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Statistical Applications of taxodist}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.align = "center"
)

library(taxodist)

vegan_available <- requireNamespace("vegan", quietly = TRUE)
ape_available <- requireNamespace("ape", quietly = TRUE)

if (is.null(taxobase$statistical_matrix)) {
  stop(
    "The installed taxobase object does not contain statistical_matrix. ",
    "Rebuild data/taxobase.rda before building this vignette."
  )
}
```

## Overview

A distance matrix produced by `taxodist` is a base R `dist` object. It can
therefore be supplied to clustering, ordination, permutation tests, and
taxonomic-diversity functions that accept pairwise dissimilarities.

These analyses operate on distances derived from classification depth. They do
not convert the source hierarchy into evolutionary time, genetic distance, or
phylogenetic branch length. Results should be described as analyses of
taxonomic hierarchy distance.

This vignette uses the reference matrix stored in `taxobase`. The matrix is
built in advance from The Taxonomicon and permits the vignette to run without
network access.

```{r data}
taxa <- taxobase$statistical_taxa
mat <- taxobase$statistical_matrix

taxobase$metadata
length(taxa)
inherits(mat, "dist")
identical(attr(mat, "Labels"), taxa)
```

The example contains 15 mammalian genera representing several broad taxonomic
groups. The selection is illustrative and is not intended to constitute a
formal ecological sample or a conservation-status dataset.

```{r taxa}
taxa
```

## Inspecting the distance matrix

```{r matrix-summary}
summary(as.vector(mat))
range(mat)

round(
  as.matrix(mat)[1:6, 1:6],
  digits = 5
)
```

For this dataset, pairwise distances range from approximately 0.01370 to
0.01695. Smaller values indicate a deeper MRCA in the source hierarchy.

The absolute scale depends on classification resolution. A value such as 0.014
should not be interpreted as a proportion of evolutionary divergence.

## Hierarchical clustering

```{r clustering, fig.width=7, fig.height=5}
clustering <- taxo_cluster(mat, method = "average")

clustering$hclust
summary(clustering$hclust$height)

plot(
  clustering,
  main = "Average-linkage clustering of taxonomic hierarchy distances",
  xlab = "",
  sub = ""
)
```

Average-linkage clustering summarizes similarities in the supplied matrix. The
topology and node heights also depend on the selected linkage method.

The resulting dendrogram is not an independently inferred phylogenetic tree.
It contains no substitution model, fossil calibration, branch-length
estimation, or statistical reconstruction of evolutionary history.

### Tree representation with ape

An `hclust` object can be converted to the `phylo` data structure used by
`ape`. This can be useful for plotting and file-format interoperability.

```{r ape-tree, eval=ape_available, fig.width=7, fig.height=5}
tree <- ape::as.phylo(clustering$hclust)

plot(
  tree,
  main = "Tree representation of a taxonomic distance dendrogram",
  cex = 0.8
)
```

Conversion to class `phylo` changes the representation of the clustering
result, not its scientific interpretation. The object remains derived from a
classification-based distance matrix.

Export can be performed when required:

```{r ape-export, eval=FALSE}
ape::write.tree(tree, file = "taxonomic-distance-dendrogram.nwk")
```

## Principal coordinates analysis

Principal coordinates analysis, or classical multidimensional scaling,
represents pairwise dissimilarities as coordinates in a Euclidean space
(Gower, 1966).

```{r pcoa}
ordination <- taxo_ordinate(mat, k = 2)

positive_eigenvalues <- ordination$eig[ordination$eig > 0]
variance_percent <- 100 * positive_eigenvalues / sum(positive_eigenvalues)

ordination_summary <- data.frame(
  Axis = c("PC1", "PC2"),
  Eigenvalue = ordination$eig[1:2],
  Variance_percent = variance_percent[1:2],
  Cumulative_percent = cumsum(variance_percent)[1:2]
)

ordination_summary
100 * ordination$GOF[1]
sum(ordination$eig < -sqrt(.Machine$double.eps))
```

```{r pcoa-plot, fig.width=7, fig.height=5}
plot(
  ordination,
  main = "PCoA of taxonomic hierarchy distances"
)
```

The first two axes explain approximately 11.46% and 9.06% of the positive
eigenvalue total, respectively. Their cumulative representation is therefore
about 20.52%.

Consequently, the two-dimensional plot is a limited projection of the complete
distance structure. Proximity in this graph should not be interpreted without
considering its relatively low goodness-of-fit.

No materially negative eigenvalues were observed for this matrix. This does
not imply that every matrix produced from missing, disconnected, or differently
processed lineages will have the same property.

## Taxonomic diversity and distinctness

Clarke and Warwick (1998, 2001) developed indices based on averaged taxonomic
distances among species or individuals. `vegan::taxondive()` accepts a
community matrix together with a `dist` object.

The following communities are hypothetical presence-absence assemblages. They
are constructed only to demonstrate the interface.

```{r community}
comm <- matrix(
  c(
    1,1,0,0,1,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,1,1,1,1,1
  ),
  nrow = 3,
  byrow = TRUE,
  dimnames = list(
    c("community_A", "community_B", "community_C"),
    taxa
  )
)

comm
```

```{r taxondive, eval=vegan_available}
taxonomic_diversity <- vegan::taxondive(comm, mat)

taxonomic_diversity_table <- data.frame(
  Species = taxonomic_diversity$Species,
  Delta = taxonomic_diversity$D,
  Delta_star = taxonomic_diversity$Dstar,
  Lambda_plus = taxonomic_diversity$Lambda,
  Delta_plus = taxonomic_diversity$Dplus,
  SD_Delta_plus = taxonomic_diversity$sd.Dplus
)

taxonomic_diversity_table
```

Here:

- \(\Delta\) is taxonomic diversity and incorporates abundances;
- \(\Delta^*\) is taxonomic distinctness conditional on individuals belonging
  to different taxa;
- \(\Delta^+\) is average taxonomic distinctness for presence-absence data;
- \(\Lambda^+\) describes variation in pairwise taxonomic distinctness.

Because these examples use presence-absence data, \(\Delta\), \(\Delta^*\),
and \(\Delta^+\) coincide within each community.

`community_A` has the largest mean distance, approximately 0.01659.
`community_B` and `community_C` have values of approximately 0.01539 and
0.01527. These differences describe only the deliberately constructed example
communities.

Traditional applications often calculate taxonomic distances from integer or
weighted steps through standardized ranks. The distances supplied here instead
use reciprocal MRCA depth. `taxondive()` permits other dissimilarities, but the
resulting values are specific to this definition and should not be compared
numerically with conventional Clarke-Warwick indices without accounting for
the different distance scale. Multiplying all distances by a constant also
rescales the Delta indices and rescales \(\Lambda^+\) by the square of that
constant.

## Mantel test

The Mantel test evaluates association between two distance matrices using
permutations (Mantel, 1967). Both matrices must refer to the same observations
in the same order.

The coordinates below are simulated and have no biological interpretation.

```{r mantel, eval=vegan_available}
set.seed(42)

coordinates <- matrix(
  rnorm(2 * length(taxa)),
  ncol = 2,
  dimnames = list(taxa, c("x", "y"))
)

geographic_distance <- stats::dist(coordinates)

mantel_result <- vegan::mantel(
  mat,
  geographic_distance,
  method = "pearson",
  permutations = 999
)

mantel_result
```

The simulated example gives a Mantel correlation of approximately -0.056 and a
permutation p-value of approximately 0.655. It therefore provides no evidence
of association between taxonomic and simulated geographic distance.

This is the expected interpretation of this example: the coordinates were
generated randomly. It is not evidence for or against a biological
biogeographic process.

In an empirical analysis, the permutation design must reflect the sampling
design. Spatial dependence, repeated measurements, blocks, and nested sampling
can invalidate unrestricted permutations.

## PERMANOVA

PERMANOVA partitions variation in a distance matrix and assesses a pseudo-\(F\)
statistic using permutations (Anderson, 2001).

The groups below are illustrative taxonomic categories. They are uneven in
size, and some contain only one observation. They are therefore suitable for
demonstrating the software interface but not for drawing substantive
inferential conclusions.

```{r groups}
groups <- factor(c(
  "xenarthran", "xenarthran",
  "carnivoran",
  "ungulate",
  "marsupial",
  "primate", "primate",
  "carnivoran", "carnivoran", "carnivoran",
  "cetacean", "cetacean",
  "sirenian",
  "ungulate", "ungulate"
))

table(groups)
```

```{r permanova, eval=vegan_available}
set.seed(42)

permanova_result <- vegan::adonis2(
  mat ~ groups,
  permutations = 999
)

permanova_result
```

For these constructed groups, the model accounts for approximately 52.44% of
the total sum of squares and gives a permutation p-value of 0.001.

That result must not immediately be interpreted as evidence that group
centroids differ. PERMANOVA can respond to differences in group location,
within-group dispersion, or both, particularly in an unbalanced design.

## Multivariate dispersion

Distance-based analysis of multivariate dispersion examines distances from
observations to their group centroid or spatial median (Anderson, 2006).
It provides an important companion analysis for PERMANOVA.

```{r dispersion, eval=vegan_available}
dispersion <- vegan::betadisper(mat, groups)

anova(dispersion)

set.seed(42)
vegan::permutest(
  dispersion,
  permutations = 999
)
```

The ordinary analysis of distances to group centers is highly significant in
this example. Thus the groups differ strongly in dispersion.

The PERMANOVA result is consequently confounded with dispersion differences and
cannot be presented as an unambiguous difference among group centroids.
Moreover, singleton and very small groups make this example unsuitable for
formal biological inference.

A real study should use independently justified groups, adequate replication,
and a permutation structure consistent with the sampling design.

## Complete live workflow

The following workflow computes a new matrix from taxon names. It is not
evaluated while the vignette is built because it requires access to The
Taxonomicon.

```{r live-workflow, eval=FALSE}
library(taxodist)

taxa <- c(
  "Tyrannosaurus",
  "Velociraptor",
  "Spinosaurus",
  "Allosaurus"
)

coverage <- check_coverage(taxa)
stopifnot(all(coverage))

mat <- distance_matrix(taxa)

clustering <- taxo_cluster(mat, method = "average")
ordination <- taxo_ordinate(mat, k = 2)

plot(clustering)
plot(ordination)
```

## Reporting recommendations

Analyses based on `taxodist` should report:

- the package version;
- the lineage retrieval or dataset-generation date;
- The Taxonomicon as the classification source;
- the distance definition;
- decisions involving ambiguous or missing taxa;
- the clustering linkage or ordination method;
- the number and structure of permutations;
- checks of multivariate dispersion when PERMANOVA is used;
- the distinction between hierarchy distance and evolutionary branch length.

The exact package and data-source citations are available with:

```{r citations, eval=FALSE}
citation("taxodist")
citation("vegan")
citation("ape")
```

## References

Anderson, M.J. (2001). A new method for non-parametric multivariate
analysis of variance. *Austral Ecology*, 26, 32--46.
doi:10.1046/j.1442-9993.2001.01070.x.

Anderson, M.J. (2006). Distance-based tests for homogeneity of multivariate
dispersions. *Biometrics*, 62, 245--253.
doi:10.1111/j.1541-0420.2005.00440.x.

Brands, S.J. (1989 onwards). *Systema Naturae 2000*. Amsterdam,
The Netherlands. The Taxonomicon.

Clarke, K.R. and Warwick, R.M. (1998). A taxonomic distinctness index and
its statistical properties. *Journal of Applied Ecology*, 35, 523--531.
doi:10.1046/j.1365-2664.1998.3540523.x.

Clarke, K.R. and Warwick, R.M. (2001). A further biodiversity index
applicable to species lists: variation in taxonomic distinctness.
*Marine Ecology Progress Series*, 216, 265--278.
doi:10.3354/meps216265.

Gower, J.C. (1966). Some distance properties of latent root and vector
methods used in multivariate analysis. *Biometrika*, 53, 325--338.
doi:10.1093/biomet/53.3-4.325.

Mantel, N. (1967). The detection of disease clustering and a generalized
regression approach. *Cancer Research*, 27, 209--220.

Paradis, E. and Schliep, K. (2019). ape 5.0: an environment for modern
phylogenetics and evolutionary analyses in R. *Bioinformatics*, 35,
526--528. doi:10.1093/bioinformatics/bty633.

---
title: "Methodological Notes and Limitations"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Methodological Notes and Limitations}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(taxodist)
```

## Scope

`taxodist` represents each taxon as an ordered lineage from the root of The
Taxonomicon classification to the queried node. The resulting distance is a
property of that classification hierarchy. It is not an estimate of divergence
time, genetic distance, or phylogenetic branch length.

## Definition

Let \(L_A\) and \(L_B\) be two ordered lineages. Let \(h(A,B)\) be the length of
their continuous common prefix, equivalent to the depth of their most recent
common ancestor (MRCA). The distance is

\[
d(A,B) =
\begin{cases}
0, & L_A = L_B, \\
1/h(A,B), & L_A \ne L_B.
\end{cases}
\]

If two lineages have no shared root, the package returns an infinite distance.
Names repeated after the first divergence are not considered evidence of shared
ancestry.

Zero is reserved for the same hierarchy node. Consequently, an ancestor and
one of its descendants have positive distance even though they are connected by
ancestry. Membership and distance answer different questions: `is_member()` and
`taxo_path()` should be used for containment and ancestry queries.

## Ultrametric property

For lineages belonging to one connected hierarchy, the distance is
non-negative, symmetric, and satisfies identity of indiscernibles. It also
satisfies the strong triangle inequality

\[
d(A,C) \leq \max\{d(A,B), d(B,C)\}.
\]

The common prefix of \(A\) and \(C\) must be at least as long as the shorter of
the common prefixes of \((A,B)\) and \((B,C)\). Therefore,

\[
h(A,C) \geq \min\{h(A,B),h(B,C)\}.
\]

Taking reciprocals gives the strong triangle inequality.

The property can also be checked on the offline reference matrix distributed
with the package.

```{r ultrametric-check}
m <- as.matrix(taxobase$statistical_matrix)
tolerance <- sqrt(.Machine$double.eps)

ultrametric_ok <- function(index) {
  i <- index[1]
  j <- index[2]
  k <- index[3]

  d_ij <- m[i, j]
  d_ik <- m[i, k]
  d_jk <- m[j, k]

  d_ij <= max(d_ik, d_jk) + tolerance &&
    d_ik <= max(d_ij, d_jk) + tolerance &&
    d_jk <= max(d_ij, d_ik) + tolerance
}

all(combn(seq_len(nrow(m)), 3, FUN = ultrametric_ok))
```

## Dependence on the source classification

The numerical value depends on the number and arrangement of nodes returned by
The Taxonomicon. Taxonomic revisions, alternative classifications, and
differences in lineage resolution can change both MRCA depth and distance.

Comparisons should therefore use lineages obtained from the same source and,
where possible, the same package and data-retrieval version.

## Ambiguous and missing taxa

Some names correspond to more than one valid biological entry. A numeric
Taxonomicon identifier can be supplied when a specific taxonomic concept is
required.

Missing lineages produce missing pairwise values. Lineages without a common
root produce infinite values. These cases should be resolved before applying
methods that require a complete finite matrix, including clustering and
ordination.

## Interpretation

Smaller values indicate a deeper shared node in the source classification.
They should not be interpreted as:

- elapsed evolutionary time;
- genetic or morphological divergence;
- phylogenetic branch length;
- probability of common ancestry.

Hierarchical clustering of these distances produces a taxonomic similarity
dendrogram, not an independently inferred phylogenetic tree.

## Recommended reporting

Analyses using `taxodist` should report:

- the package version;
- the lineage retrieval date;
- The Taxonomicon as the source classification;
- the distance definition;
- decisions involving ambiguous or missing taxa;
- any clustering, ordination, or permutation method subsequently applied.

Formatted citations for the package and The Taxonomicon are available through
`citation("taxodist")`.

---
title: "Introduction to taxodist"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to taxodist}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.align = "center"
)

library(taxodist)

if (is.null(taxobase$matrix)) {
  stop(
    "The installed taxobase object does not contain the reference matrix. ",
    "Rebuild data/taxobase.rda before building this vignette."
  )
}
```

## Overview

`taxodist` retrieves ordered taxonomic lineages from The Taxonomicon and uses
them to compare named nodes in a classification hierarchy.

The package provides functions for:

- lineage retrieval;
- most-recent-common-ancestor queries;
- taxonomic hierarchy distances;
- clade membership and filtering;
- paths between hierarchy nodes;
- pairwise distance matrices;
- clustering and principal coordinates analysis;
- session and persistent cache management.

Inputs may represent species, genera, families, orders, named clades, or other
nodes available in the source classification.

## What the distance represents

Let \(L_A\) and \(L_B\) be the ordered lineages of two hierarchy nodes. Let
\(h(A,B)\) be the depth of their most recent common ancestor, defined from their
continuous common prefix.

The distance is

\[
d(A,B) =
\begin{cases}
0, & L_A = L_B, \\
1/h(A,B), & L_A \ne L_B.
\end{cases}
\]

A deeper shared ancestor produces a smaller distance. Within a connected
hierarchy, finite distances lie between zero and one.

If two lineages have no shared root, the package returns an infinite distance.
Missing lineages instead produce missing pairwise values in a distance matrix.

The measure is an ultrametric within each connected hierarchy. Its mathematical
properties and assumptions are discussed in:

```{r methodology-link, eval=FALSE}
vignette("methodological-notes", package = "taxodist")
```

## Distance and membership are different

A descendant belongs to each clade represented among its ancestors. However,
the descendant and ancestor remain distinct hierarchy nodes.

For example, *Tyrannosaurus* is a member of *Dinosauria*, but the distance
between the two nodes is positive:

\[
d(\textit{Tyrannosaurus},\textit{Dinosauria})
=
\frac{1}{\operatorname{depth}(\textit{Dinosauria})}.
\]

Use:

- `taxo_distance()` to compare hierarchy nodes;
- `is_member()` to test containment within a clade;
- `taxo_path()` to inspect the complete ancestry path between nodes.

The distance represents classification depth. It is not evolutionary time,
genetic distance, morphological divergence, or phylogenetic branch length.

## Reproducible packaged examples

The package includes `taxobase`, a reference object generated from The
Taxonomicon. It allows documentation and examples to run without contacting an
external service.

```{r taxobase}
names(taxobase)
taxobase$metadata
```

The metadata records the source, generation date, package version, and distance
definition used to construct the object.

### Stored lineages

```{r stored-lineages}
tail(taxobase$lineage_tyrannosaurus)
tail(taxobase$lineage_homo)
```

Lineages are ordered from the root of the source classification to the queried
node.

### Stored pairwise comparison

```{r stored-pairwise}
taxobase$pairwise
```

A `taxodist_result` records:

- the distance;
- the MRCA and its depth;
- the depth of each complete lineage;
- the two queried names.

```{r result-structure}
class(taxobase$pairwise)
names(taxobase$pairwise)
```

### Reference distance matrix

```{r reference-matrix}
reference_matrix <- taxobase$matrix

inherits(reference_matrix, "dist")
attr(reference_matrix, "Size")
head(attr(reference_matrix, "Labels"))

round(
  as.matrix(reference_matrix)[1:6, 1:6],
  digits = 4
)
```

A matrix returned by `distance_matrix()` is a standard R `dist` object. It is
symmetric, has a zero diagonal, and can be supplied to functions accepting
pairwise dissimilarities.

## Live lineage queries

The following examples contact The Taxonomicon and are not evaluated while the
vignette is built.

### Retrieving a lineage

```{r live-lineage, eval=FALSE}
get_lineage("Tyrannosaurus")
get_lineage("Drosophila melanogaster")
```

Numeric Taxonomicon identifiers can also be supplied directly:

```{r lineage-id, eval=FALSE}
get_lineage("67263")
```

Using an identifier is useful when a name corresponds to more than one valid
taxonomic entry.

### Computing a pairwise distance

```{r live-distance, eval=FALSE}
result <- taxo_distance(
  "Tyrannosaurus",
  "Velociraptor"
)

result
result$distance
result$mrca
```

### Finding the MRCA

```{r live-mrca, eval=FALSE}
mrca("Tyrannosaurus", "Velociraptor")
mrca("Tyrannosaurus", "Triceratops")
mrca("Tyrannosaurus", "Homo")
```

The returned MRCA depends on the classification currently provided by The
Taxonomicon and may change following taxonomic revisions.

### Inspecting the path between nodes

```{r live-path, eval=FALSE}
taxo_path(
  "Tyrannosaurus",
  "Triceratops"
)

taxo_path(
  "Dinosauria",
  "Tyrannosaurus"
)
```

The path ascends from the first node to the MRCA and descends toward the second
node. For an ancestor-descendant pair, the ancestor itself is the MRCA.

## Working with multiple taxa

### Constructing a distance matrix

```{r live-matrix, eval=FALSE}
taxa <- c(
  "Tyrannosaurus",
  "Velociraptor",
  "Spinosaurus",
  "Allosaurus"
)

mat <- distance_matrix(
  taxa,
  progress = TRUE
)

mat
```

Lineages are retrieved once and cached for subsequent comparisons.

### Finding the closest candidates

A stored example is available in `taxobase`:

```{r stored-closest}
taxobase$closest
```

For a live query:

```{r live-closest, eval=FALSE}
closest_relative(
  "Carnotaurus",
  c(
    "Aucasaurus",
    "Velociraptor",
    "Triceratops",
    "Brachiosaurus"
  )
)
```

Candidates are sorted from smallest to largest distance. A missing candidate
is retained with a missing distance.

### Focal distances

```{r live-focal, eval=FALSE}
focal_distances(
  focal = "Tyrannosaurus",
  community = c(
    "Velociraptor",
    "Triceratops",
    "Spinosaurus"
  )
)
```

This produces a table containing the candidate name, distance, MRCA, and MRCA
depth.

## Clade operations

### Testing membership

```{r live-membership, eval=FALSE}
is_member("Tyrannosaurus", "Dinosauria")
is_member("Tyrannosaurus", "Theropoda")
is_member("Tyrannosaurus", "Ornithischia")
```

Matching is case-insensitive, ignores surrounding whitespace, and requires the
complete clade name rather than a partial substring.

### Filtering taxa

The package data include a stored filtering example:

```{r stored-filter}
taxobase$filter
```

A live operation can be performed with:

```{r live-filter, eval=FALSE}
taxa <- c(
  "Tyrannosaurus",
  "Carnotaurus",
  "Triceratops",
  "Velociraptor",
  "Homo",
  "Drosophila"
)

filter_clade(taxa, "Dinosauria")
filter_clade(taxa, "Theropoda")
```

### Shared clades and lineage comparison

```{r live-comparison, eval=FALSE}
shared_clades(
  "Tyrannosaurus",
  "Triceratops"
)

compare_lineages(
  "Carnotaurus",
  "Tyrannosaurus"
)
```

Shared ancestry is determined from the continuous common lineage prefix.
A name repeated after the first divergence is not treated as a shared ancestor.

## Clustering, ordination, and visualization

The following examples use a subset of the packaged matrix and therefore
require no network access.

```{r analysis-matrix}
labels <- attr(reference_matrix, "Labels")[1:8]

example_matrix <- stats::as.dist(
  as.matrix(reference_matrix)[labels, labels]
)

example_matrix
```

### Hierarchical clustering

```{r clustering, fig.width=7, fig.height=5}
clustering <- taxo_cluster(
  example_matrix,
  method = "average"
)

plot(
  clustering,
  main = "Taxonomic hierarchy distance clustering",
  xlab = "",
  sub = ""
)
```

The result is a clustering dendrogram, not an independently inferred
phylogenetic tree.

### Principal coordinates analysis

```{r ordination, fig.width=7, fig.height=5}
ordination <- taxo_ordinate(
  example_matrix,
  k = 2
)

summary(ordination)
plot(
  ordination,
  main = "Taxonomic hierarchy distance space"
)
```

The goodness-of-fit and eigenvalues should be examined before interpreting a
low-dimensional representation.

### Heatmap

```{r heatmap, fig.width=7, fig.height=6}
taxo_heatmap(
  example_matrix,
  main = "Taxonomic hierarchy distance matrix"
)
```

Additional statistical examples and interpretation are provided in:

```{r statistics-link, eval=FALSE}
vignette("statistical-applications", package = "taxodist")
```

## Name resolution and coverage

Taxonomic names may be missing, ambiguous, redirected, or associated with
non-biological homonyms.

`taxo_search()` returns candidate Taxonomicon entries:

```{r live-search, eval=FALSE}
taxo_search("Panthera")
taxo_search("Bacteria")
```

When several valid biological entries exist, the package reports the available
identifiers. A numeric identifier should be supplied when the intended
taxonomic concept is known.

Before calculating a large matrix, coverage can be checked explicitly:

```{r live-coverage, eval=FALSE}
taxa <- c(
  "Tyrannosaurus",
  "Velociraptor",
  "Quercus",
  "Not_a_real_taxon"
)

check_coverage(taxa)
```

Unresolved taxa generate missing pairwise distances. These values should be
examined before clustering, ordination, or other methods requiring complete
finite matrices.

## Cache management

Taxonomicon identifiers and lineages are cached in memory during an R session.
This avoids repeated requests for the same taxa.

```{r live-cache, eval=FALSE}
cache_info()
clear_cache()
```

The cache can also be saved and restored:

```{r persistent-cache, eval=FALSE}
save_cache("taxodist-cache.rds")
clear_cache()
load_cache("taxodist-cache.rds")
```

Cache files preserve retrieved lineages across sessions. They should be
regenerated when an analysis is intended to incorporate subsequent revisions
of the source classification.

## Interpretation and reproducibility

The numerical distance depends on the classification supplied by The
Taxonomicon. Changes in lineage resolution or taxonomic concepts can alter MRCA
depths and therefore alter distances.

A reproducible analysis should record:

- the `taxodist` version;
- the retrieval or dataset-generation date;
- the source classification;
- numeric identifiers used to resolve ambiguous names;
- missing or excluded taxa;
- the distance definition;
- downstream clustering or ordination settings.

Saved caches and distance matrices can preserve the exact input used in an
analysis.

## Data source and citation

All retrieved lineage data originate from The Taxonomicon, based on *Systema
Naturae 2000*.

Formatted references for both the package and the data source are available
with:

```{r citation, eval=FALSE}
citation("taxodist")
```

Both the software and the underlying taxonomic source should be cited in
published analyses.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{shared_clades}
\alias{shared_clades}
\title{List all clades shared between two taxa}
\usage{
shared_clades(taxon_a, taxon_b, verbose = FALSE)
}
\arguments{
\item{taxon_a}{A character string giving the first taxon name.}

\item{taxon_b}{A character string giving the second taxon name.}

\item{verbose}{Logical. If \code{TRUE}, prints progress messages. Default \code{FALSE}.}
}
\value{
A character vector of shared clade names ordered from root to MRCA,
or \code{NULL} if either taxon cannot be found.
}
\description{
Returns the vector of clade names forming the shared trunk of two taxa's
lineages, from root down to (and including) their MRCA.
}
\examples{
\donttest{
shared_clades("Tyrannosaurus", "Velociraptor")
shared_clades("Tyrannosaurus", "Homo")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{get_lineage}
\alias{get_lineage}
\title{Retrieve the full taxonomic lineage of a taxon by name}
\usage{
get_lineage(taxon, clean = TRUE, verbose = FALSE)
}
\arguments{
\item{taxon}{A character string giving the taxon name.}

\item{clean}{Logical. If \code{TRUE} (default), removes philosophical root nodes
and cleans formatting markers.}

\item{verbose}{Logical. If \code{TRUE}, prints progress messages. Default \code{FALSE}.}
}
\value{
A character vector of clade names ordered from root to tip, or
\code{NULL} if the taxon cannot be found.
}
\description{
A convenience wrapper that combines \code{\link[=get_taxonomicon_id]{get_taxonomicon_id()}} and
\code{\link[=get_lineage_by_id]{get_lineage_by_id()}} into a single call. Given a taxon name, returns
its complete lineage from root to tip.
}
\examples{
\donttest{
get_lineage("Tyrannosaurus")
get_lineage("Homo sapiens")
get_lineage("Quercus robur")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{load_cache}
\alias{load_cache}
\title{Load a previously saved taxodist cache from disk}
\usage{
load_cache(file)
}
\arguments{
\item{file}{Path to an \code{.rds} file created by \code{\link[=save_cache]{save_cache()}}.}
}
\value{
Invisibly returns \code{NULL}.
}
\description{
Restores lineage data saved with \code{\link[=save_cache]{save_cache()}} into the current session
cache, avoiding network requests for taxa already retrieved in a previous
session. The file structure is validated before the current cache is
modified.
}
\examples{
\donttest{
tmp <- tempfile(fileext = ".rds")
save_cache(tmp)
load_cache(tmp)
unlink(tmp)
}
}
\seealso{
\code{\link[=save_cache]{save_cache()}}, \code{\link[=clear_cache]{clear_cache()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{get_taxonomicon_id}
\alias{get_taxonomicon_id}
\title{Find the Taxonomicon ID for a taxon name}
\usage{
get_taxonomicon_id(taxon, verbose = FALSE)
}
\arguments{
\item{taxon}{A character string giving the taxon name to search for.
Typically a genus name (e.g., \code{"Tyrannosaurus"}) but species and higher
ranks are also supported.}

\item{verbose}{Logical. If \code{TRUE}, prints status messages during retrieval.
Default is \code{FALSE}.}
}
\value{
A character string containing the Taxonomicon numeric ID, or \code{NULL}
if the taxon is not found.
}
\description{
Queries The Taxonomicon (taxonomy.nl) to retrieve the internal numeric
identifier for a given taxon name. The search filters out non-biological
entities such as astronomical objects that may share the same name.
}
\details{
The function queries the static search endpoint at
\code{taxonomicon.taxonomy.nl/TaxonList.aspx} and parses the resulting HTML
to extract the taxon ID from the hierarchy link. When multiple matches
exist (e.g., a genus name shared with an astronomical object), biological
entries are prioritised by filtering for entries annotated as dinosaur,
reptile, archosaur, animal, plant, fungus, or bacterium.
}
\examples{
\donttest{
get_taxonomicon_id("Tyrannosaurus")   # returns "50841"
get_taxonomicon_id("Homo")
get_taxonomicon_id("Quercus")
}
}
\seealso{
\code{\link[=get_lineage]{get_lineage()}}, \code{\link[=taxo_distance]{taxo_distance()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{taxo_distance}
\alias{taxo_distance}
\title{Compute the taxonomic hierarchy distance between two taxa}
\usage{
taxo_distance(taxon_a, taxon_b, verbose = FALSE)
}
\arguments{
\item{taxon_a}{A character string giving the first taxon name.}

\item{taxon_b}{A character string giving the second taxon name.}

\item{verbose}{Logical. If \code{TRUE}, prints progress messages. Default \code{FALSE}.}
}
\value{
A named list of class \code{"taxodist_result"} with the following elements:
\describe{
\item{\code{distance}}{Numeric. The distance between the two taxa. Returns 0
only when their complete lineages are identical.}
\item{\code{mrca}}{Character. The name of the most recent common ancestor.}
\item{\code{mrca_depth}}{Integer. The depth of the MRCA node.}
\item{\code{depth_a}}{Integer. The lineage depth of taxon A.}
\item{\code{depth_b}}{Integer. The lineage depth of taxon B.}
\item{\code{taxon_a}}{Character. Name of the first taxon.}
\item{\code{taxon_b}}{Character. Name of the second taxon.}
}
Returns \code{NULL} if either taxon cannot be found.
}
\description{
Given two taxon names, retrieves their lineages from The Taxonomicon and
computes a taxonomic distance based on the depth of their most recent
common ancestor (MRCA):
}
\details{
\deqn{d(A,A) = 0}{d(A,A) = 0}
\deqn{d(A,B) = 1 / h(A,B)}{d(A,B) = 1 / h(A,B)}

Here, \code{h(A,B)} is the depth of the MRCA of \code{A} and \code{B}. The deeper the
shared ancestor, the smaller (closer to zero) the distance.
Distinct nodes always have positive distance, including when one taxon is
an ancestor of the other. This makes the measure an ultrametric on each
connected taxonomic hierarchy. Use \code{\link[=is_member]{is_member()}} when the question is
whether one taxon belongs to a clade rather than how far apart the two
hierarchy nodes are.
}
\examples{
\donttest{
# Distance between two theropods
taxo_distance("Tyrannosaurus", "Velociraptor")

# Distance between very distantly related taxa
taxo_distance("Tyrannosaurus", "Quercus")

# Distance between two oviraptorid genera
taxo_distance("Nomingia", "Huanansaurus")
}
}
\references{
Brands, S.J. (1989 onwards). Systema Naturae 2000. Amsterdam, The
Netherlands. Retrieved from The Taxonomicon,
\url{http://taxonomicon.taxonomy.nl}.
}
\seealso{
\code{\link[=mrca]{mrca()}}, \code{\link[=distance_matrix]{distance_matrix()}}, \code{\link[=get_lineage]{get_lineage()}}, \code{\link[=is_member]{is_member()}},
\code{\link[=taxo_path]{taxo_path()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{closest_relative}
\alias{closest_relative}
\title{Find the closest relative of a taxon among a set of candidates}
\usage{
closest_relative(taxon, candidates, verbose = FALSE)
}
\arguments{
\item{taxon}{A character string giving the query taxon name.}

\item{candidates}{A character vector of candidate taxon names to compare
against.}

\item{verbose}{Logical. If \code{TRUE}, prints progress messages. Default \code{FALSE}.}
}
\value{
A data frame with columns \code{taxon} (candidate name) and \code{distance}
(tree metric distance), sorted by distance ascending. Returns \code{NULL} if
the query taxon cannot be found.
}
\description{
Given a query taxon and a vector of candidate taxa, returns the candidate
with the smallest phylogenetic distance to the query.
}
\examples{
\donttest{
closest_relative("Tyrannosaurus",
  c("Velociraptor", "Triceratops", "Brachiosaurus", "Allosaurus"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{clear_cache}
\alias{clear_cache}
\title{Clear the taxodist lineage cache}
\usage{
clear_cache()
}
\value{
Invisibly returns \code{NULL}.
}
\description{
Clears all cached lineages stored in the current R session. Useful when
you suspect cached data is stale or want to force fresh retrieval.
}
\examples{
\donttest{
clear_cache()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{plot.taxodist_ord}
\alias{plot.taxodist_ord}
\title{Plot method for taxodist_ord objects}
\usage{
\method{plot}{taxodist_ord}(
  x,
  main = "Taxonomic Ordination (PCoA)",
  xlab = "PC1",
  ylab = "PC2",
  labels = if (!is.null(x$points)) rownames(x$points) else NULL,
  ...
)
}
\arguments{
\item{x}{A \code{taxodist_ord} object from \code{\link[=taxo_ordinate]{taxo_ordinate()}}.}

\item{main}{Plot title. Default: \code{"Taxonomic Ordination (PCoA)"}.}

\item{xlab}{X-axis label. Default: \code{"PC1"}.}

\item{ylab}{Y-axis label. Default: \code{"PC2"}.}

\item{labels}{Character vector of labels. Default: rownames of points.}

\item{...}{Additional arguments passed to \code{\link[graphics:text]{graphics::text()}}.}
}
\value{
Invisibly returns \code{x}. Called for side effects (plotting).
}
\description{
Plots the PCoA ordination produced by \code{\link[=taxo_ordinate]{taxo_ordinate()}}.
}
\examples{
\donttest{
taxa <- c("Tyrannosaurus", "Velociraptor", "Triceratops", "Brachiosaurus")
ord <- taxo_ordinate(taxa)
plot(ord)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{taxo_cluster}
\alias{taxo_cluster}
\title{Cluster taxa by taxonomic distance}
\usage{
taxo_cluster(taxa, method = "average", ...)
}
\arguments{
\item{taxa}{A character vector of taxon names, or a \code{dist} object from
\code{\link[=distance_matrix]{distance_matrix()}}.}

\item{method}{Clustering method passed to \code{\link[stats:hclust]{stats::hclust()}}. Default
\code{"average"} (UPGMA), which works well with taxonomic distances.}

\item{...}{Additional arguments passed to \code{\link[=distance_matrix]{distance_matrix()}} (e.g.
\code{verbose}, \code{progress}).}
}
\value{
An object of class \code{"taxodist_cluster"} — a list with:
\describe{
\item{\code{hclust}}{The \code{\link[stats:hclust]{stats::hclust()}} result.}
\item{\code{dist}}{The underlying distance matrix.}
}
}
\description{
Computes pairwise taxonomic distances and performs hierarchical clustering.
}
\examples{
\donttest{
taxa <- c("Tyrannosaurus", "Velociraptor", "Triceratops", "Brachiosaurus")
cl <- taxo_cluster(taxa)
if (!is.null(cl$hclust)) {
  plot(cl)
}
}
}
\seealso{
\code{\link[=taxo_ordinate]{taxo_ordinate()}}, \code{\link[=distance_matrix]{distance_matrix()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxodist-package.R
\docType{package}
\name{taxodist-package}
\alias{taxodist-package}
\alias{taxodist}
\title{taxodist: Taxonomic Hierarchy Distances and Lineage Analysis}
\description{
taxodist retrieves taxonomic lineages from The Taxonomicon (taxonomy.nl)
and computes distances from the depth of the most recent common ancestor.
The resulting values describe separation within the source classification;
they are not estimates of evolutionary time or phylogenetic branch length.
\subsection{Core functions}{
\itemize{
\item \code{\link[=get_lineage]{get_lineage()}} — retrieve the full lineage of any taxon
\item \code{\link[=taxo_distance]{taxo_distance()}} — compute the ultrametric distance between two taxa
\item \code{\link[=mrca]{mrca()}} — find the most recent common ancestor
\item \code{\link[=distance_matrix]{distance_matrix()}} — compute all pairwise distances for a set of taxa
\item \code{\link[=closest_relative]{closest_relative()}} — find the closest relative among candidates
\item \code{\link[=compare_lineages]{compare_lineages()}} — print a side-by-side lineage comparison
\item \code{\link[=shared_clades]{shared_clades()}} — list clades shared between two taxa
\item \code{\link[=is_member]{is_member()}} — test clade membership
\item \code{\link[=filter_clade]{filter_clade()}} — filter taxa by clade membership
\item \code{\link[=check_coverage]{check_coverage()}} — check Taxonomicon coverage for a list of taxa
\item \code{\link[=lineage_depth]{lineage_depth()}} — get the lineage depth of a taxon
\item \code{\link[=clear_cache]{clear_cache()}} — clear the session lineage cache
}
}

\subsection{Mathematical background}{

The distance metric is based on the depth of the most recent common
ancestor (MRCA):

\deqn{d(A,A) = 0}{d(A,A) = 0}
\deqn{d(A,B) = 1 / h(A,B)}{d(A,B) = 1 / h(A,B)}

Here, \code{h(A,B)} denotes the depth of the MRCA. The deeper the shared
ancestor, the smaller the distance. Zero is reserved
for identical hierarchy nodes; distinct ancestor-descendant pairs therefore
have positive distance. The measure is an ultrametric on each connected
taxonomic hierarchy. Its numerical values depend on the resolution of the
classification returned by the data source.
}

\subsection{Data source}{

All lineage data is sourced from The Taxonomicon (taxonomy.nl), based on
Systema Naturae 2000 by S.J. Brands (1989 onwards). Please cite this
resource when using taxodist in published work.
}
}
\references{
Brands, S.J. (1989 onwards). Systema Naturae 2000. Amsterdam,
The Netherlands. Retrieved from The Taxonomicon,
\url{http://taxonomicon.taxonomy.nl}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/rodrigosqrt3/taxodist}
  \item Report bugs at \url{https://github.com/rodrigosqrt3/taxodist/issues}
}

}
\author{
\strong{Maintainer}: Rodrigo Fonseca Villa \email{rodrigo03.villa@gmail.com} (\href{https://orcid.org/0009-0005-2938-2270}{ORCID})

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{is_member}
\alias{is_member}
\title{Test whether one taxon is nested within another}
\usage{
is_member(taxon, clade, verbose = FALSE)
}
\arguments{
\item{taxon}{A character string giving the taxon name to test.}

\item{clade}{A character string giving the clade name to test membership in.}

\item{verbose}{Logical. If \code{TRUE}, prints progress messages. Default \code{FALSE}.}
}
\value{
A logical value, or \code{NULL} if the taxon cannot be found.
}
\description{
Returns \code{TRUE} if \code{taxon} is a member of \code{clade} — i.e., if the clade name
appears in the taxon's lineage.
}
\examples{
\donttest{
is_member("Tyrannosaurus", "Theropoda")    # TRUE
is_member("Triceratops", "Theropoda")      # FALSE
is_member("Homo", "Amniota")               # TRUE
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{focal_distances}
\alias{focal_distances}
\title{Compute distances from a focal taxon to a community of taxa}
\usage{
focal_distances(focal, community, verbose = FALSE, progress = TRUE)
}
\arguments{
\item{focal}{A character string giving the focal taxon name.}

\item{community}{A character vector of community taxon names to compare
against. The focal taxon may be included — it will receive a distance
of \code{0}.}

\item{verbose}{Logical. If \code{TRUE}, prints progress messages. Default \code{FALSE}.}

\item{progress}{Logical. If \code{TRUE}, shows a progress bar. Default \code{TRUE}.}
}
\value{
A data frame of class \code{"taxodist_focal"} with columns:
\describe{
\item{\code{taxon}}{Character. Community taxon name.}
\item{\code{distance}}{Numeric. Taxonomic distance to the focal taxon.}
\item{\code{mrca}}{Character. Name of the most recent common ancestor.}
\item{\code{mrca_depth}}{Integer. Depth of the MRCA node.}
}
Rows are sorted by \code{distance} ascending (closest relatives first).
Returns \code{NULL} if the focal taxon cannot be found. An empty community
returns an empty object with the same columns and class.
}
\description{
Given a focal taxon and a vector of community taxa, retrieves all lineages
and computes the taxonomic distance from the focal taxon to each member of
the community. Returns a sorted data frame that also reports the most recent
common ancestor (MRCA) for each pair, making it easy to interpret \emph{why} a
taxon is close or distant.
}
\examples{
\donttest{
community <- c("Velociraptor", "Triceratops", "Brachiosaurus",
               "Allosaurus", "Homo", "Quercus")
focal_distances("Tyrannosaurus", community)
}
}
\seealso{
\code{\link[=closest_relative]{closest_relative()}}, \code{\link[=distance_matrix]{distance_matrix()}}, \code{\link[=taxo_distance]{taxo_distance()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{lineage_depth}
\alias{lineage_depth}
\title{Get the lineage depth of a taxon}
\usage{
lineage_depth(taxon, verbose = FALSE)
}
\arguments{
\item{taxon}{A character string giving the taxon name.}

\item{verbose}{Logical. If \code{TRUE}, prints progress messages. Default \code{FALSE}.}
}
\value{
An integer giving the lineage depth, or \code{NULL} if the taxon cannot
be found.
}
\description{
Returns the number of nodes in the lineage of a taxon, from root to tip.
This reflects how deeply nested the taxon is within the taxonomic hierarchy.
}
\examples{
\donttest{
lineage_depth("Tyrannosaurus")   # deep — many intermediate clades
lineage_depth("Biota")           # shallow — near root
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{taxobase}
\alias{taxobase}
\title{Reference Taxonomic Dataset (taxobase)}
\format{
A structured list containing the following pre-computed components:
\describe{
\item{taxa}{Character vector of 52 candidate genera.}
\item{found_taxa}{Character vector of candidates found in The Taxonomicon.}
\item{coverage}{Named logical vector indicating database presence of each taxon.}
\item{matrix}{A pre-computed taxonomic \code{dist} object for \code{found_taxa}.}
\item{pairwise}{A pre-computed distance result for \emph{Tyrannosaurus} and \emph{Homo}.}
\item{lineage_homo}{Taxonomic lineage returned for \emph{Homo}.}
\item{lineage_tyrannosaurus}{Taxonomic lineage returned for \emph{Tyrannosaurus}.}
\item{closest}{Result of a closest-relative query for \emph{Tyrannosaurus}.}
\item{filter}{Taxa in \code{found_taxa} assigned to Theropoda.}
\item{search}{Reference search result for the term \code{"Bacteria"}.}
\item{statistical_taxa}{Character vector used by the statistical applications vignette.}
\item{statistical_matrix}{Pre-computed \code{dist} object for \code{statistical_taxa}.}
\item{metadata}{Data source, generation date, package version, and distance definition.}
}
}
\source{
\url{http://taxonomicon.taxonomy.nl}
}
\usage{
data(taxobase)
}
\description{
A pre-computed reference dataset for 52 candidate genera distributed across
major taxonomic groups. The data were retrieved from The Taxonomicon using
the package functions and are intended for examples, documentation, and
offline inspection of package output structures.
}
\examples{
data(taxobase)
length(taxobase$found_taxa)
head(labels(taxobase$matrix))
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{cache_info}
\alias{cache_info}
\title{Inspect the current taxodist lineage cache}
\usage{
cache_info()
}
\value{
Invisibly returns a list with elements:
\describe{
\item{\code{n_lineages}}{Integer. Number of cached lineages.}
\item{\code{n_ids}}{Integer. Number of cached taxon IDs.}
\item{\code{taxa}}{Character vector of taxa with cached lineages.}
\item{\code{size_bytes}}{Numeric. Total memory used by the cache.}
}
}
\description{
Reports the number of cached entries, their total memory footprint, and
the names of all taxa whose lineages are stored in the current session.
Useful for understanding what has already been retrieved before running
further computations.
}
\examples{
\donttest{
get_lineage("Tyrannosaurus")
get_lineage("Velociraptor")
cache_info()
}
}
\seealso{
\code{\link[=save_cache]{save_cache()}}, \code{\link[=load_cache]{load_cache()}}, \code{\link[=clear_cache]{clear_cache()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{print.taxodist_result}
\alias{print.taxodist_result}
\title{Print method for taxodist distance results}
\usage{
\method{print}{taxodist_result}(x, ...)
}
\arguments{
\item{x}{A list returned by \code{\link[=taxo_distance]{taxo_distance()}}.}

\item{...}{Additional arguments (ignored).}
}
\value{
Invisibly returns \code{x}. Called for side effects (printing).
}
\description{
Print method for taxodist distance results
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{print.taxodist_path}
\alias{print.taxodist_path}
\title{Print method for taxodist_path objects}
\usage{
\method{print}{taxodist_path}(x, ...)
}
\arguments{
\item{x}{A \code{taxodist_path} object from \code{\link[=taxo_path]{taxo_path()}}.}

\item{...}{Additional arguments (ignored).}
}
\value{
Invisibly returns \code{x}. Called for side effects (printing).
}
\description{
Print method for taxodist_path objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{taxo_search}
\alias{taxo_search}
\title{Search The Taxonomicon for a taxon name}
\usage{
taxo_search(taxon, verbose = FALSE)
}
\arguments{
\item{taxon}{A character string giving the taxon name to search for.}

\item{verbose}{Logical. If \code{TRUE}, prints status messages. Default \code{FALSE}.}
}
\value{
A data frame of class \code{"data.frame"} with columns:
\describe{
\item{\code{id}}{Character. The numeric Taxonomicon ID.}
\item{\code{name}}{Character. The full taxon description, including rank and author.}
}
Returns \code{NULL} if no matches are found.
}
\description{
Queries The Taxonomicon database and returns a data frame of all available
biological entries matching the search string. This is particularly useful
for exploring homonyms, historical ranks, or taxonomic synonyms before
computing distances.
}
\examples{
\donttest{
taxo_search("Bacteria")
taxo_search("Nereis")
taxo_search("Tyrannosaurus")
}
}
\seealso{
\code{\link[=get_lineage]{get_lineage()}}, \code{\link[=taxo_distance]{taxo_distance()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{summary.taxodist_ord}
\alias{summary.taxodist_ord}
\title{Summary method for taxodist_ord objects}
\usage{
\method{summary}{taxodist_ord}(object, ...)
}
\arguments{
\item{object}{A \code{taxodist_ord} object from \code{\link{taxo_ordinate}}.}

\item{...}{Additional arguments (ignored).}
}
\value{
Invisibly returns a data frame containing the eigenvalues and
variance explained by each dimension.
}
\description{
Prints a summary of the PCoA ordination, including the goodness-of-fit
and the proportion of variance explained by each principal coordinate.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{check_coverage}
\alias{check_coverage}
\title{Check whether a taxon is covered by The Taxonomicon}
\usage{
check_coverage(taxa, verbose = FALSE)
}
\arguments{
\item{taxa}{A character vector of one or more taxon names.}

\item{verbose}{Logical. If \code{TRUE}, prints progress messages. Default \code{FALSE}.}
}
\value{
A named logical vector. \code{TRUE} indicates the taxon was found,
\code{FALSE} indicates it was not.
}
\description{
Queries The Taxonomicon for a taxon name and returns a logical indicating
whether the taxon was found. Useful for pre-screening a list of names
before running distance computations.
}
\examples{
\donttest{
check_coverage(c("Tyrannosaurus", "Velociraptor", "Fakeosaurus"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{taxo_ordinate}
\alias{taxo_ordinate}
\title{Ordinate taxa in taxonomic distance space}
\usage{
taxo_ordinate(taxa, k = 2, ...)
}
\arguments{
\item{taxa}{A character vector of taxon names, or a \code{dist} object from
\code{\link[=distance_matrix]{distance_matrix()}}.}

\item{k}{Number of dimensions. Default \code{2}.}

\item{...}{Additional arguments passed to \code{\link[=distance_matrix]{distance_matrix()}}.}
}
\value{
An object of class \code{"taxodist_ord"} — a list with:
\describe{
\item{\code{points}}{A matrix of coordinates (taxa x k dimensions).}
\item{\code{dist}}{The underlying distance matrix.}
\item{\code{GOF}}{Goodness-of-fit from\code{\link[stats:cmdscale]{stats::cmdscale()}}.}
\item{\code{eig}}{The eigenvalues computed during PCoA.}
}
}
\description{
Computes pairwise taxonomic distances and applies classical multidimensional
scaling (PCoA) to project taxa into a low-dimensional space.
}
\examples{
\donttest{
taxa <- c("Tyrannosaurus", "Velociraptor", "Triceratops", "Brachiosaurus")
ord <- taxo_ordinate(taxa)
if (!is.null(ord$points)) {
  plot(ord$points, type = "n")
  text(ord$points, labels = rownames(ord$points))
}
}
}
\seealso{
\code{\link[=taxo_cluster]{taxo_cluster()}}, \code{\link[=distance_matrix]{distance_matrix()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{taxo_path}
\alias{taxo_path}
\title{Get the taxonomic path between two taxa}
\usage{
taxo_path(taxon_a, taxon_b, verbose = FALSE)
}
\arguments{
\item{taxon_a}{A character string giving the first taxon name.}

\item{taxon_b}{A character string giving the second taxon name.}

\item{verbose}{Logical. If \code{TRUE}, prints progress messages. Default \code{FALSE}.}
}
\value{
A data frame of class \code{"taxodist_path"} with columns:
\describe{
\item{\code{node}}{Character. The clade or taxon name at this step.}
\item{\code{depth}}{Integer. The depth of this node in the full lineage of
its side (or the shared lineage for the MRCA).}
\item{\code{direction}}{Character. One of \code{"a"} (ascending from taxon A to
MRCA), \code{"mrca"} (the shared ancestor), or \code{"b"} (descending from MRCA
to taxon B).}
}
Returns \code{NULL} if either taxon cannot be found or if their lineages have no
common ancestor.
}
\description{
Returns the full node-by-node path from one taxon up to their most recent
common ancestor (MRCA) and back down to the other taxon. The result is a
data frame with one row per node, making it easy to inspect, filter, or
pipe into other functions.
}
\examples{
\donttest{
taxo_path("Tyrannosaurus", "Velociraptor")
taxo_path("Tyrannosaurus", "Homo")
}
}
\seealso{
\code{\link[=mrca]{mrca()}}, \code{\link[=shared_clades]{shared_clades()}}, \code{\link[=compare_lineages]{compare_lineages()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{compare_lineages}
\alias{compare_lineages}
\title{Compare lineages of two taxa side by side}
\usage{
compare_lineages(taxon_a, taxon_b, verbose = FALSE)
}
\arguments{
\item{taxon_a}{A character string giving the first taxon name.}

\item{taxon_b}{A character string giving the second taxon name.}

\item{verbose}{Logical. If \code{TRUE}, prints progress messages. Default \code{FALSE}.}
}
\value{
Invisibly returns a list with elements \code{lineage_a}, \code{lineage_b},
and \code{mrca_depth}.
}
\description{
Prints the lineages of two taxa aligned at their most recent common
ancestor, making the point of divergence easy to identify.
}
\examples{
\donttest{
compare_lineages("Tyrannosaurus", "Velociraptor")
compare_lineages("Tyrannosaurus", "Triceratops")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{print.taxodist_focal}
\alias{print.taxodist_focal}
\title{Print method for taxodist_focal objects}
\usage{
\method{print}{taxodist_focal}(x, ...)
}
\arguments{
\item{x}{A \code{taxodist_focal} object from \code{\link[=focal_distances]{focal_distances()}}.}

\item{...}{Additional arguments (ignored).}
}
\value{
Invisibly returns \code{x}. Called for side effects (printing).
}
\description{
Print method for taxodist_focal objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{mrca}
\alias{mrca}
\title{Compute the most recent common ancestor of two taxa}
\usage{
mrca(taxon_a, taxon_b, verbose = FALSE)
}
\arguments{
\item{taxon_a}{A character string giving the first taxon name.}

\item{taxon_b}{A character string giving the second taxon name.}

\item{verbose}{Logical. If \code{TRUE}, prints progress messages. Default \code{FALSE}.}
}
\value{
A character string giving the name of the MRCA, or \code{NULL} if
either taxon cannot be found or no common ancestor exists.
}
\description{
Retrieves the lineages of two taxa and returns the name of their most
recent common ancestor (MRCA) — the deepest node shared by both lineages.
}
\examples{
\donttest{
mrca("Tyrannosaurus", "Velociraptor")   # "Tyrannoraptora"
mrca("Tyrannosaurus", "Triceratops")    # "Dinosauria"
mrca("Tyrannosaurus", "Homo")           # "Amniota"
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{plot.taxodist_cluster}
\alias{plot.taxodist_cluster}
\title{Plot method for taxodist_cluster objects}
\usage{
\method{plot}{taxodist_cluster}(
  x,
  main = "Taxonomic Clustering",
  xlab = "",
  sub = paste("Method:", if (!is.null(x$hclust)) x$hclust$method else "None"),
  ...
)
}
\arguments{
\item{x}{A \code{taxodist_cluster} object from \code{\link[=taxo_cluster]{taxo_cluster()}}.}

\item{main}{Plot title. Default: \code{"Taxonomic Clustering"}.}

\item{xlab}{X-axis label. Default: \code{""}.}

\item{sub}{Subtitle. Default: clustering method used.}

\item{...}{Additional arguments passed to \code{\link[stats:hclust]{stats::plot.hclust()}}.}
}
\value{
Invisibly returns \code{x}. Called for side effects (plotting).
}
\description{
Plots the hierarchical clustering dendrogram produced by \code{\link[=taxo_cluster]{taxo_cluster()}}.
}
\examples{
\donttest{
taxa <- c("Tyrannosaurus", "Velociraptor", "Triceratops", "Brachiosaurus")
cl <- taxo_cluster(taxa)
plot(cl)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{save_cache}
\alias{save_cache}
\title{Save the taxodist lineage cache to disk}
\usage{
save_cache(file)
}
\arguments{
\item{file}{Path to the \code{.rds} file to write.}
}
\value{
Invisibly returns \code{NULL}.
}
\description{
Serialises the current session cache to an \code{.rds} file so it can be
restored in a future session with \code{\link[=load_cache]{load_cache()}}. Useful for
reproducibility and for avoiding repeated network requests.
}
\examples{
\donttest{
tmp <- tempfile(fileext = ".rds")
save_cache(tmp)
unlink(tmp)
}
}
\seealso{
\code{\link[=load_cache]{load_cache()}}, \code{\link[=clear_cache]{clear_cache()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{taxo_heatmap}
\alias{taxo_heatmap}
\title{Plot a taxonomic heatmap}
\usage{
taxo_heatmap(taxa, ...)
}
\arguments{
\item{taxa}{A character vector of taxon names, or a \code{dist} object from
\code{\link{distance_matrix}}.}

\item{...}{Additional arguments passed to \code{\link[stats]{heatmap}}.}
}
\value{
Invisibly returns the underlying \code{dist} object.
Called primarily for its side effect (plotting).
}
\description{
Computes pairwise taxonomic distances and plots a heatmap with hierarchical
clustering dendrograms on the margins. Darker/hotter colors typically
represent smaller distances (closer relatives).
}
\examples{
\donttest{
taxa <- c("Tyrannosaurus", "Velociraptor", "Homo", "Panthera", "Quercus")
taxo_heatmap(taxa)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{filter_clade}
\alias{filter_clade}
\title{Filter a vector of taxa to those belonging to a given clade}
\usage{
filter_clade(taxa, clade, verbose = FALSE)
}
\arguments{
\item{taxa}{A character vector of taxon names.}

\item{clade}{A character string giving the clade to filter by.}

\item{verbose}{Logical. If \code{TRUE}, prints progress messages. Default \code{FALSE}.}
}
\value{
A character vector of taxa that are members of the specified clade.
}
\description{
Given a vector of taxon names and a clade name, returns only those taxa
whose lineage includes the specified clade.
}
\examples{
\donttest{
taxa <- c("Tyrannosaurus", "Triceratops", "Velociraptor",
          "Brachiosaurus", "Homo")
filter_clade(taxa, "Theropoda")
filter_clade(taxa, "Dinosauria")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{distance_matrix}
\alias{distance_matrix}
\title{Compute pairwise taxonomic distances for a set of taxa}
\usage{
distance_matrix(taxa, verbose = FALSE, progress = TRUE)
}
\arguments{
\item{taxa}{A character vector of taxon names.}

\item{verbose}{Logical. If \code{TRUE}, prints progress for each pair.
Default \code{FALSE}.}

\item{progress}{Logical. If \code{TRUE}, shows a progress bar. Default \code{TRUE}.}
}
\value{
A symmetric numeric matrix of class \code{"dist"} containing pairwise
distances. Row and column names are set to the input taxon names.
Taxa that could not be found are included with \code{NA} distances.
An empty input returns an empty \code{"dist"} object.
}
\description{
Given a vector of taxon names, computes all pairwise taxonomic hierarchy distances
and returns a symmetric distance matrix. Lineages are cached after first
retrieval to minimise redundant network requests.
}
\examples{
\donttest{
theropods <- c("Tyrannosaurus", "Velociraptor", "Spinosaurus",
               "Allosaurus", "Carnotaurus")
mat <- distance_matrix(theropods)
print(mat)
}
}
\seealso{
\code{\link[=taxo_distance]{taxo_distance()}}, \code{\link[=closest_relative]{closest_relative()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{get_lineage_by_id}
\alias{get_lineage_by_id}
\title{Retrieve the full taxonomic lineage of a taxon}
\usage{
get_lineage_by_id(taxon_id, clean = TRUE, verbose = FALSE)
}
\arguments{
\item{taxon_id}{A numeric or character string giving the Taxonomicon ID.
Obtain this with \code{\link[=get_taxonomicon_id]{get_taxonomicon_id()}}.}

\item{clean}{Logical. If \code{TRUE} (default), removes philosophical root nodes
above \code{Biota} (i.e., Natura, actualia, Mundus, naturalia) and strips
dagger and superscript markers from names.}

\item{verbose}{Logical. If \code{TRUE}, prints status messages. Default \code{FALSE}.}
}
\value{
A character vector of clade names from root to tip, or \code{NULL} if
retrieval fails.
}
\description{
Given a Taxonomicon numeric ID, retrieves and parses the complete
hierarchical lineage from root (Natura) to the taxon itself. The lineage
is returned as a character vector ordered from root to tip.
}
\details{
Lineage data is sourced from The Taxonomicon, which is based on
Systema Naturae 2000 (Brands, S.J., 1989 onwards). Lineage resolution varies
among taxonomic groups and may change when the source classification is
updated. Because hierarchy distances depend on lineage depth, analyses
should record the retrieval date and package version.
}
\examples{
\donttest{
id <- get_taxonomicon_id("Tyrannosaurus")
lin <- get_lineage_by_id(id)
print(lin)
}
}
\seealso{
\code{\link[=get_lineage]{get_lineage()}}, \code{\link[=taxo_distance]{taxo_distance()}}
}
